{-# LANGUAGE DeriveDataTypeable, FlexibleContexts, FlexibleInstances, GeneralizedNewtypeDeriving,
             MultiParamTypeClasses, TypeSynonymInstances, TemplateHaskell, UndecidableInstances #-}
{-# OPTIONS -Wwarn -fno-warn-orphans -fno-warn-missing-signatures -fno-warn-name-shadowing #-}
-- |It would be nice to move these symbols to Appraisal.ReportStore,
-- but the full name Appraisal.Report.Types.ReportStore appears in the
-- checkpoints file (as does Appraisal.Report.State.State.)
module Test.Report where

--import Appraisal.HtmlForm(FormLabel(toLabel, ofLabels), rdLabels)
--import Appraisal.Permissions (Permissions, Ownable(..))
--import Appraisal.Report.Item(Item)
--import Appraisal.Report.Params(Params(topDir))
import Data.Generics
import Data.Map(Map)
import Data.Maybe(Maybe, maybe)
import Document.Conversion(Conversion(..))
import Document.Wiki(Wiki(NoWiki, NewLine, Text, Wiki))
import Document.Wiki.Conversion ()    -- Instances only
import Extra.Unicode(Unicode)
import Happstack.Data (Default(..), deriveNewData, deriveAll, defaultValue)
import Happstack.Data.IxSet(noCalcs, Indexable(..), inferIxSet)
import Happstack.Data.IxSet.Merge (threeWayMerge')
import qualified Happstack.Data.IxSet.Revision as R
import Happstack.Data.User.Permissions
import Extra.Image
import Extra.ImageFile
import Happstack.Data.IxSet.Revision (NodeStatus(..))
import Happstack.Data.User.Permissions (Permissions(..), Ownable(..))
import Happstack.Server.Account (Username(..))
import Extra.File
import Extra.Image
import Extra.ImageFile
import Test.HUnit
import Debug.Trace

class PixmapShape a where
    pixmapHeight :: a -> Int
    pixmapWidth :: a -> Int
    pixmapMaxVal :: a -> Int

data ImageFieldName
          = ReportImageTopCrop
          | ReportImageLeftCrop
          | ReportImageBottomCrop
          | ReportImageRightCrop
          | ReportImageRotation
          | ReportImageFile
          | ReportImageURI
          | ReportImageCaption
          | ReportImageMeasure
          | ReportImageSize
          | ReportImageUnits
          deriving (Eq, Read, Show, Ord, Data, Typeable)

-- |This is all the information we store about an image in the report.
data ReportImage
          = Pic
            { picSize :: ImageSize
            , picCrop :: ImageCrop
            , picCaption :: Unicode
            , picImage :: Maybe ImageFile
            , picEdited :: Maybe ImageFile
            , picThumb :: Maybe ImageFile
            , picPrinter :: Maybe ImageFile
            , picMessages :: Unicode
            }
    deriving (Show, Read, Eq, Ord, Data, Typeable)

data Item
          = Item { itemName :: Unicode
                     -- It would be nice to use a Map here, but the template stuff makes it less straightforward.
                   , fields :: [(ItemFieldName, Unicode)]
                   , images :: [ReportImage]
                   } deriving (Show, Read, Eq, Ord, Data, Typeable)

      -- The Object fields should be split out of this and the rest moved back to Types.hs.
data ItemFieldName
          = ItemInspector
          | ItemClient
          | ItemProjectNumber
          | ItemInspectionDate
          | ItemLocation
          | ItemDataSheetNumber
          | ItemTypeOfObject
          | ItemArtistOrMaker
          | ItemArtistDate
          | ItemCountryOrNationality
          | ItemSubjectOrTitle
          | ItemItemDate
          | ItemSignatureInscriptionsMarkings
          | ItemRecto
          | ItemVerso
          | ItemMediumMaterialsTechniques
          | ItemDateOrPeriod
          | ItemSupportBaseFrame
          | ItemFrameMeasurement
          | ItemFrameHeight
          | ItemFrameWidth
          | ItemFrameDepth
          | ItemItemMeasurement
          | ItemItemHeight
          | ItemItemWidth
          | ItemItemDepth
          | ItemProvenance
          | ItemAddress
          | ItemCity
          | ItemTel
          | ItemFax
          | ItemCost
          | ItemCondition
          | ItemDescription
          | ItemMarketAnalysis
          | ItemExhibitionsAndPublications
          | ItemPhotographNumbers
          | ItemAdditionalNotes
          | ItemCashValue
          deriving (Eq, Read, Show, Ord, Data, Typeable)


data ReportFieldLabel
    = ReportName
    | ReportDate
    | ReportContractDate
    | ReportInspectionDate
    | ReportEffectiveDate
    | ReportAuthor
    | ReportPreparer
    | ReportPreparerEIN
    | ReportPreparerAddress
    | ReportPreparerEMail
    | ReportPreparerWebsite
    | ReportTitle
    | ReportIntendedUse
    | ReportValueType
    | ReportValueApproach
    | ReportClient
    | ReportClientGreeting
    | ReportItemsOwnerFull
    | ReportItemsOwner
    | ReportBriefItems
    | ReportInspectionLocation
    | ParagraphText
    | ReportElemType
    deriving (Read, Show, Eq)

data Author
    = Author { authorName :: Unicode , authorCredentials :: Unicode }
    deriving (Read, Show, Eq, Ord, Typeable, Data)

data AuthorFieldLabel
    = AuthorName
    | AuthorCredentials
    deriving (Read, Show, Eq, Ord, Typeable, Data)

data ReportValueType
    = MarketableCashValue
    | FairMarketValue
    | ReplacementValue
    deriving (Read, Show, Eq, Ord, Typeable, Data)
      
data ReportIntendedUse
    = SalesAdvisory
    | EstatePlanning
    | Insurance
    | CharitableDonation
    deriving (Read, Show, Eq, Ord, Typeable, Data)
      
data ReportValueApproach
    = MarketComparisonApproach
    | IncomeApproach
    | CostApproach
    deriving (Read, Show, Eq, Ord, Typeable, Data)

data ReportElem
    = ReportItem {elemItem :: Item}
    | ReportParagraph {elemText :: Unicode}
    | ReportUndecided
    deriving (Read, Show, Eq, Ord, Typeable, Data)

data Report
    = Report { reportFolder :: FilePath	-- Includes trailing /, omits localPrefix
             , reportName :: Unicode
             , reportDate :: Unicode
             , reportContractDate :: Unicode
             , reportInspectionDate :: Unicode
             , reportEffectiveDate :: Unicode
             , reportAuthors :: [Author]
             , reportPreparer :: Unicode
             , reportPreparerEIN :: Unicode
             , reportPreparerAddress :: Unicode
             , reportPreparerEMail :: Unicode
             , reportPreparerWebsite :: Unicode
             , reportAbbrevs :: [(Unicode, Unicode)]
             , reportTitle :: Unicode
             , reportIntendedUse :: Maybe ReportIntendedUse
             , reportValueType :: Maybe ReportValueType
             , reportValueApproach :: Maybe ReportValueApproach
             , reportClient :: Unicode
             , reportClientGreeting :: Unicode
             , reportItemsOwnerFull :: Unicode
             , reportItemsOwner :: Unicode
             , reportBriefItems :: Unicode
             , reportInspectionLocation :: Unicode
             , reportBody :: [ReportElem]
             , reportGlossary :: [(Unicode, Unicode)]
             , reportSources :: [(Unicode, Unicode)]
             , reportLetterOfTransmittal :: Unicode
             , reportScopeOfWork :: Unicode
             , reportCertification :: [Unicode]
             , reportLimitingConditions :: [Unicode]
             , reportPrivacyPolicy :: Unicode
             , reportMessages :: Unicode
             , reportPerms :: Maybe Permissions
             , revisionInfo :: R.RevisionInfo
             }
    deriving (Read, Show, Eq, Ord, Typeable, Data)

-- reportId :: Report -> R.Ident
-- reportId report = R.ident . R.revision . revisionInfo $ report

instance R.Revisable Report where
    getRevisionInfo = revisionInfo
    putRevisionInfo i x = x {revisionInfo = i}

instance Ownable Report where
    getPermissions = reportPerms
    putPermissions p report = report {reportPerms = p}

-- Identifies all revisions of a particular report
-- newtype ReportID = ReportID Integer deriving (Read, Show, Eq, Ord, Typeable, Data)
-- Uniquely identifies a particular revision of a particular report
-- data ReportRev = ReportRev Ident Integer deriving (Read, Show, Eq, Ord, Typeable, Data)

-- $(inferIxSet "ReportSet" ''Report 'noCalcs [''R.Ident, ''R.Revision, ''R.NodeStatus])
-- MUST STAY IN THIS MODULE!  ReportStore is used as a Component, and
-- therefore its full module name gets embedded in the serialized
-- representation.
-- data ReportStore = ReportStore R.Ident ReportSet deriving (Show, Read, Typeable)

{-
$(deriveNewData [''ReportStore, ''Report, ''Author,
                 ''AuthorFieldLabel, ''ReportValueType, ''ReportIntendedUse,
                 ''ReportValueApproach, ''ReportElem])


addMessage :: Unicode -> Report -> Report
addMessage message report =
    report {reportMessages = 
                case reportMessages report of
                  "" -> message
                  s -> s ++ convert (Wiki [NewLine, convert message])}

clearMessages :: Report -> Report
clearMessages report = report {reportMessages = ""}
-}

data ReportElemTypeName
    = UndecidedElem
    | ItemElem
    | CommentaryElem
    deriving (Read, Show, Eq)

o = Report {reportFolder = "", reportName = "Laurie Hanson", reportDate = "", reportContractDate = "", reportInspectionDate = "", reportEffectiveDate = "", reportAuthors = [Author {authorName = "Lydia Thompson, Ph.D.", authorCredentials = "====Education\n * Ph.D., New York University (NYU), Institute of Fine Arts (IFA), New York, NY, 1998.  Art and Archaeology.  Specialized China\n * M.A., NYU, Institute of Fine Arts (IFA), New York, NY, 1989.  Art and Archaeology\n * B.A., Middlebury College, Middlebury, VT, 1983.  East Asian Studies\n====Professional Study\n * University of California, Irvine, Certificate Program in Appraisal Studies, UCI in partnership with American Society of Appraisers (ASA).  Completion of all P.O.V. courses in certificate program\n * Shandong University, Jinan, Shandong, PRC.  Department of Archaeology, 1993-94.\n====Awards\n * Kress Foundation, Dissertation Fellowship, 1995-96.  Center for Advanced Study in the Visual Arts (CASVA), National Gallery of Art, Ittleson Fellow, 1993-95. Samuel F.B. Morse Fellowship, 1991-92\n====Memberships\n * Candidate Member, American Society of Appraisers\n====Selected List of Professional Experience 2000-2008\n * President, Thompson and Martinez Fine Art Appraisals, Inc., La Jolla, CA\n * Curator and Docent Educator for the exhibition: ``ymbols of Power:  Masterpieces from the Nanjing Museum'', the Bowers Museum of Cultural Art, Santa Ana, CA\n * Adjunct Faculty in Art History, UC Los Angeles and UC San Diego, CA.  \n====Selected Publications\n * ``Ritual, Art and Agency:  Consecrating the Burial Ground in the Han Period'' <em>Recarving China's Past:  Art, Archaeology, and Architecture of the Wu Family Shrines</em>.  Proceedings of a symposium held at the Princeton University Art Museum, (Princeton U.P.:  forthcoming)\n * Entries on artists and translation of essay by Art Critic Li Xianting, Betti-Sue Hertz ed. <em>Past in Reverse:  Contemporary Art in East Asia</em> (San Diego:  San Diego Museum of Art, 2004)\n * Co-author with Xu Huping, <em>Symbols of Power:  Masterpieces from the Nanjing Museum</em>, (Santa Ana:  The Bowers Museum of Cultural Art, 2002)\n"},Author {authorName = "Natasha Bonilla Martinez, MA", authorCredentials = "====Education\n * Certificate, University of California, Irvine Extension, Certificate Program in Appraisal Studies, UCI in partnership with American Society of Appraisers (ASA), 2007\n * Certificate, USPAP, University of California, Irvine Extension, Certificate Program in Appraisal Studies, UCI in partnership with American Society of Appraisers (ASA), 2006.\n * M.A. San Diego State University, Latin American Studies, San Diego, CA. 2001 Specialization in Latin American and Latino photography and art.\n * Graduate study. New York University, Department of Anthropology, New York, NY. 1982 to 1986, photographic anthropology and Native American art and culture.\n * B.A.  Vassar College, Poughkeepsie, NY, 1981. Independent Major.\n====Awards\n * Gateways Rockefeller Fellowship, 1997, Guadalupe Cultural Arts Center, San Antonio, TX, for the study of contemporary photography by Mexican and Mexican American women.\n====Selected Professional Experience 2000-2008\n * Vice-President, Thompson and Martinez Fine Art Appraisals, Inc., La Jolla, CA.\n * Director of Education, Director of Visual Arts and Education and Director of Program Development, successively, California Center for the Arts, Escondido, Escondido CA. \n * Co-Curator of Niki de Saint Phalle: California Dreaming, California Center for the Arts, Escondido, Escondido CA.\n * Co-Curator, Spirit Capture: Photography and Native America exhibition, National Museum of the American Indian, Smithsonian Institution, New York, NY.\n * Adjunct faculty, Department of Art, University of San Diego, San Diego, CA. Course on US/Mexico border art.\n====Selected Publications\n * ``An Indian Americas'' and ``Camera Shots'' in <em>Spirit Capture: Photographs from the National Museum of the American Indian</em>. (Washington, D.C.: Smithsonian Press, 1998).\n====Memberships\n * Candidate member, American Society of Appraisers\n * American Association of Museums \n * Association for Latin American Art, College Art Association\n"}], reportPreparer = "", reportPreparerEIN = "(preparer EIN)", reportPreparerAddress = "", reportPreparerEMail = "", reportPreparerWebsite = "", reportAbbrevs = [("[EffectiveDate]","September 17th, 2008"),("agree","agrees"),("appraisers","Appraiser"),("appraisers'","Appraiser's"),("are","is"),("assume","assumes"),("certify","certifies"),("Client","Mr. Ron Smith"),("EIN","20-8308057"),("have","has"),("InspectionAddress","17427 Francisco Drive\nSan Diego, CA 92128"),("intendeduse","Insurance purposes"),("LOTInspectedBy","Lydia Thompson, Ph.D."),("Our","My"),("our","my"),("Preparer","Thompson & Martinez Fine Art Appraisals, Inc.\nEIN: 20-8308057\n"),("preparer Email","appraiser@thompsonandmartinez.com"),("preparer Website","http://www.thompsonandmartinez.com"),("properties","properties"),("ReportType","summary"),("ReportUsers","Mr. Ron Smith and his insurance company"),("soucysmith","(from Soucy and Smith, eds, <em>The Appraisal of Personal Property - Principles, Theories, and \nPractice Methods for the Professional Appraiser\n</em>, 1994.)"),("take","takes"),("These","This"),("these","these"),("theseappraisers","this appraiser"),("Us","Us"),("us","us"),("USPAP","<em>2008-2009 Uniform Standards of Professional Practice (USPAP)</em>"),("USPAPShort","USPAP 2008-2009"),("ValueTypeLongDescription","Replacement Value is the price required to replace a property with another of similar age, quality, origin, appearance and condition, within a reasonable length of time in an appropriate and relevant market (from Soucy and Smith, eds, <em>The Appraisal of Personal Property - Principles, Theories, and \nPractice Methods for the Professional Appraiser\n</em>, 1994.) \n\nBecause of the time element involved, the appropriate market for replacement is deemed to be that of primary sources, such as art galleries carrying comparable and immediately available personal properties. If the only source for replacement can be demonstrated to be that of auction, that is the source utilized. Replacement Value is used as the basis for insurance purposes and damage/loss reports. "),("We","I"),("we","I"),("were","was")], reportTitle = "", reportIntendedUse = Just Insurance, reportValueType = Just ReplacementValue, reportValueApproach = Just MarketComparisonApproach, reportClient = "", reportClientGreeting = "Ms Hanson", reportItemsOwnerFull = "", reportItemsOwner = "", reportBriefItems = "", reportInspectionLocation = "", reportBody = [ReportParagraph {elemText = "Since the inception of the financial crisis and economic downturn in the fall of 2008, the art market has dropped precipitously. Well-known auction houses have seen their sales volume drop 70% or more since their highs in the spring of 2008 and, in several categories of art, prices have returned to 2002-2003 levels (5/21/09 Editorial by Swann President Nicholas Lowry). \n\nSome categories of art, however, have fared better in the downturn than others. Contemporary art, which realized spectacular price increases in the past 5 years, has been the hardest hit with average prices of high-end contemporary art dropping two-thirds in the evening sales. On the other hand, average prices of antiquities and Old Masters have softened but are still double what they were in 2003 (Art & Auction, June 2009, databank, p. 204). Sales of the impressionists at spring auctions showed surprising strength generally exceeding expectations (Art & Auction, July/August, 2009) Auction Reviews, p. 72.\n\nMs. Hanson's collection consists of prints by impressionists and old Masters as well as prints and paintings by well-known 20th century artists. Also included is a decorative Chinese painted screen.\n\nThere is a limited market for antique European prints in southern California, with only a few galleries on the West Coast, with one of the largest the Davidson Gallery in Seattle. The majority of galleries carrying this type of material is based in the East Coast of the United States and Europe. Although the customary practice is to seek comparable sales in the market where the insured shops, the unavailability of Old Master prints in the region required the Appraisers to seek comparable sales elsewhere in the national market. Given the quality of the works comparables were sought among dealers of high quality prints and members of the International Fine Art Print Dealer's Association, as well as in the auction market as appropriate.\n\nA significant element of the Hanson collection are several works by Hans Burkhardt (1904-1994), the Swiss American abstract impressionist, including a painting, works on paper and prints. A well known figure in Southern California, Burkhardt is especially known for his large scale anti-war paintings that expresses his response to the Spanish Civil War, World War II, Vietnam and Desert Storm. Recent major installations by the Jack Rutberg Gallery in Los Angeles and the California State University Northridge gallery have brought renewed attention to Burkhardt's work, which appears to be reflected in interest and activity at recent auction offerings at Sotheby's and Christie's.\n\n"},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,""),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,"Reverse composition features abstracted seated nude female figure on red field. Emphasis is on breasts and thighs with no detailing of face or lower extremities."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"purchased from the artist ca. 1985"),(ItemCondition,"Excellent"),(ItemMediumMaterialsTechniques,"woodcut on handmade paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"12/16"),(ItemSignatureInscriptionsMarkings,"Numbered lower left margin; signed and dated lower right margin; embossed stamp lower right margin"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Untitled"),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1972"),(ItemTypeOfObject,"Print"),(ItemLocation,"Dining room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"Hanson #26")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SoSr8iQyWJI/AAAAAAAAB0E/hscFnMAb58c/s512/IMG00013.jpg", fileChksum = "bec5bacaf92f4d696a29e2e9e7744f01", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 384, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SoSr8iQyWJI/AAAAAAAAB0E/hscFnMAb58c/s512/IMG00013.jpg", fileChksum = "bec5bacaf92f4d696a29e2e9e7744f01", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 384, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "8e585423d482abea53f36ac243f87b67", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 137, imageFileHeight = 183, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "33d652f5435a93658161916f5ef91d11", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1350, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_EV171eW_Cmg/SoSr8iQyWJI/AAAAAAAAB0E/hscFnMAb58c/s512/IMG00013.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"print acquired from the artist in 1985 according to Ms. Hanson"),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,"Three color composition features man wearing hat bending down to pick up a pepper on a sidewalk. The pepper is red and yellow, while the balance of the composition is black and white."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"purchased from the artist ca 1985"),(ItemCondition,"Very Good"),(ItemMediumMaterialsTechniques,"woodcut"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"8/12"),(ItemSignatureInscriptionsMarkings,"signed and numbered, lower right margin; embossed stamp lower right margin"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"\"Mark Tobey\""),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1975"),(ItemTypeOfObject,"Print"),(ItemLocation,"Dining room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"Hanson #27")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_EV171eW_Cmg/SoSrvxqq2JI/AAAAAAAAB0A/8y9y53D2YKo/s512/IMG00016_2.jpg", fileChksum = "79221c6448716b08b8ab19085b6bc582", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 392, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_EV171eW_Cmg/SoSrvxqq2JI/AAAAAAAAB0A/8y9y53D2YKo/s512/IMG00016_2.jpg", fileChksum = "79221c6448716b08b8ab19085b6bc582", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 392, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "90c3fc977a59abfda729d9fb50b0137f", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 138, imageFileHeight = 181, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "d21aad082a06bb4e5790de5ed9166967", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1378, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_EV171eW_Cmg/SoSrvxqq2JI/AAAAAAAAB0A/8y9y53D2YKo/s512/IMG00016_2.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Venator & Hanstein KG (Butcher Graphik Autographen Auuktion 100) 3/23/07 #548 estimated $1330 (EU 1000) bought in\n\nSame print currently available for $950 at the Davidson Gallery.\n\nAuction records of print published in Gazette de Beux-Arts indicate this print sell for between $431, $447, $632.\n\nDelteil 1/iv\n\nC&J Goodfriend offers Souvenir de Toscane 2nd state print, dated 1845 for $2500\n\nMasterworks Fine Art Inc. has recently sold a 4th state Souvenir de Toscane but don't know price.\nRecently sold (4th state, A. Salmon printer, Gazette des Beaux-Arts at Masterworks Fine Art Inc. \n"),(ItemCashValue,"$950"),(ItemMarketAnalysis,"Jean-Baptiste-Camille Corot was the leading painter of the Barbizon school of France in the mid-nineteenth century. He is a pivotal figure in landscape painting and his vast output simultaneously references the Neo-Classical tradition and anticipates the plein-air innovations of Impressionism. He was also an engraver. ``Souvenir de Toscane'' was originally executed in 1845 and produced in four states.\n\nValue characteristics include: state, edition, subject matter and condition. State is a significant characteristic of value for this print with impressions published before the Gazette des Beaux-Arts edition (state 4) being especially scarce. A second state version of Souvenir de Toscane is available at C&J. Goodfriend Galleries for $2500. The subject property is an example of the 4th state published in Gazette des Beaux-Arts in 1875 whA second state version of Souvenir de Toscane is available at C&J. Goodfriend Galleries for $2500. between $250 and $630.  A 4th state print in similar condition and impression to the subject property is currently offered for sale at the Davidson Galleries for $950 and is the basis for estimating its replacement value."),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Gallery"),(ItemCondition,"Very Good: dark spot (mold?) lower margin center. Paper appears toned, beige, almost sepia"),(ItemMediumMaterialsTechniques,"Etching on laid paper "),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Edition of 750, 4th state, Delteil 1, iv., Melot 1,v."),(ItemSignatureInscriptionsMarkings,"Bottom left: Corot Del. & Next line Gazette des Beaux-Arts bottom right: Imp. A. Salmon, Paris."),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Plate impression: 5\" x 7\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"11 3/4\" x 14 1/2\""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Souvenir de Toscane"),(ItemCountryOrNationality,"France"),(ItemArtistDate,"1796-1875"),(ItemArtistOrMaker,"Jean Baptiste Camille Corot"),(ItemItemDate,""),(ItemDateOrPeriod,"c. 1843, published April 1st, 1875. Printed by A. Salmon, Paris"),(ItemTypeOfObject,"Print"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"1 Hanson #41")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/Shlbi8G5awI/AAAAAAAABoU/NxQ-H2USxAo/s800/DSC_0281.JPG", fileChksum = "b743dc61d9721ee357c2b0d8555d5dac", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/Shlbi8G5awI/AAAAAAAABoU/NxQ-H2USxAo/s800/DSC_0281.JPG", fileChksum = "b743dc61d9721ee357c2b0d8555d5dac", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "30cbc4b81ec19aa1ba6b1441a264df7b", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 194, imageFileHeight = 129, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "da929ce67ba02b3f079f047d0a712aba", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2707, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/Shlbi8G5awI/AAAAAAAABoU/NxQ-H2USxAo/s800/DSC_0281.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Gallery label on back. Original print 1904. Published in 1919.\n\nAccording to the Catalog Raisonne, the original lithograph is from 1904.\n\nThis print has sold at auction for between $1000 and $6000 mostly from this edition.\n\nA single print from the edition of 950 (``on wove paper'') sold at Christie's New York 5/2/07 for $6000.\n\nAnother print ``Ambroise Vollard\" of the print-maker Vollard from the same series: ``Douze Lithographies Originales\" of 1919 is available from Davidson for $3800.\n\nMasterworks Fine Art: Claude Renoir, La Tete Baissee. Original print on Japon paper, edition of 50. Offered for $8700.\n\nArmstrong Fine Art\n``I currently have an impression of \"Claude Renoir, la Tete Baissee\" in pristine condition for $5,500.  It is a beautiful impression on a full sheet of Arches ``wove paper'' (implies vellum??, see above).  There are no tears, creases, discolorations, etc.  There are only two acid-free cloth hinges on the verso at the top of the sheet, which could easily be removed.  Aside from that, it is perfect.''\n\nIf you have further questions, please do not hesitate to contact me.\n \nSincerely, Bernard\n \nBernard Derroitte\nArmstrong Fine Art\n200 East Walton Place\nChicago, IL  60611\nT.  312.664.9312; F.  312.664.9862\nwww.armstrongfineart.com\n\nLouis Valtat. c. 1904. Lithograph. Delteil, Stella 38. 11 5/8 x 9 1/4 (sheet 13 x 9 3/4). Series: Douze Lithographies Originales de Pierre-Auguste Renoir. A fine impression apart from the edition of 1,000 printed on cream laid paper with deckle edges on three sides. Signed in the stone. $4,500.\n"),(ItemCashValue,"$4500"),(ItemMarketAnalysis,"Renowned French Impressionist painter Pierre-Auguste Renoir took up printmaking later in life after 1890, and is thought to have executed some 59 lithographs and etchings, some of which are regarded as some of the best of the period. Renoir was the first of the Impressionists to enjoy popular success and his prints were produced in recognition of that success. Stella lists 60 prints (25 etchings and 35 lithographs) in his catalogue of Renoir's prints. Michel Melot, The Impressionist Print notes that most of Renoir's prints were produced after he had became famous and represent an attempt to satisfy the demand for his works, almost 3/4 of them having been produced after 1900 for inclusion in books and portfolios produced by Theodore Duret, the chief chronicler of the Impressionists, and Ambroise Vollard, Renoir's dealer and a great lover of the art of printmaking. One of Renoir's subjects of choice was portraits of children and friends.\n\nValue characteristics for Renoir lithographs include quality of impression, state and condition and type of paper. Executed in 1904, this is an impression from 2nd and final state which was published for ``L'Album des Douze Lithographies Originales de Pierre August Renoir\" by A. Vollard (Paris, 1919) in and edition of 1000, with 50 printed on Japon paper and 950 printed on Arches wove paper. Some images from this series were also printed separately on laid paper. At Allinson Gallery, a print from this series on laid paper is available for $4500.\n\n\nExamples of the print are available at retail with the print from the edition of 950 available for  $5500, and the two from the edition of 50 for $6500 and $8700."),(ItemDescription,""),(ItemCost,"$785"),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Galleries"),(ItemCondition,"Very good. Discoloration to upper horizontal and right vertical margins. No tape residue on reverse. UV light damage to the sheet. Deckled edges."),(ItemMediumMaterialsTechniques,"On laid paper. No watermark."),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Second state. Edition: Ambroise Vollard, Paris, 1919. Printer: Auguste Clot. Edition of 1000. One from the series, \"Douze Lithographies Originales\". D. 39"),(ItemSignatureInscriptionsMarkings,"Signature L/R, signed in stone"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Mat opening: 9 3/4'' x 8''\nSheet: 13\" x 10\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"21 5/8'' x 17 5/8''"),(ItemSupportBaseFrame,"Glazed gilt wood frame"),(ItemSubjectOrTitle,"Claude Renoir, La Tete Baissee"),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1841-1919"),(ItemArtistOrMaker,"Auguste Renoir"),(ItemItemDate,""),(ItemDateOrPeriod,"c. 1904"),(ItemTypeOfObject,"Print (Davidson)"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"2 Hanson #37")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SmTiIV_kBoI/AAAAAAAAC-4/GfT1RJYXZXc/s800/Renoir.jpg", fileChksum = "340ae2f791e548f90b9a345b32cf894e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 600, imageFileHeight = 678, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SmTiIV_kBoI/AAAAAAAAC-4/GfT1RJYXZXc/s800/Renoir.jpg", fileChksum = "340ae2f791e548f90b9a345b32cf894e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 600, imageFileHeight = 678, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "d8a06582fa6ab7548c1a1d3f3a1c8a03", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 149, imageFileHeight = 168, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "7b3d6e27ca4467778a146c3df16d84b2", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1593, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/SmTiIV_kBoI/AAAAAAAAC-4/GfT1RJYXZXc/s800/Renoir.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"\"A Basement Shop in Cracow\" $2,000. Lockport Street Gallery\n\nAu Pied du Sinai: Schlome Fuss A La Synague, ?, Lewis and Maese Auctions, Houston, TX, auction est $3,750 - $4,250 lithograph in black-brown on china paper, signed in pencil Lower right. Rare - 25 proofs, deluxe book edition iii.\n"),(ItemCashValue,"$2,000"),(ItemMarketAnalysis,"Henri Toulouse-Lautrec was a French painter, printmaker, and illustrator best known for his depictions of life in the bohemian world of the Parisian theater and its environs at the end of the 19th century. The subject property is from a series of works depicting Jewish life that illustrated a volume of sketches or travel essays on Eastern European Jewish life by Georges Clemenceau, who later served as prime minister of France. Toulouse-Lautrec is said to have observed Jewish life in Paris as the basis for his illustrations.\n\nThere is one state only of this lithograph, one of a set of eight published by Henry Floury. The group of eight lithographs were published in a deluxe book edition with four variations, and in a regular book edition in two variations.\n\nValue characteristics include edition and condition. The subject property is from the regular book edition, var. ii, as indicated by its sheet size, and is a good quality impression in good condition. This print occasionally appears at auction and at retail, as part of the set of eight and also as a single work. A national survey of print dealers located two offerings ranging from $2000 to $4500.  The higher comparable is for a print from the Deluxe book edition, and the lower comparable for a print from the Regular Book edition similar to the subject property. The lower comparable is the basis for the value conclusion of Replacement Value Comparable of $2,000."),(ItemDescription,"The composition features an elderly man with a long beard and wearing a coat and hat, looking out of a basement level window adjacent to a sidewalk. His hands are resting on the window ledge at the same level as the sidewalk. A small black cat sits at the edge of the window near the man's right hand. The window is lined with large blocks, evoking a an older stone building. The composition emphasizes shades of gray, in the figure of the man window and building, so that the black cat serves as a focal point."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Good impression in very good condition. Minor UV light damage. No adhesive or tape on reverse."),(ItemMediumMaterialsTechniques,"Lithograph on wove paper, no watermark"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Series: \"Au Pied du Sinai\", 1 state only"),(ItemSignatureInscriptionsMarkings,"Unsigned"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Image impression: 170 mm x 140 mm; Sheet size: 260 mm x 196mm"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"16\" x 13 5/8\""),(ItemSupportBaseFrame,"Gilt wood frame, glazed"),(ItemSubjectOrTitle,"\"Une Arriere-boutique a Cracovie\" (A Basement Shop) from the series \"Au pied du Sinai\"(From the Foot of Sinai)"),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1864-1901"),(ItemArtistOrMaker,"Henri Toulouse-Lautrec"),(ItemItemDate,""),(ItemDateOrPeriod,"1898"),(ItemTypeOfObject,"Print"),(ItemLocation,"Living room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"4 Hanson #36")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "\"Une Arriere-boutique a Cracovie\"", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SmThoUC_IxI/AAAAAAAAC-o/T4jaLSFLz0c/s400/Toulouse%20Lautrec1.jpg", fileChksum = "ed9232e7fa5b87ba495cacce48c03ee9", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 321, imageFileHeight = 400, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SmThoUC_IxI/AAAAAAAAC-o/T4jaLSFLz0c/s400/Toulouse%20Lautrec1.jpg", fileChksum = "ed9232e7fa5b87ba495cacce48c03ee9", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 321, imageFileHeight = 400, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "9e82c2f080523b6b49be954d3aea9a4e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 142, imageFileHeight = 177, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "20b6fc9e14bd2f4041c1c071c35a6b10", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1445, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_s-XE562xJtM/SmThoUC_IxI/AAAAAAAAC-o/T4jaLSFLz0c/s400/Toulouse%20Lautrec1.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Etching, \8220Tampoco\8221 from \8220Desastres de la Guerra\8221,  is $2,250., David Tunick, Inc.\n\nEtching \8220Bien te esta\8221 from from \8220Desastres de la Guerra\8221, proof impression is $140,000, Harris Schrank\n\n"),(ItemCashValue,""),(ItemMarketAnalysis,"Francisco de Goya, the Spanish master painter, illustrator and engraver of the 19th century is well known for his series of prints describing the human cost of the Napoleonic Wars in Spain called \8220Los desastres de la Guerra,\8221 (The Disasters of War). The series includes some 80 prints of which only some proofs and individual prints were made during the artist\8217s lifetime. The first published edition was issued posthumously in 1864, and ultimately seven editions were printed, in varying degrees of quality.\n\nValue characteristics for Goya prints from this series include edition, condition, quality of impression, and plate or subject. At auction the value range is dramatic, with a rare proof of \"Bien te esta\" reaching more than $140,000 to later edition lesser quality impressions selling for less than $500. The subject property, an impression of plate 45 titled \8220Tambien estos\8221, is from the first edition. It  is a good quality impression in very good condition. A national survey of galeries did not locate any prints of this plate, but did locate several first edition prints of other plates in the series. These range from ? to ?. The subject property, is a very good impression in generally good condition.\n"),(ItemDescription,"The composition depicts a room full of wounded soldiers, some lying on the floor and on beds, others seated and standing. Three figures form a triangle in the composition: the seated figure on a bed in the foreground, the standing figure looking over his shoulder in the center and the figure lying on the right in the middle."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very good impression in good condition. Upper and lower margins have been trimmed. Mat burn but no foxing and UV light exposure. Remnants of tape residue on reverse of the sheet. Inscribed along bottom margin. \"1-13-37\". Inscription on reverse of sheet L/R corner: ``Disasters of war, 1st edition''. "),(ItemMediumMaterialsTechniques,"Etching on wove paper with no watermark."),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Series: Los Desastres de la Guerra, first edition"),(ItemSignatureInscriptionsMarkings,"Unsigned, Composition L/L; no. 25 in U/R margin"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Image size: 4 1/2'' x 7 1/2'' (11 1/2 cm. x 19.2 cm); \nPlate impression: 6 1/8'' x 9 1/2'' (15 1/2 cm x 23 1/2 cm)\nSheet size: ???"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"19\" x 20 3/4''"),(ItemSupportBaseFrame,"Glaze gilt wood frame"),(ItemSubjectOrTitle,"\"Tambien Estos\" (Also these) from the \"Los desastres de la guerra\" (Disasters of War) series"),(ItemCountryOrNationality,"Spanish"),(ItemArtistDate,"1746-1828"),(ItemArtistOrMaker,"Francisco de Goya "),(ItemItemDate,""),(ItemDateOrPeriod,"1898"),(ItemTypeOfObject,"Print"),(ItemLocation,"Living room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"5 Hanson #32")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/SmThoXPWNmI/AAAAAAAAC-I/4U5lZrDuWkA/s640/Goya1.jpg", fileChksum = "10c5286b55fa1707077e866f3c8bbad5", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 640, imageFileHeight = 479, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/SmThoXPWNmI/AAAAAAAAC-I/4U5lZrDuWkA/s640/Goya1.jpg", fileChksum = "10c5286b55fa1707077e866f3c8bbad5", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 640, imageFileHeight = 479, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "9ad9cea61d166acdf7a95b491f2faa36", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 183, imageFileHeight = 137, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "f3b70a8be1092118de3857829f4f6465", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2405, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh4.ggpht.com/_s-XE562xJtM/SmThoXPWNmI/AAAAAAAAC-I/4U5lZrDuWkA/s640/Goya1.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Gallery label on back\n\nJayne Haslem Gallery, The Zither, 9 7/8'' x 10 3/4\" Aquatint and etching, $1500\n\nThe Annex Gallery, The Zither, 24/60 published 1982. $750.\n\nKiechel Fine Art, August, 1940 etching 9 3/4\" x 8\" $1200\n\nEtchings and acquatints of the same era have sold at auction for between $418 and $3200 with the majority in the $1100- $2400 range."),(ItemCashValue,"$1500"),(ItemMarketAnalysis,"Born in Beverly, Massachusettes in 1911 American contemporary artist, Will Barnet is an acclaimed painter and printmaker, who has also taught art at the leading art schools in the U.S.\n\nTrained at the Boston Museum of Fine Arts School and the Art Students League New York, he later taught art at such leading American schools as Yale University, Cornell University, and the Art Students League.\n\nA prolific graphic artist, Barnet changed his style significantly at different points in his career. His earliest works were influenced by expressionism; they were followed by abstract works in the 1950s and 1960s, and finally evolved into more figurative works of silhouetted forms set against geometrically designed backgrounds.\n\nHis work has been exhibited in prominent museums and galleries in the United States and Canada and is included in many prestigious collections, including the Metropolitan Museum of Art and the Guggenheim Museum in New York and the Boston Museum of Fine Arts. The subject property is an example of Barnet's expressionistic early work. A search of auctions and galleries did not turn up any available examples of ``The Bath''. Other aquatints and etchings of this period have sold at auction for between $418 - $3200. At retail, prints of this era have sold for between $750 and $1500 depending on edition. Because ``The Bath'' is currently scarce, the subject property is valued at the high end of this range."),(ItemDescription,""),(ItemCost,"$107"),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Galleries. Mercer Island, WA"),(ItemCondition,"Very good"),(ItemMediumMaterialsTechniques,"Acquatint etching"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Edition of 12"),(ItemSignatureInscriptionsMarkings,"L/R signature, title II ``The Bath''; Edition 12"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 9 3/4\" x   12\"; Impression: 6 1/2\" x 8 3/4\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"The Bath"),(ItemCountryOrNationality,"American"),(ItemArtistDate,"b. 1911"),(ItemArtistOrMaker,"Will Barnet"),(ItemItemDate,""),(ItemDateOrPeriod,"1937, Cole 46"),(ItemTypeOfObject,"Print, Davidson"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"6")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShlbvJW0ffI/AAAAAAAABqc/kv21ucxKpug/s800/DSC_0343.JPG", fileChksum = "2741073cabe851ab5f85e641b3127b12", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 632, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShlbvJW0ffI/AAAAAAAABqc/kv21ucxKpug/s800/DSC_0343.JPG", fileChksum = "2741073cabe851ab5f85e641b3127b12", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 632, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "571790e957786b7b125ee1197cb262be", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 178, imageFileHeight = 141, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "39b9eb87169cf6ce97850c8c4b23e3ae", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2278, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/ShlbvJW0ffI/AAAAAAAABqc/kv21ucxKpug/s800/DSC_0343.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Male Figure, lithograph, 1902, edition of 50 offered for $950 at Davidsons.\nDeux Nus recently sold at Davidson's for $950\nDeux Nus offered for sale at Figaro Gallery for $1800.\nDavidson's recently sold a print for $950"),(ItemCashValue,"$1800"),(ItemMarketAnalysis,"A French painter, illustrator and printmaker, member of the group of artists called the Nabis and afterward a leader of the Intimists; he is generally regarded as one of the greatest colourists of modern art. He also produced lithographs and dry-point. In the 1920s he undertook a series of paintings on one of his most famous themes\8212a nude in a bath, the subject of this lithograph. Les Baigneuses  was probably engraved for Ambrose Vollard who was planing an Album de Nus with various artists. Value characteristics for Bonnard prints include quality of impression, state and condition. A search of auction prices indicate that lithographs of nudes by Bonnard have sold for between $1200 and $2880. At the retail level lithographs of nudes, including 2 examples of Les Baigneuses, have sold and are available at the retail level for between $950 and $1800. Les Baigneuses is currently available the The Figaro Gallery and is the basis for estimating the subject property's Replacement Value. \n\n"),(ItemDescription,"Two nude women are situated in a landscape. One lies on the grass while the other stands in profile next to a small tree or bush. Rendered in scratch"),(ItemCost,"$232"),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Galleries, Mercer Island, WA"),(ItemCondition,"Very Good. Nothing to note."),(ItemMediumMaterialsTechniques,"Drypoint and roulette. Handmade wove paper (Japanese paper)"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Edition of 50"),(ItemSignatureInscriptionsMarkings,""),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 26 cm x 31 cm 10\" x 12\"; Impression:  16 cm x 23.2 cm  6 1/2\" x 23.2 cm. 9 1/4\" \t\t\t\n"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Deux Nus (Les Baigneuses)"),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1867-1947"),(ItemArtistOrMaker,"Pierre Bonnard"),(ItemItemDate,""),(ItemDateOrPeriod,"c. 1927-29"),(ItemTypeOfObject,"Print (Davidson)"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"7, Hanson #18")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/ShlbvtyXvFI/AAAAAAAABqk/cUWgS7DuEH8/s800/DSC_0345.JPG", fileChksum = "86cc8540086f6ff5101b3f370297ae92", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 624, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/ShlbvtyXvFI/AAAAAAAABqk/cUWgS7DuEH8/s800/DSC_0345.JPG", fileChksum = "86cc8540086f6ff5101b3f370297ae92", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 624, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "33c6d8738ebaf0cf5ede0f8b1903896d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 179, imageFileHeight = 140, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "66cd9fdb2fc7a56c64c2fc4f760b85e3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2308, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh3.ggpht.com/_s-XE562xJtM/ShlbvtyXvFI/AAAAAAAABqk/cUWgS7DuEH8/s800/DSC_0345.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,""),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Good: Ares of loss upper right, upper left quadrant and center"),(ItemMediumMaterialsTechniques,"Oil on canvas"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,""),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"9 1/2\" x 13 1/12\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"15 1/8\" x 19 1/4\""),(ItemSupportBaseFrame,"Glazed gilt wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,""),(ItemArtistDate,""),(ItemArtistOrMaker,"Hudson River School"),(ItemItemDate,""),(ItemDateOrPeriod,""),(ItemTypeOfObject,"Painting"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"14, Hanson #43")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SiBTr4P6xmI/AAAAAAAAB4Q/hY6AhRf1thQ/s720/DSC_0374.JPG", fileChksum = "6bf4434d44c3c6ee243eea34bfdf73e3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 479, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SiBTr4P6xmI/AAAAAAAAB4Q/hY6AhRf1thQ/s720/DSC_0374.JPG", fileChksum = "6bf4434d44c3c6ee243eea34bfdf73e3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 479, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "4547a89fcc964f7278edf8f86427dcd3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 194, imageFileHeight = 129, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "1b0078d0ead16f4279a50a75ce3f7f41", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2706, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/SiBTr4P6xmI/AAAAAAAAB4Q/hY6AhRf1thQ/s720/DSC_0374.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Oil on panel works available at auction in the Netherlands and Germany as well as Sotheby's and Christie's, Amsterdam, Bonham's and Butterfields, Leslie Hindman in the USfor between $250 and $6000 (Bonham's, 5/16/07 oil on board 15 3/4\" x 12\" A Fisherman on a dock) with the majority selling in the $1000-$2500 range. \n\nlots 606, 607 oil on panel, portraits'65 Leslie Hindman:  1/18/09, $1464\nGlerum, Amsterdam, boy with goat, 1970 $930\nGlerum, Amsterdam, 6/2/08 $1085\n\n\n\nA painting by the artist is available at Kamp Galleries in Winnetka, IL. \nDear Lydia,\nThanks for your inquiry about our Van d. Berg Still Life.  This is the only painting we have by the artist\nand we believe it to be an early example, between the wars, ca 1925.  The dealer price is $2,900.00\n(Retail $5,900.00).  It's a beautiful example of his work, and a good representation of art from that era.\nYou should buy it.\nRegards,\nNick Vahlkamp, mgr.\n- Show quoted text -\n-- \nKamp Gallery\nDrake Hotel Arcade\n140 East Walton Street\nChicago, IL 60611\n312-664-0090\nfax 312-664-0014\n\n\nHis name is also listed on Covington Fine Arts Gallery in Tuscon, AZ.\n6530 E Tanque Verde Rd Suite 140\nTucson Arizona, 85715 USA\nTel: \t(520) 298-7878\nAlso see M. Klein Auctions in Chicago. \n400 W. Hubbard \nChicago Illinois, 60622 USA\nTel: \t312-948-0080\nFax: \t312-948-0090\nListed on Artnet\n"),(ItemCashValue,"$2,900"),(ItemMarketAnalysis,"Willem van den Berg studied art at the Academie voor Beeldende Kunst in The Hague and traveled extensively to Belgium, France, Italy, Austria and England. \nKnown as  a draughtsman, painter and graphic artist, Van Den Berg's style is characterized by a preference for strong closed contours which reflects his interest in Japanese wood block prints. He was director of the Rijksacademie in Amsterdam from 1938 to 1953. He died in Amsterdam in 1970. His paintings are in the collections of museums in Holland, Europe and the United States. The majority of Van Den Berg's works are available through auctions in Amsterdam and Germany. They have sold for between $732 (7.5\" x 5.25\"), Portrait on panel and $6000. Van Den Berg is particularly known for his oil on panel portraits of fishermen and farmers, and these images have brought the highest prices with majority n the $1000-$2500 range. Paintings by the artist comparable to the subject property's pastoral subject that have come up at auction in recent years have all failed to sell indicating a weak market for this subject. A search for Van Den Berg's works for sale in galleries the United States and Europe turned up a single example - an oil on canvas painting still life of garden flowers which is available for $2900 and is the basis for the value conclusion for the subject property."),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very good: surface crackelure"),(ItemMediumMaterialsTechniques,"oil on panel"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"W. V. D. Berg"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"11 1/4\" x 15 1/4\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"16 3/4\" x 20 3/4\""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,"Holland/Netherlands"),(ItemArtistDate,"1886-1970"),(ItemArtistOrMaker,"Willem van den Berg"),(ItemItemDate,""),(ItemDateOrPeriod,"early 20th century"),(ItemTypeOfObject,"Painting"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"15")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwmvncL_sI/AAAAAAAABww/33e5fwELMLY/s720/DSC03318.JPG", fileChksum = "49dd80467b4515462303f794b29ffe58", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 504, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwmvncL_sI/AAAAAAAABww/33e5fwELMLY/s720/DSC03318.JPG", fileChksum = "49dd80467b4515462303f794b29ffe58", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 504, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "90b383820e81928940f738d65cf7a70f", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 189, imageFileHeight = 132, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "91cfb0af5d0e0f17e6f33b88eb4842d6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2571, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/ShwmvncL_sI/AAAAAAAABww/33e5fwELMLY/s720/DSC03318.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Label on back has biography of artist\nSothebys 2006 1860 \"Garden views with green houses, figures, and gardeners\", 5.3 in x 8.3 in, watercolor over black chalk, $1,566\nthere are additional drawing comps on Artprice, also for Jean-Victor Bertin"),(ItemCashValue,"$1,566?"),(ItemMarketAnalysis,"Charles Nicolas Ransonnette was a painter, engraver and draughtsman active in the mid-19th century. He served as official draughtsman and engraver to the Duchess of Berry, and exhibited regularly at the French Salon from 1822 to 1857. He was a pupil of neo-classical painter Jean-Victor Bertin, and his father, Pierre Nicolas Ransonette, also a painter, engraver, and draughtsman, served the French royal family. \n\nRansonnette worked in a finely detailed and charming style, with many works in small scale. Works by Ransonnette rarely appear in the American market. No current retail offerings could be located and only a few auction sales. The most recent, is for a large scale watercolor over chalk on paper landscape, 5.3 in. x 8.3 in., which sold for $1,566. The subject property is comprised of two ink on paper works framed together. Each work is approximately one half of the comparable. "),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very good"),(ItemMediumMaterialsTechniques,"pen and ink and wash on laid paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"said to be signed on verso under dust jacket"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Mat opening: Upper and Lower: 4\" x 5"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"20 1/2\" x 15 1/2\" "),(ItemSupportBaseFrame,"Gilt and black paint glazed wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1793-1877"),(ItemArtistOrMaker,"Charles-Nicolas Ransonnette"),(ItemItemDate,""),(ItemDateOrPeriod,"1856"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"16, Hanson #38")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5A3amjI/AAAAAAAABx0/IKqO3Zh6LHU/s512/DSC03332.JPG", fileChksum = "e1aca3d54e41127c581be92762f4b8bd", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5A3amjI/AAAAAAAABx0/IKqO3Zh6LHU/s512/DSC03332.JPG", fileChksum = "e1aca3d54e41127c581be92762f4b8bd", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "9e7179bac603aebdd740d2f5d8bd9074", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 129, imageFileHeight = 194, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "cd0cd9d06928104877468460ea075fc6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1199, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh3.ggpht.com/_s-XE562xJtM/Shwm5A3amjI/AAAAAAAABx0/IKqO3Zh6LHU/s512/DSC03332.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Jean-Luc Baroni, Ltd.of London has a drawing for sale. As yet I have not located galleries that sell his prints..\n\nMost recent sale of print, Elefanten, ein Lowe unde ein Eber, 33.2 x 47.7 cm. Bassenge, 04/06/09 $1050\n\nEtchings by are available at Spaightwood Galleries for between $1750- $4500. Most are much smaller than the subject property. "),(ItemCashValue,"$4,500"),(ItemMarketAnalysis,"Known as a Mannerist painter, draughtsman and printmaker, Battista Franco spent the first years of his independent career in Rome, but by 1536 had settled in Florence. Franco is better known, and has usually been more highly regarded, as a draughtsman rather than as a painter. He is also well known for his etching having produced over 100 during his life-time. An assiduous copyist, among his drawings and etchings are numerous studies of antique subjects, usually taken from reliefs on Roman sarcophagi. According to the Illustrated Bartsch, The Deluge was likely after a work designed by Polidoro da Caravaggio.\n\nValue characteristics of prints include state, quality of impression, edition size, subject matter and condition. The subject property is in good condition and is a fine impression. A search for etchings by this artist indicate that the primary  market for this artist is in Europe with most sales taking place at Galerie Bassenge Kunst Und Buchauktionene and Reiss & Sohn in Berlin.  Prints of similar biblical subjects, size, quality of impression, condition and subject matter range from $1401-$2424. The subject property is most comparable in quality of impression to prints at the high end of this range. A search of etchings by the artist in galleries in the United States only turned up a single venue, Spaightwood Galleries in Upton, Massachusetts where a number etchings are available for between $1750-$4500. Most of the etchings available for sale are considerably smaller than the subject property. The etching most closely approximating the subject property in size and quality of impression is available for $4,500 and is the basis for estimating the subject property's replacement value\n"),(ItemDescription,"A biblical scene of the flood."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Good: Some stains and creases. Good impression."),(ItemMediumMaterialsTechniques,"Etching"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signed in the plate lower left, Battista Franco Inu"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 10 1/2\" x 17\" (27 cm x 43 cm)"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"25 1/4\" x 30 1/4\""),(ItemSupportBaseFrame,"Glazed silver gilt wood frame"),(ItemSubjectOrTitle,"The Deluge, Le deluge universel, 3, Bartsch"),(ItemCountryOrNationality,"Italy"),(ItemArtistDate,"1498-1561"),(ItemArtistOrMaker,"Giovanni Battista Franco Inu"),(ItemItemDate,""),(ItemDateOrPeriod,"16th century"),(ItemTypeOfObject,"Print"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"17, Hanson #66")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG", fileChksum = "dac291fe9972381edd30621a4739a12c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG", fileChksum = "dac291fe9972381edd30621a4739a12c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "a699a3975a3ee1fb393e47e1ab5b3ca1", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 194, imageFileHeight = 129, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "592c3b006543a18d7373ab57901c5832", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2707, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"pastel \" Standing Female Nude \" is available, framed and in excellent conditions and as described on our site. The asking price is $ 9.500, but because we are having a special summer promotion, I can give you a discount of 25%, for a net of $ 7.125, including packing, shipping and insurance Pasquale Ianetti Gallery, San Francisco, CA\n\npastel Nude, was offered for $14,000 at Jack Rutberg Gallery, Los Angeles, CA\n"),(ItemCashValue,"$10,000?"),(ItemMarketAnalysis,"Hans Burhardt worked in a variety of media during his long career, including painting, drawing and print making. Known primarily as an abstract impressionist, Burkhardt also worked in a figurative style. Born in Switzerland, Burkhardt came to the United States in ??. He moved to Los Angeles in 1937 and was a well known and well regarded figure in the West Coast art scene. Burkhardt\8217s work has received recent attention due to two major projects, a retrospective of his work mounted at the University Gallery at California State University Northridge, and a comprehensive exhibition of his 1960s paintings presented at the Jack Rutberg Gallery in Los Angeles in 2008-2009. These projects have brought renewed public attention to the artists\8217s work reflected in a greater wolume of sales at auction.\n\nValue characteristics for Burhardt prints include date, size, condition and quality. In the case of his works on paper, a review of auction results in the past five years indicates that earlier works from the 1930s and 1940s appear to be more desirable, reaching higher prices than later drawings from the 1970s and 1980s. The subject property is an earlier work in the abstract style in very good condition. A survey of retail offerings located two pastel works on paper, one offered for $7,125 and the second for $14,000\n"),(ItemDescription,"Abstract composition of three figures rendered in colors on ivory paper"),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good"),(ItemMediumMaterialsTechniques,"graphite and pastel on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signed, dated lower center quadrant"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"26\" x 19 1/2\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"27\" x 20 3/4\""),(ItemSupportBaseFrame,"Red painted and gilt wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,""),(ItemArtistDate,""),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1969"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"18 Hanson #64")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5yzTHVI/AAAAAAAABx4/WErzxTrjaa8/s512/DSC03333.JPG", fileChksum = "a65da9e6609f6590ef0449079fbef5f1", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 400, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5yzTHVI/AAAAAAAABx4/WErzxTrjaa8/s512/DSC03333.JPG", fileChksum = "a65da9e6609f6590ef0449079fbef5f1", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 400, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "f3a8178401afff79c8e85a4dd92155de", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 140, imageFileHeight = 179, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "43029cacdeccdc39f5725592e3fca02f", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1406, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh3.ggpht.com/_s-XE562xJtM/Shwm5yzTHVI/AAAAAAAABx4/WErzxTrjaa8/s512/DSC03333.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes," \8220Flowers and Tears\8221 excellent condition offered for $8,500, The Caldwell Gallery\n\8220Pink Sage\8221 1960, 18 x 25, excellent condition, offered for $34,000, Jack Rutberg Gallery, Los Angeles\nLandscape, 1968, 20 x 24, excellent condition, offered for $?, Jack Rutberg Gallery, Los Angeles\n"),(ItemCashValue,""),(ItemMarketAnalysis,"Hans Burhardt worked in a variety of media during his long career, including painting, drawing and print making. Known primarily as an abstract impressionist, Burkhardt also worked in a figurative style. Born in Switzerland, Burkhardt came to the United States in ??. He moved to Los Angeles in 1937 and was a well known and well regarded figure in the West Coast art scene. Burkhardt\8217s work has received recent attention due to two major projects, a retrospective of his work mounted at the University Gallery at California State University Northridge, and a comprehensive exhibition of his 1960s paintings presented at the Jack Rutberg Gallery in Los Angeles in 2008-2009. These projects have brought renewed public attention to the artists\8217s work reflected in a greater wolume of sales at auction.\n\nValue characteristics for Burhardt prints include date, size, condition and quality. In the case of his works on paper, a review of auction results in the past five years indicates that earlier works from the 1930s and 1940s appear to be more desirable, reaching higher prices than later drawings from the 1970s and 1980s. The subject property is an earlier work in the abstract style in very good condition. A survey of retail offerings located \n"),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good. No evidence of damage"),(ItemMediumMaterialsTechniques,"oil on canvas"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signature, date, lower left quadrant"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"15 1/2\" x 19 1/4\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"23 1/2\" x 27 1/2\""),(ItemSupportBaseFrame,"Gilt wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,""),(ItemArtistDate,""),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1975"),(ItemTypeOfObject,"Painting"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"19 Hanson #63")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SnoTIWHxquI/AAAAAAAABzk/D_Os7LpS43U/IMG00021_2.jpg", fileChksum = "61f9a7508ceddcbb780fe757b3cc48bf", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 622, imageFileHeight = 511, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SnoTIWHxquI/AAAAAAAABzk/D_Os7LpS43U/IMG00021_2.jpg", fileChksum = "61f9a7508ceddcbb780fe757b3cc48bf", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 622, imageFileHeight = 511, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "c13a71aec66c814dd585fc6e417fad20", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 174, imageFileHeight = 143, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "ed1f081afe85b605eace3b171c14252c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2191, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_EV171eW_Cmg/SnoTIWHxquI/AAAAAAAABzk/D_Os7LpS43U/IMG00021_2.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"pastel \" Standing Female Nude \" is available, framed and in excellent conditions and as described on our site. The asking price is $ 9.500, but because we are having a special summer promotion, I can give you a discount of 25%, for a net of $ 7.125, including packing, shipping and insurance Pasquale Ianetti Gallery, San Francisco\n\npastel Nude, was offered for $14,000\n"),(ItemCashValue,""),(ItemMarketAnalysis,"Hans Burhardt worked in a variety of media during his long career, including painting, drawing and print making. Known primarily as an abstract impressionist, Burkhardt also worked in a figurative style. Born in Switzerland, Burkhardt came to the United States in ??. He moved to Los Angeles in 1937 and was a well known and well regarded figure in the West Coast art scene. Burkhardt\8217s work has received recent attention due to two major projects, a retrospective of his work mounted at the University Gallery at California State University Northridge, and a comprehensive exhibition of his 1960s paintings presented at the Jack Rutberg Gallery in Los Angeles in 2008-2009. These projects have brought renewed public attention to the artists\8217s work reflected in a greater wolume of sales at auction.\n\nValue characteristics for Burhardt prints include date, size, condition and quality. In the case of his works on paper, a review of auction results in the past five years indicates that earlier works from the 1930s and 1940s appear to be more desirable, reaching higher prices than later drawings from the 1970s and 1980s. The subject property is an earlier work in the abstract style in very good condition. A survey of retail offerings located .. \n"),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good"),(ItemMediumMaterialsTechniques,"Pastel and graphite on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signature, date, lower right quadrant"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"23 1/2\" x 17 1/2\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"25 1/2\" x 19 1/2\""),(ItemSupportBaseFrame,"Glazed silver gilt frame"),(ItemSubjectOrTitle,"Untitled"),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1976"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"20, Hanson #62")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/ShwnA8lJxmI/AAAAAAAAByg/3D8KSo25_5Q/s512/DSC03343.JPG", fileChksum = "eb55c9e8f2b87448ad3fe1b7a5aaf54e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/ShwnA8lJxmI/AAAAAAAAByg/3D8KSo25_5Q/s512/DSC03343.JPG", fileChksum = "eb55c9e8f2b87448ad3fe1b7a5aaf54e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "fdd47b358ebab5ec1b0b0a51cbe41fbc", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 129, imageFileHeight = 194, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "940e99372ee848707da068ff9a2e03a9", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1199, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_s-XE562xJtM/ShwnA8lJxmI/AAAAAAAAByg/3D8KSo25_5Q/s512/DSC03343.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"pastel \" Standing Female Nude \" is available, framed and in excellent conditions and as described on our site. The asking price is $ 9.500, but because we are having a special summer promotion, I can give you a discount of 25%, for a net of $ 7.125, including packing, shipping and insurance Pasquale Ianetti Gallery, San Francisco\n\npastel Nude, was offered for $14,000\n"),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good: Some creasing along the top edge and bottom left corner"),(ItemMediumMaterialsTechniques,"Charcoal on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signature, date L/R"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 17 1/4\" x 23\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"27 1/4\" x 32 3/4\""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Untitled"),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1941"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"21 Hanson #61")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwnC1KiGAI/AAAAAAAABys/EtMJnBzM5AU/s576/DSC03346.JPG", fileChksum = "10d359b37f2a0b3c109762e7f7cd7de6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 576, imageFileHeight = 469, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwnC1KiGAI/AAAAAAAABys/EtMJnBzM5AU/s576/DSC03346.JPG", fileChksum = "10d359b37f2a0b3c109762e7f7cd7de6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 576, imageFileHeight = 469, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "6365c7b50dbfa8a39588c266a3123f1d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 175, imageFileHeight = 143, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "3d50f7e328332edf554de26815f2b282", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2211, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/ShwnC1KiGAI/AAAAAAAABys/EtMJnBzM5AU/s576/DSC03346.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Paid $150 to Cal State Northridge\nOne was offered by Delorme-Collin du Bocage in Paris on 1 July 2009. It was estimated at 2,500-3,000 euros but failed to sell. The second example was offered by Christie's New York on 24 November 2008. It was estimated at $4,000-$5,000, and sold at $3,000.\n\nOne was offered by Delorme-Collin du Bocage in Paris on 1 July 2009. It was estimated at 2,500-3,000 euros but failed to sell. The second example was offered by Christie's New York on 24 November 2008. It was estimated at $4,000-$5,000, and sold at $3,000.\n\n"),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,"Portrait of Native American activist and celebrity Russell Means depicted in a bust view, wearing braids and a traditional bone choker. The background features shades of blue that contrast with the brown skinned figure and the peach colored shirt he is wearing."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,""),(ItemMediumMaterialsTechniques,""),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signed vertically, Andy Warhol"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"49\" x 27\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"61\" x 38 1/2\""),(ItemSupportBaseFrame,"Stained wood frame, glazed"),(ItemSubjectOrTitle,"\"Russell Means\" from the \"American Indian\" series"),(ItemCountryOrNationality,"American"),(ItemArtistDate,""),(ItemArtistOrMaker,"Andy Warhol"),(ItemItemDate,""),(ItemDateOrPeriod,"1972"),(ItemTypeOfObject,"Poster"),(ItemLocation,"Stairwell"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"22 Hanson #65")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/ShwnGFc7K9I/AAAAAAAABzA/0hx7n8lKdoI/s800/DSC03351.JPG", fileChksum = "a5c3128491db2ae76606a856354db49d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 534, imageFileHeight = 800, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/ShwnGFc7K9I/AAAAAAAABzA/0hx7n8lKdoI/s800/DSC03351.JPG", fileChksum = "a5c3128491db2ae76606a856354db49d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 534, imageFileHeight = 800, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "437579f98bb162682cd04bb705fac5be", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 129, imageFileHeight = 194, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "df596a1cec9291a0afd8a0e3cfed0fcb", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1202, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh4.ggpht.com/_s-XE562xJtM/ShwnGFc7K9I/AAAAAAAABzA/0hx7n8lKdoI/s800/DSC03351.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Painted screens available at Ark Antiques are and Edo period, 6 panel screen, 4'6\" high in very good condition and very good quality offered for $2600.\nA small screen, approx. 3'5\" high, 6 panels,  Japanese, 19th century of birds, duck, chickens and foliage available for $1800.\n\n8 panel painted Chinese screen sold at Millea Brothers for $2000 + BP 10/29/06\n\n"),(ItemCashValue,"$2000"),(ItemMarketAnalysis,"This screen was produced for export to the Western markets.  Vintage screens from China are available from importers and at estate sales and regional auction houses. Value characteristics include age, quality of painting, size and condition. The subject property was purchased at an art sale at California State Northridge, therefore the appropriate market for this piece is high-end consignment stores and regional auction houses. Vintage screens comparable to the subject property in size, age and quality were available at local consignment stores and at regional auction houses throughout the country for between $575 and $3,500. A pair of four panel painted Chinese screens depicting boaters on river amid mountains sold at an auction house in New Jersey, Millea Brothers, for $2000 is the basis for the subject property's replacement value.\n "),(ItemDescription,"A six-panel painted screen of ink and colors on paper with a lacquer frame depicting a landscape of misty mountains, tiny villages and boaters along a river."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"California State University at Northridge"),(ItemCondition,"Good: First panel: buckling, a hole in upper quarter; second panel, some slight buckling; third panel: one small buckle in center; fourth panel: horizontal buckling with creases and some loss of pigment; fifth panel: some buckling in center and lower half; sixth panel: slight abrasions towards left center border; gilt paint border: abrasions along areas of borders. "),(ItemMediumMaterialsTechniques,"ink and colors on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"L/L of leftmost panel, signature and title\nSame 2 seals on each panel, alternates lower right and lower left."),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"53 1/2\" x 20\" each of 6 panels 120\" wide overall"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"68\" x 24 1/2\" (each panel)"),(ItemSupportBaseFrame,"Lacquer frame and silk backing"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,"China"),(ItemArtistDate,""),(ItemArtistOrMaker,""),(ItemItemDate,""),(ItemDateOrPeriod,"20th century"),(ItemTypeOfObject,"Chinese screen"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"24")], images = []}},ReportParagraph {elemText = ""},ReportItem {elemItem = Item {itemName = "", fields = [], images = []}},ReportUndecided], reportGlossary = [("Aquatint","The whole plate is covered with grains of resin called an aquatint ground, allowing acid to bite into the entire area, creating an overall grainy, tonal effect. This technique is often combined with etching."),("Drypoint","As in an engraving, the drypoint needle draws the image directly onto the plate. The residue copper is left on the side of the etched lines, which then collect the ink, creating a furry effect called burr."),("Etching","The plate is covered in an acid-resistant layer of wax called an etching ground. The image is then drawn into this surface with an etching needle. When covered with printing ink the lines hold the ink whilst the rest of the plate repels it."),("Lithograph","From the Greek lithos, stone and graphe, writing. This printing process is unlike both intaglio and relief processes, both of which involve cutting into the plate. Lithography relies on the principle that grease and water will repel each other. The image is drawn in a greasy substance onto a lithographic stone. The stone is then dampened with water and the greasy printing ink adheres only to the drawing."),("Second edition","Often an artist will work on a composition to a certain point, and stop to print an impression of it. This single stage in the evolution of this image is called a state. Each time the composition is changed a new state of the print is created. These changes can range from the addition of a plate signature to drastic alterations in the composition. Today artist\8217s will frequently choose to edition a state before moving on with the composition. "),("State","Often an artist will work on a composition to a certain point, and stop to print an impression of it. This single stage in the evolution of this image is called a state. Each time the composition is changed a new state of the print is created. These changes can range from the addition of a plate signature to drastic alterations in the composition. Today artist\8217s will frequently choose to edition a state before moving on with the composition. ")], reportSources = [("Experts Consulted","Jennifer M. Wood\nConsulting and Appraisal Services\n3122 Arizona Avenue, Suite C\nSanta Monica, CA  90404\nPhone: 310 633 1736\nEmail: jenniferwood@roadrunner.com\n"),("Internet Sources","Artfact.com\nhttp://www.artfact.com\n\nArtnet.com\nhttp://www.artnet.com\n\nInternational Fine Print Dealers Association\nhttp://www.printdealers.com\n\nTrocadero.com\nhttp://www.trocadero.com"),("Print Sources","Bartsch, Adam, Zerner, Henri ed., <em> The Illustrated Bartsch: Italian Artists of the Sixteenth Century School of Fountainbleau<\\em> New York: Abaris Books, 1979.\n\nBouvet, Francis <em>Bonnard: The Complete Graphic Work </em> New York: Rizzoli, 1981. \n\nDelteil, Loys, <em> Le Peintre-Graveur Illustre (XIX et XX Siecles) Tome Cinquieme, Corot <\\em> republication of the 1910 Paris edition, New York: Da Capo Press, 1968.\n\nIves, Colta, Giambruni, Newman Sasha M. <em> Pierre Bonnard: The Graphic Art<\\em> New York: The Metropolitan Museum of Art, 1989\n\nLeymarie, Jean, Melot, Michel, <em> The Graphic Works of the Impressionists: Manet, Pissarro, Renoir, Cezanne, Sisley<\\em> New York: Harry N. Abrams, Inc., 1971.\n\nSoucy, Patricia and Janella N. Smyth, eds. <em>The Appraisal of Personal Property: Principles, Theories and Practice Methods for the Professional Appraiser.</em> Washington, D.C.: American Society of Appraisers, 1994.\n\nStella, Joseph G., <em> The Graphic Works of Renoir </em> London: Lund Humphries, 1975.\n\n<em>Uniform Standards of Professional Appraisal Practice, the 2008-2009 Edition</em>(USPAP)(Washington, D.C.: The Appraisal Foundation, 2008)."),("Galleries and Auction Houses","Allinson Gallery, Inc.\n3 Dunham Pond Road East \nStorrs, CT 06268-2022 \n860-429-2322\nhttp://www.allinsongallery.com\n\nBonham's & Butterfield's\n220 San Bruno Avenue\nSan Francisco, CA 94203\n415-86-7500\nhttp://www.bonhams.com\n\nChristie's\n20 Rockefeller Plaza\nNew York, NY 10020\n212-636-2000\nhttp://www.christies.com\n\nDavidson Galleries\n313 Occidental Ave S.\nSeattle, WA 98104\n206 624.6700\nhttp://www.davidsongalleries.com\n\nFigaro Gallery\n703-491-1000\nwww.justfineart.com\nE-mail: inquire@figarogallery.com\n\nGalerie Bassenge Kunst-und Buchauktionen\nErdener Strabe 5a\n14193 Berlin-Grunewald\nGerman\n+49-(0)30-893 80 29-0\n\nI. M. Chait Gallery\n9330 Civic Center Drive\nBeverly Hills, CA 90210\n(800) 775-5020\nhttp://www.chait.com\n\nKamp Galleries\n563 Lincoln Avenue\nWinnetka, IL 60093\n847-441-7999\nhttp://www.kampgallery.com\n\nLeslie Hindman Auctioneers\n\nMasterworks Fine Art Inc.\n13470 Campus Drive\nOakland, CA 94619\nhttp://www.masterworksfineart.com\n510-777-9970\nSpaightwood Galleries")], reportLetterOfTransmittal = "According to our contract signed by you on [ContractDate], [we] have\nprepared the following [ReportType] report for the property\nrepresented as belonging to you.  The subject properties consists of:\n[BriefDescription].\n\nYou provided [us] with the information that indicates that you own the\nsubject [properties] listed above in their entirety.  [We] examined the\n[properties] in the presence of [Client] at [InspectionAddress]\non [InspectionDate].\n\nThis appraisal report provides the appraisal conclusion of\n[ValueType] for the intended use of [IntendedUse].  \n[ValueTypeDescription]\n\nThis report will function as documented evidence of [theseappraisers]\nhaving witnessed the subject [properties], and identified, described and\nvalued the subject [properties] in the context of estimating [ValueType].\n\n\nAfter thorough consideration of available facts and statistics, on\n[EffectiveDate], the [ValueType] for the subject [properties] is\n<b>[TotalValueAsText] ([TotalValue])</b>.\nThe effective date of value for this appraisal is [EffectiveDate].\nThis figure does not reflect a consideration for related sales\ntaxes and delivery/installation charges, which can vary.  This\nprofessional opinion is based upon [our] knowledge and experience in the\nvaluation of fine arts.  This appraisal is based on [our] research and\nanalysis of the [properties] and knowledge of relevant markets where the\n[properties] [are] traded.\n\nThis is a [ReportType] report that was prepared to conform to\nthe Code of Ethics of the American Society of Appraisers and the\ncurrent version of the [USPAP],\npublished by the Appraisal Foundation.  We have\nno past, present or contemplated future interest in the [properties]\nincluded in this report.  Our fee was not contingent on any value\nfinding.\n\nThis [ReportLength] page report consists of a letter of transmittal, scope of\nwork, definitions, privacy statement, assumptions and limiting\nconditions, certification, description, photographs, narrative,\nglossary, bibliography and credentials.  All these elements must be\npresent for the value conclusion to be valid.\n\nThe information in this report is confidential.  All notes,\nelectronic, work copies and photographs remain the property of the\nappraiser and are retained for five years, and/or two years after the\nconclusion of any litigation.  All documents are shredded or otherwise\ndefaced prior to disposal.\n\nThe submission of this report completes [our] assignment and\nobligations.  Any further consultation, preparation of additional\nreports, testimony, or appearances in court or before any governmental\nforum in reference to the [properties] described in this appraisal will\nrequire additional fees to cover the required services.  The\naforementioned fees will be billed at the rate commensurate with the\ntiming of the appraisal.\n\nThank you for allowing [us] to be of service to you.\n", reportScopeOfWork = "Thompson & Martinez Fine Art Appraisals, Inc. was hired by [Client] to reach an opinion of the [ValueType] of [BriefDescription], for the intended use of [IntendedUse].  Based on [our] discussion with [Client], as well as the intended use of the report, the type and location of the subject [properties] and the effective date of valuation, the [appraisers] [have] selected the type of value, the relevant market, the information researched and the analysis to be applied.  To complete this assignment [we] consulted authoritative publications and on-line sources, reviewed published and on-line market data and interviewed knowledgeable dealers and others regarding the artists, their work, and the relevant markets. Kathi Jablonsky ISA, CAPP assisted in the research of the display cabinet and chairs. Jo Ellen Cole of Cole Appraisal Services consulted on the identification of the jade and bowenite sculptures.\n\nThis report is intended to be an objective, independent, unbiased and credible analysis of known facts.  During the course of our examination, research and analysis, [we] have endeavored to be accurate and to use due diligence.  This opinion of value was developed using appraisal methods that are within the accepted standard of care for the appraisal profession and within the boundaries of typical practice. \n\nThis appraisal document summarizes the majority of information analyzed, the appraisal procedures followed, and the reasoning that supports the analysis, opinions and conclusions under the requirements of the [USPAP].  It is considered a summary report.  The documentation substantiating the value reported is held in the appraiser's file. ", reportCertification = ["To the best of [our] knowledge and belief, the statements of fact continued in this report are true and correct.  The value conclusions are based on [our] background and experience outlined in [our] credentials attached at the end of this report.","The reported analyses, opinions and conclusions are limited only by the reported assumptions and liming conditions.  They are [our] personal, impartial and unbiased professional analyses, opinions and conclusions.","[Our] analyses, opinions and conclusions in this report were developed and prepared in conformity with the [USPAP]. ","Unless otherwise stated, the value given in the Appraisal Report represents the opinion of value as of the date of inspection.","[We] have no bias with respect to the property that is the subject of this report, or to the parties involved with this assignment.","[Our] compensation for completing this assignment is not contingent upon development or reporting of a predetermined value or direction in value that favors the cause of the client, the amount of the value opinion, the attainment of a stipulated result, or the occurrence of a subsequent event directly related to the intended use of this appraisal.","[We] have made a personal inspection of the property that is the subject of this report.","[These] [appraisers] prepared most conclusions and opinions set forth in this Appraisal Report. Kathi Jablonsky, ISA, CAPP assisted in the research of the display cabinet and pair of chairs.","Although all due care has been taken to ensure that any statement as to the authorship, attribution, origin, date, age, provenance, condition, and value of any item of personal property included in this report is accurate, all such statements are to the best of the knowledge of the [appraisers] who completed this report, and are not to be relied upon as representation of fact."], reportLimitingConditions = ["Title to the property and present ownership of the property is assumed to be good and marketable unless otherwise stated.  Further, the appraisal is made with the understanding that the present ownership of the subject property includes all rights that may be lawfully owned and is not subject to any encumbrances or indebtedness.  The Clients stated that they own the property; ownership was not researched or confirmed.","The [appraisers] [have] no present or prospective interest in the property that is the subject of this report, and no specified personal interest with respect to the parties involved.  The reported analyses, opinions and conclusions are limited only by the reported assumptions and limiting conditions, and are personal, impartial and unbiased professional analyses, opinions and conclusions.","The [appraisers'] analyses, opinions, and conclusions [were] developed, and this report has been prepared, in conformity with the Code of Ethics as set forth by the American Society of Appraisers and the [USPAP] of the Appraisal Foundation, Washington, D.C.","The Appraisal is based on visual inspection of the appraised item.  No tests of any kind have been administered unless so stated.  Measurements and photographs are made under ``field conditions,'' and therefore will be as accurate as possible.  Measurements are listed in the order of height, width and depth, unless otherwise indicated.","The opinion of value expressed in this Appraisal Report is the result of and is subject to the data and research described in detail in the Appraisal Report.  In the preparation of this report, the [appraisers] [have] consulted a variety of sources that are documented either in the report or are available in the [appraisers'] files.  These sources are assumed to be reliable and accurate; however, the [appraisers] [take] no responsibility for errors or omissions contained in these sources.","This Appraisal Report, or a copy of this Appraisal Report, may be transmitted to a third party or legal entity only in its entirety and only with the expressed written consent of the appraisal client.","Disclosure of the contents of the Appraisal Report is governed by the Bylaws and Regulations of the Appraisal Foundation [USPAPShort].  Neither all nor any part of the contents of this report (especially any conclusions as to the value or the identity of the Appraisers) shall be disseminated to the public through advertising, public relations, news, sales, or the media without the prior written consent and approval of the appraiser.","Information furnished by the owner regarding property history is believed to be reliable; however, no warranty is given for its accuracy.  ","The [appraisers] assumes no responsibility for unforeseen changes in market conditions, nor is the inability to locate a potential buyer or replacement item, at the appraised value, if such is attempted by the client or any other person. The value conclusion is not an offer to buy or sell the property.","In the event of a dispute involving interpretation or application of this report, the American Society of Appraisers recommends that such a dispute shall be referred to a neutral third-party mediation service.  The cost of such mediation shall be borne equally by both parties.  In the event mediation is not successful, the parties should agree to submit the dispute to binding arbitration under the laws of the State of California.","The [appraisers] [have] personally inspected the subject property.  The condition of the property is described sufficiently for its type, with any visible and/or relevant repairs noted. The condition of the property may change after inspection and the effective date of valuation.","The opinion of value expressed in the Appraisal Report is the result of and is subject to the data and research described in detail in the Appraisal Report and on data retained in the [appraisers'] files.  The [appraisers] [have] not knowingly withheld any significant information.","Appraisers witness, identify and value subject property but do not claim authentication of the subject properties, unless otherwise noted.","This Appraisal report is valid only for the intended use, intended users and for the type of value stated in the appraisal.  The values reported in this appraisal are valid only on the effective date of the appraisal [EffectiveDate].","The submission of this report completes [our] assignment and obligations.  Any further consultation, preparation of additional reports, testimony, or appearances in court or before any governmental forum in reference to the property described in this appraisal will require additional fees to cover the required services."], reportPrivacyPolicy = "Thompson & Martinez Fine Art Appraisals, Inc. makes significant efforts to keep all appraisal-related data confidential in compliance with the current version of the [USPAP] as published by The Appraisal Foundation. We will not disclose confidential appraisal data to any third party without the written permission of the rightful owner of that data. We will not release a copy of an appraisal you have contracted through Thompson & Martinez Fine Art Appraisals, Inc. to any third party without the written permission of the client.\n\nOur policy specifies the protection of client information in the following ways:\n * We limit access to all information only to those who have a business or professional reason to know and only to nonaffiliated parties as required by law.\n * We maintain a secure office and computer environment to ensure that your information is not placed at unreasonable risk.\n * The categories of non-public personal information that we collect from a client depend upon the scope of the client's appraisal engagement.  This includes information about personal property assets, information about tax identification numbers, and other non-public information necessary.  Additionally, it may include information about transactions between clients and third parties and information form third party sources.\n * Unaffiliated third parties that require access to our clients' personal information, such as Federal and State tax regulators, insurance companies, and the like may only review company records as permitted under law.\n * We do not provide client information to mailing list vendors or solicitors for any reason.\n * Personally identifiable information will be maintained during the time a client is a client of the company and for 5 years after the appraisal has been completed and/or 2 years after litigation as well as the ethical considerations as specified by [USPAPShort].  After this required period of record retention, all such information will be destroyed.\n", reportMessages = "", reportPerms = Just (Permissions {owner = Username {unUsername = "lydiat"}, writers = [Username {unUsername = "Natasha"},Username {unUsername = "david"},Username {unUsername = "lydia"}], readers = []}), revisionInfo = defaultValue}
l = Report {reportFolder = "", reportName = "Laurie Hanson", reportDate = "", reportContractDate = "", reportInspectionDate = "", reportEffectiveDate = "", reportAuthors = [Author {authorName = "Lydia Thompson, Ph.D.", authorCredentials = "====Education\n * Ph.D., New York University (NYU), Institute of Fine Arts (IFA), New York, NY, 1998.  Art and Archaeology.  Specialized China\n * M.A., NYU, Institute of Fine Arts (IFA), New York, NY, 1989.  Art and Archaeology\n * B.A., Middlebury College, Middlebury, VT, 1983.  East Asian Studies\n====Professional Study\n * University of California, Irvine, Certificate Program in Appraisal Studies, UCI in partnership with American Society of Appraisers (ASA).  Completion of all P.O.V. courses in certificate program\n * Shandong University, Jinan, Shandong, PRC.  Department of Archaeology, 1993-94.\n====Awards\n * Kress Foundation, Dissertation Fellowship, 1995-96.  Center for Advanced Study in the Visual Arts (CASVA), National Gallery of Art, Ittleson Fellow, 1993-95. Samuel F.B. Morse Fellowship, 1991-92\n====Memberships\n * Candidate Member, American Society of Appraisers\n====Selected List of Professional Experience 2000-2008\n * President, Thompson and Martinez Fine Art Appraisals, Inc., La Jolla, CA\n * Curator and Docent Educator for the exhibition: ``ymbols of Power:  Masterpieces from the Nanjing Museum'', the Bowers Museum of Cultural Art, Santa Ana, CA\n * Adjunct Faculty in Art History, UC Los Angeles and UC San Diego, CA.  \n====Selected Publications\n * ``Ritual, Art and Agency:  Consecrating the Burial Ground in the Han Period'' <em>Recarving China's Past:  Art, Archaeology, and Architecture of the Wu Family Shrines</em>.  Proceedings of a symposium held at the Princeton University Art Museum, (Princeton U.P.:  forthcoming)\n * Entries on artists and translation of essay by Art Critic Li Xianting, Betti-Sue Hertz ed. <em>Past in Reverse:  Contemporary Art in East Asia</em> (San Diego:  San Diego Museum of Art, 2004)\n * Co-author with Xu Huping, <em>Symbols of Power:  Masterpieces from the Nanjing Museum</em>, (Santa Ana:  The Bowers Museum of Cultural Art, 2002)\n"},Author {authorName = "Natasha Bonilla Martinez, MA", authorCredentials = "====Education\n * Certificate, University of California, Irvine Extension, Certificate Program in Appraisal Studies, UCI in partnership with American Society of Appraisers (ASA), 2007\n * Certificate, USPAP, University of California, Irvine Extension, Certificate Program in Appraisal Studies, UCI in partnership with American Society of Appraisers (ASA), 2006.\n * M.A. San Diego State University, Latin American Studies, San Diego, CA. 2001 Specialization in Latin American and Latino photography and art.\n * Graduate study. New York University, Department of Anthropology, New York, NY. 1982 to 1986, photographic anthropology and Native American art and culture.\n * B.A.  Vassar College, Poughkeepsie, NY, 1981. Independent Major.\n====Awards\n * Gateways Rockefeller Fellowship, 1997, Guadalupe Cultural Arts Center, San Antonio, TX, for the study of contemporary photography by Mexican and Mexican American women.\n====Selected Professional Experience 2000-2008\n * Vice-President, Thompson and Martinez Fine Art Appraisals, Inc., La Jolla, CA.\n * Director of Education, Director of Visual Arts and Education and Director of Program Development, successively, California Center for the Arts, Escondido, Escondido CA. \n * Co-Curator of Niki de Saint Phalle: California Dreaming, California Center for the Arts, Escondido, Escondido CA.\n * Co-Curator, Spirit Capture: Photography and Native America exhibition, National Museum of the American Indian, Smithsonian Institution, New York, NY.\n * Adjunct faculty, Department of Art, University of San Diego, San Diego, CA. Course on US/Mexico border art.\n====Selected Publications\n * ``An Indian Americas'' and ``Camera Shots'' in <em>Spirit Capture: Photographs from the National Museum of the American Indian</em>. (Washington, D.C.: Smithsonian Press, 1998).\n====Memberships\n * Candidate member, American Society of Appraisers\n * American Association of Museums \n * Association for Latin American Art, College Art Association\n"}], reportPreparer = "", reportPreparerEIN = "(preparer EIN)", reportPreparerAddress = "", reportPreparerEMail = "", reportPreparerWebsite = "", reportAbbrevs = [("[EffectiveDate]","September 17th, 2008"),("agree","agrees"),("appraisers","Appraiser"),("appraisers'","Appraiser's"),("are","is"),("assume","assumes"),("certify","certifies"),("Client","Mr. Ron Smith"),("EIN","20-8308057"),("have","has"),("InspectionAddress","17427 Francisco Drive\nSan Diego, CA 92128"),("intendeduse","Insurance purposes"),("LOTInspectedBy","Lydia Thompson, Ph.D."),("Our","My"),("our","my"),("Preparer","Thompson & Martinez Fine Art Appraisals, Inc.\nEIN: 20-8308057\n"),("preparer Email","appraiser@thompsonandmartinez.com"),("preparer Website","http://www.thompsonandmartinez.com"),("properties","properties"),("ReportType","summary"),("ReportUsers","Mr. Ron Smith and his insurance company"),("soucysmith","(from Soucy and Smith, eds, <em>The Appraisal of Personal Property - Principles, Theories, and \nPractice Methods for the Professional Appraiser\n</em>, 1994.)"),("take","takes"),("These","This"),("these","these"),("theseappraisers","this appraiser"),("Us","Us"),("us","us"),("USPAP","<em>2008-2009 Uniform Standards of Professional Practice (USPAP)</em>"),("USPAPShort","USPAP 2008-2009"),("ValueTypeLongDescription","Replacement Value is the price required to replace a property with another of similar age, quality, origin, appearance and condition, within a reasonable length of time in an appropriate and relevant market (from Soucy and Smith, eds, <em>The Appraisal of Personal Property - Principles, Theories, and \nPractice Methods for the Professional Appraiser\n</em>, 1994.) \n\nBecause of the time element involved, the appropriate market for replacement is deemed to be that of primary sources, such as art galleries carrying comparable and immediately available personal properties. If the only source for replacement can be demonstrated to be that of auction, that is the source utilized. Replacement Value is used as the basis for insurance purposes and damage/loss reports. "),("We","I"),("we","I"),("were","was")], reportTitle = "", reportIntendedUse = Just Insurance, reportValueType = Just ReplacementValue, reportValueApproach = Just MarketComparisonApproach, reportClient = "", reportClientGreeting = "Ms Hanson", reportItemsOwnerFull = "", reportItemsOwner = "", reportBriefItems = "", reportInspectionLocation = "", reportBody = [ReportParagraph {elemText = "Since the inception of the financial crisis and economic downturn in the fall of 2008, the art market has dropped precipitously. Well-known auction houses have seen their sales volume drop 70% or more since their highs in the spring of 2008 and, in several categories of art, prices have returned to 2002-2003 levels (5/21/09 Editorial by Swann President Nicholas Lowry). \n\nSome categories of art, however, have fared better in the downturn than others. Contemporary art, which realized spectacular price increases in the past 5 years, has been the hardest hit with average prices of high-end contemporary art dropping two-thirds in the evening sales. On the other hand, average prices of antiquities and Old Masters have softened but are still double what they were in 2003 (Art & Auction, June 2009, databank, p. 204). Sales of the impressionists at spring auctions showed surprising strength generally exceeding expectations (Art & Auction, July/August, 2009) Auction Reviews, p. 72.\n\nMs. Hanson's collection consists of prints by impressionists and old Masters as well as prints and paintings by well-known 20th century artists. Also included is a decorative Chinese painted screen.\n\nThere is a limited market for antique European prints in southern California, with only a few galleries on the West Coast, with one of the largest the Davidson Gallery in Seattle. The majority of galleries carrying this type of material is based in the East Coast of the United States and Europe. Although the customary practice is to seek comparable sales in the market where the insured shops, the unavailability of Old Master prints in the region required the Appraisers to seek comparable sales elsewhere in the national market. Given the quality of the works comparables were sought among dealers of high quality prints and members of the International Fine Art Print Dealer's Association, as well as in the auction market as appropriate.\n\nA significant element of the Hanson collection are several works by Hans Burkhardt (1904-1994), the Swiss American abstract impressionist, including a painting, works on paper and prints. A well known figure in Southern California, Burkhardt is especially known for his large scale anti-war paintings that expresses his response to the Spanish Civil War, World War II, Vietnam and Desert Storm. Recent major installations by the Jack Rutberg Gallery in Los Angeles and the California State University Northridge gallery have brought renewed attention to Burkhardt's work, which appears to be reflected in interest and activity at recent auction offerings at Sotheby's and Christie's.\n\n"},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,""),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,"Reverse composition features abstracted seated nude female figure on red field. Emphasis is on breasts and thighs with no detailing of face or lower extremities."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"purchased from the artist ca. 1985"),(ItemCondition,"Excellent"),(ItemMediumMaterialsTechniques,"woodcut on handmade paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"12/16"),(ItemSignatureInscriptionsMarkings,"Numbered lower left margin; signed and dated lower right margin; embossed stamp lower right margin"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Untitled"),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1972"),(ItemTypeOfObject,"Print"),(ItemLocation,"Dining room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"Hanson #26")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SoSr8iQyWJI/AAAAAAAAB0E/hscFnMAb58c/s512/IMG00013.jpg", fileChksum = "bec5bacaf92f4d696a29e2e9e7744f01", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 384, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SoSr8iQyWJI/AAAAAAAAB0E/hscFnMAb58c/s512/IMG00013.jpg", fileChksum = "bec5bacaf92f4d696a29e2e9e7744f01", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 384, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "8e585423d482abea53f36ac243f87b67", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 137, imageFileHeight = 183, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "33d652f5435a93658161916f5ef91d11", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1350, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_EV171eW_Cmg/SoSr8iQyWJI/AAAAAAAAB0E/hscFnMAb58c/s512/IMG00013.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"print acquired from the artist in 1985 according to Ms. Hanson"),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,"Three color composition features man wearing hat bending down to pick up a pepper on a sidewalk. The pepper is red and yellow, while the balance of the composition is black and white."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"purchased from the artist ca 1985"),(ItemCondition,"Very Good"),(ItemMediumMaterialsTechniques,"woodcut"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"8/12"),(ItemSignatureInscriptionsMarkings,"signed and numbered, lower right margin; embossed stamp lower right margin"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"\"Mark Tobey\""),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1975"),(ItemTypeOfObject,"Print"),(ItemLocation,"Dining room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"Hanson #27")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_EV171eW_Cmg/SoSrvxqq2JI/AAAAAAAAB0A/8y9y53D2YKo/s512/IMG00016_2.jpg", fileChksum = "79221c6448716b08b8ab19085b6bc582", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 392, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_EV171eW_Cmg/SoSrvxqq2JI/AAAAAAAAB0A/8y9y53D2YKo/s512/IMG00016_2.jpg", fileChksum = "79221c6448716b08b8ab19085b6bc582", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 392, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "90c3fc977a59abfda729d9fb50b0137f", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 138, imageFileHeight = 181, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "d21aad082a06bb4e5790de5ed9166967", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1378, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_EV171eW_Cmg/SoSrvxqq2JI/AAAAAAAAB0A/8y9y53D2YKo/s512/IMG00016_2.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Venator & Hanstein KG (Butcher Graphik Autographen Auuktion 100) 3/23/07 #548 estimated $1330 (EU 1000) bought in\n\nSame print currently available for $950 at the Davidson Gallery.\n\nAuction records of print published in Gazette de Beux-Arts indicate this print sell for between $431, $447, $632.\n\nDelteil 1/iv\n\nC&J Goodfriend offers Souvenir de Toscane 2nd state print, dated 1845 for $2500\n\nMasterworks Fine Art Inc. has recently sold a 4th state Souvenir de Toscane but don't know price.\nRecently sold (4th state, A. Salmon printer, Gazette des Beaux-Arts at Masterworks Fine Art Inc. \n"),(ItemCashValue,"$950"),(ItemMarketAnalysis,"Jean-Baptiste-Camille Corot was the leading painter of the Barbizon school of France in the mid-nineteenth century. He is a pivotal figure in landscape painting and his vast output simultaneously references the Neo-Classical tradition and anticipates the plein-air innovations of Impressionism. He was also an engraver. ``Souvenir de Toscane'' was originally executed in 1845 and produced in four states.\n\nValue characteristics include: state, edition, subject matter and condition. State is a significant characteristic of value for this print with impressions published before the Gazette des Beaux-Arts edition (state 4) being especially scarce. A second state version of Souvenir de Toscane is available at C&J. Goodfriend Galleries for $2500. The subject property is an example of the 4th state published in Gazette des Beaux-Arts in 1875 whA second state version of Souvenir de Toscane is available at C&J. Goodfriend Galleries for $2500. between $250 and $630.  A 4th state print in similar condition and impression to the subject property is currently offered for sale at the Davidson Galleries for $950 and is the basis for estimating its replacement value."),(ItemDescription,"Landscape of Tuscan countryside with rolling hills and trees in the foreground framing a domed structure in the center middle-ground."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Gallery"),(ItemCondition,"Very Good: dark spot (mold?) lower margin center. Paper appears toned, beige, almost sepia"),(ItemMediumMaterialsTechniques,"Etching on laid paper "),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Edition of 750, 4th state, Delteil 1, iv., Melot 1,v."),(ItemSignatureInscriptionsMarkings,"Bottom left: Corot Del. & Next line Gazette des Beaux-Arts bottom right: Imp. A. Salmon, Paris."),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Plate impression: 5\" x 7\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"11 3/4\" x 14 1/2\""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Souvenir de Toscane"),(ItemCountryOrNationality,"France"),(ItemArtistDate,"1796-1875"),(ItemArtistOrMaker,"Jean Baptiste Camille Corot"),(ItemItemDate,""),(ItemDateOrPeriod,"c. 1843, published April 1st, 1875. Printed by A. Salmon, Paris"),(ItemTypeOfObject,"Print"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"1 Hanson #41")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/Shlbi8G5awI/AAAAAAAABoU/NxQ-H2USxAo/s800/DSC_0281.JPG", fileChksum = "b743dc61d9721ee357c2b0d8555d5dac", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/Shlbi8G5awI/AAAAAAAABoU/NxQ-H2USxAo/s800/DSC_0281.JPG", fileChksum = "b743dc61d9721ee357c2b0d8555d5dac", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "30cbc4b81ec19aa1ba6b1441a264df7b", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 194, imageFileHeight = 129, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "da929ce67ba02b3f079f047d0a712aba", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2707, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/Shlbi8G5awI/AAAAAAAABoU/NxQ-H2USxAo/s800/DSC_0281.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Gallery label on back. Original print 1904. Published in 1919.\n\nAccording to the Catalog Raisonne, the original lithograph is from 1904.\n\nThis print has sold at auction for between $1000 and $6000 mostly from this edition.\n\nA single print from the edition of 950 (``on wove paper'') sold at Christie's New York 5/2/07 for $6000.\n\nAnother print ``Ambroise Vollard\" of the print-maker Vollard from the same series: ``Douze Lithographies Originales\" of 1919 is available from Davidson for $3800.\n\nMasterworks Fine Art: Claude Renoir, La Tete Baissee. Original print on Japon paper, edition of 50. Offered for $8700.\n\nArmstrong Fine Art\n``I currently have an impression of \"Claude Renoir, la Tete Baissee\" in pristine condition for $5,500.  It is a beautiful impression on a full sheet of Arches ``wove paper'' (implies vellum??, see above).  There are no tears, creases, discolorations, etc.  There are only two acid-free cloth hinges on the verso at the top of the sheet, which could easily be removed.  Aside from that, it is perfect.''\n\nIf you have further questions, please do not hesitate to contact me.\n \nSincerely, Bernard\n \nBernard Derroitte\nArmstrong Fine Art\n200 East Walton Place\nChicago, IL  60611\nT.  312.664.9312; F.  312.664.9862\nwww.armstrongfineart.com\n\nLouis Valtat. c. 1904. Lithograph. Delteil, Stella 38. 11 5/8 x 9 1/4 (sheet 13 x 9 3/4). Series: Douze Lithographies Originales de Pierre-Auguste Renoir. A fine impression apart from the edition of 1,000 printed on cream laid paper with deckle edges on three sides. Signed in the stone. $4,500.\n"),(ItemCashValue,"$4500"),(ItemMarketAnalysis,"Renowned French Impressionist painter Pierre-Auguste Renoir took up printmaking later in life after 1890, and is thought to have executed some 59 lithographs and etchings, some of which are regarded as some of the best of the period. Renoir was the first of the Impressionists to enjoy popular success and his prints were produced in recognition of that success. Stella lists 60 prints (25 etchings and 35 lithographs) in his catalogue of Renoir's prints. Michel Melot, The Impressionist Print notes that most of Renoir's prints were produced after he had became famous and represent an attempt to satisfy the demand for his works, almost 3/4 of them having been produced after 1900 for inclusion in books and portfolios produced by Theodore Duret, the chief chronicler of the Impressionists, and Ambroise Vollard, Renoir's dealer and a great lover of the art of printmaking. One of Renoir's subjects of choice was portraits of children and friends.\n\nValue characteristics for Renoir lithographs include quality of impression, state and condition and type of paper. Executed in 1904, this is an impression from 2nd and final state which was published for ``L'Album des Douze Lithographies Originales de Pierre August Renoir\" by A. Vollard (Paris, 1919) in and edition of 1000, with 50 printed on Japon paper and 950 printed on Arches wove paper. Some images from this series were also printed separately on laid paper. At Allinson Gallery, a print from this series on laid paper is available for $4500.\n\n\nExamples of the print are available at retail with the print from the edition of 950 available for  $5500, and the two from the edition of 50 for $6500 and $8700."),(ItemDescription,""),(ItemCost,"$785"),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Galleries"),(ItemCondition,"Very good. Discoloration to upper horizontal and right vertical margins. No tape residue on reverse. UV light damage to the sheet. Deckled edges."),(ItemMediumMaterialsTechniques,"On laid paper. No watermark."),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Second state. Edition: Ambroise Vollard, Paris, 1919. Printer: Auguste Clot. Edition of 1000. One from the series, \"Douze Lithographies Originales\". D. 39"),(ItemSignatureInscriptionsMarkings,"Signature L/R, signed in stone"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Mat opening: 9 3/4'' x 8''\nSheet: 13\" x 10\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"21 5/8'' x 17 5/8''"),(ItemSupportBaseFrame,"Glazed gilt wood frame"),(ItemSubjectOrTitle,"Claude Renoir, La Tete Baissee"),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1841-1919"),(ItemArtistOrMaker,"Auguste Renoir"),(ItemItemDate,""),(ItemDateOrPeriod,"c. 1904"),(ItemTypeOfObject,"Print (Davidson)"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"2 Hanson #37")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SmTiIV_kBoI/AAAAAAAAC-4/GfT1RJYXZXc/s800/Renoir.jpg", fileChksum = "340ae2f791e548f90b9a345b32cf894e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 600, imageFileHeight = 678, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SmTiIV_kBoI/AAAAAAAAC-4/GfT1RJYXZXc/s800/Renoir.jpg", fileChksum = "340ae2f791e548f90b9a345b32cf894e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 600, imageFileHeight = 678, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "d8a06582fa6ab7548c1a1d3f3a1c8a03", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 149, imageFileHeight = 168, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "7b3d6e27ca4467778a146c3df16d84b2", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1593, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/SmTiIV_kBoI/AAAAAAAAC-4/GfT1RJYXZXc/s800/Renoir.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"\"A Basement Shop in Cracow\" $2,000. Lockport Street Gallery\n\nAu Pied du Sinai: Schlome Fuss A La Synague, ?, Lewis and Maese Auctions, Houston, TX, auction est $3,750 - $4,250 lithograph in black-brown on china paper, signed in pencil Lower right. Rare - 25 proofs, deluxe book edition iii.\n"),(ItemCashValue,"$2,000"),(ItemMarketAnalysis,"Henri Toulouse-Lautrec was a French painter, printmaker, and illustrator best known for his depictions of life in the bohemian world of the Parisian theater and its environs at the end of the 19th century. The subject property is from a series of works depicting Jewish life that illustrated a volume of sketches or travel essays on Eastern European Jewish life by Georges Clemenceau, who later served as prime minister of France. Toulouse-Lautrec is said to have observed Jewish life in Paris as the basis for his illustrations.\n\nThere is one state only of this lithograph, one of a set of eight published by Henry Floury. The group of eight lithographs were published in a deluxe book edition with four variations, and in a regular book edition in two variations.\n\nValue characteristics include edition and condition. The subject property is from the regular book edition, var. ii, as indicated by its sheet size, and is a good quality impression in good condition. This print occasionally appears at auction and at retail, as part of the set of eight and also as a single work. A national survey of print dealers located two offerings ranging from $2000 to $4500.  The higher comparable is for a print from the Deluxe book edition, and the lower comparable for a print from the Regular Book edition similar to the subject property. The lower comparable is the basis for the value conclusion of Replacement Value Comparable of $2,000."),(ItemDescription,"The composition features an elderly man with a long beard and wearing a coat and hat, looking out of a basement level window adjacent to a sidewalk. His hands are resting on the window ledge at the same level as the sidewalk. A small black cat sits at the edge of the window near the man's right hand. The window is lined with large blocks, evoking a an older stone building. The composition emphasizes shades of gray, in the figure of the man window and building, so that the black cat serves as a focal point."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Good impression in very good condition. Minor UV light damage. No adhesive or tape on reverse."),(ItemMediumMaterialsTechniques,"Lithograph on wove paper, no watermark"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Series: \"Au Pied du Sinai\", 1 state only"),(ItemSignatureInscriptionsMarkings,"Unsigned"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Image impression: 170 mm x 140 mm; Sheet size: 260 mm x 196mm"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"16\" x 13 5/8\""),(ItemSupportBaseFrame,"Gilt wood frame, glazed"),(ItemSubjectOrTitle,"\"Une Arriere-boutique a Cracovie\" (A Basement Shop) from the series \"Au pied du Sinai\"(From the Foot of Sinai)"),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1864-1901"),(ItemArtistOrMaker,"Henri Toulouse-Lautrec"),(ItemItemDate,""),(ItemDateOrPeriod,"1898"),(ItemTypeOfObject,"Print"),(ItemLocation,"Living room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"4 Hanson #36")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "\"Une Arriere-boutique a Cracovie\"", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SmThoUC_IxI/AAAAAAAAC-o/T4jaLSFLz0c/s400/Toulouse%20Lautrec1.jpg", fileChksum = "ed9232e7fa5b87ba495cacce48c03ee9", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 321, imageFileHeight = 400, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SmThoUC_IxI/AAAAAAAAC-o/T4jaLSFLz0c/s400/Toulouse%20Lautrec1.jpg", fileChksum = "ed9232e7fa5b87ba495cacce48c03ee9", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 321, imageFileHeight = 400, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "9e82c2f080523b6b49be954d3aea9a4e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 142, imageFileHeight = 177, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "20b6fc9e14bd2f4041c1c071c35a6b10", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1445, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_s-XE562xJtM/SmThoUC_IxI/AAAAAAAAC-o/T4jaLSFLz0c/s400/Toulouse%20Lautrec1.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Etching, \8220Tampoco\8221 from \8220Desastres de la Guerra\8221,  is $2,250., David Tunick, Inc.\n\nEtching \8220Bien te esta\8221 from from \8220Desastres de la Guerra\8221, proof impression is $140,000, Harris Schrank\n\n"),(ItemCashValue,""),(ItemMarketAnalysis,"Francisco de Goya, the Spanish master painter, illustrator and engraver of the 19th century is well known for his series of prints describing the human cost of the Napoleonic Wars in Spain called \8220Los desastres de la Guerra,\8221 (The Disasters of War). The series includes some 80 prints of which only some proofs and individual prints were made during the artist\8217s lifetime. The first published edition was issued posthumously in 1864, and ultimately seven editions were printed, in varying degrees of quality.\n\nValue characteristics for Goya prints from this series include edition, condition, quality of impression, and plate or subject. At auction the value range is dramatic, with a rare proof of \"Bien te esta\" reaching more than $140,000 to later edition lesser quality impressions selling for less than $500. The subject property, an impression of plate 45 titled \8220Tambien estos\8221, is from the first edition. It  is a good quality impression in very good condition. A national survey of galeries did not locate any prints of this plate, but did locate several first edition prints of other plates in the series. These range from ? to ?. The subject property, is a very good impression in generally good condition.\n"),(ItemDescription,"The composition depicts a room full of wounded soldiers, some lying on the floor and on beds, others seated and standing. Three figures form a triangle in the composition: the seated figure on a bed in the foreground, the standing figure looking over his shoulder in the center and the figure lying on the right in the middle."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very good impression in good condition. Upper and lower margins have been trimmed. Mat burn but no foxing and UV light exposure. Remnants of tape residue on reverse of the sheet. Inscribed along bottom margin. \"1-13-37\". Inscription on reverse of sheet L/R corner: ``Disasters of war, 1st edition''. "),(ItemMediumMaterialsTechniques,"Etching on wove paper with no watermark."),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Series: Los Desastres de la Guerra, first edition"),(ItemSignatureInscriptionsMarkings,"Unsigned, Composition L/L; no. 25 in U/R margin"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Image size: 4 1/2'' x 7 1/2'' (11 1/2 cm. x 19.2 cm); \nPlate impression: 6 1/8'' x 9 1/2'' (15 1/2 cm x 23 1/2 cm)\nSheet size: ???"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"19\" x 20 3/4''"),(ItemSupportBaseFrame,"Glaze gilt wood frame"),(ItemSubjectOrTitle,"\"Tambien Estos\" (Also these) from the \"Los desastres de la guerra\" (Disasters of War) series"),(ItemCountryOrNationality,"Spanish"),(ItemArtistDate,"1746-1828"),(ItemArtistOrMaker,"Francisco de Goya "),(ItemItemDate,""),(ItemDateOrPeriod,"1898"),(ItemTypeOfObject,"Print"),(ItemLocation,"Living room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"5 Hanson #32")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/SmThoXPWNmI/AAAAAAAAC-I/4U5lZrDuWkA/s640/Goya1.jpg", fileChksum = "10c5286b55fa1707077e866f3c8bbad5", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 640, imageFileHeight = 479, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/SmThoXPWNmI/AAAAAAAAC-I/4U5lZrDuWkA/s640/Goya1.jpg", fileChksum = "10c5286b55fa1707077e866f3c8bbad5", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 640, imageFileHeight = 479, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "9ad9cea61d166acdf7a95b491f2faa36", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 183, imageFileHeight = 137, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "f3b70a8be1092118de3857829f4f6465", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2405, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh4.ggpht.com/_s-XE562xJtM/SmThoXPWNmI/AAAAAAAAC-I/4U5lZrDuWkA/s640/Goya1.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Gallery label on back\n\nJayne Haslem Gallery, The Zither, 9 7/8'' x 10 3/4\" Aquatint and etching, $1500\n\nThe Annex Gallery, The Zither, 24/60 published 1982. $750.\n\nKiechel Fine Art, August, 1940 etching 9 3/4\" x 8\" $1200\n\nEtchings and acquatints of the same era have sold at auction for between $418 and $3200 with the majority in the $1100- $2400 range."),(ItemCashValue,"$1500"),(ItemMarketAnalysis,"Born in Beverly, Massachusettes in 1911 American contemporary artist, Will Barnet is an acclaimed painter and printmaker, who has also taught art at the leading art schools in the U.S.\n\nTrained at the Boston Museum of Fine Arts School and the Art Students League New York, he later taught art at such leading American schools as Yale University, Cornell University, and the Art Students League.\n\nA prolific graphic artist, Barnet changed his style significantly at different points in his career. His earliest works were influenced by expressionism; they were followed by abstract works in the 1950s and 1960s, and finally evolved into more figurative works of silhouetted forms set against geometrically designed backgrounds.\n\nHis work has been exhibited in prominent museums and galleries in the United States and Canada and is included in many prestigious collections, including the Metropolitan Museum of Art and the Guggenheim Museum in New York and the Boston Museum of Fine Arts. The subject property is an example of Barnet's expressionistic early work. A search of auctions and galleries did not turn up any available examples of ``The Bath''. Other aquatints and etchings of this period have sold at auction for between $418 - $3200. At retail, prints of this era have sold for between $750 and $1500 depending on edition. Because ``The Bath'' is currently scarce, the subject property is valued at the high end of this range."),(ItemDescription,""),(ItemCost,"$107"),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Galleries. Mercer Island, WA"),(ItemCondition,"Very good"),(ItemMediumMaterialsTechniques,"Acquatint etching"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Edition of 12"),(ItemSignatureInscriptionsMarkings,"L/R signature, title II ``The Bath''; Edition 12"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 9 3/4\" x   12\"; Impression: 6 1/2\" x 8 3/4\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"The Bath"),(ItemCountryOrNationality,"American"),(ItemArtistDate,"b. 1911"),(ItemArtistOrMaker,"Will Barnet"),(ItemItemDate,""),(ItemDateOrPeriod,"1937, Cole 46"),(ItemTypeOfObject,"Print, Davidson"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"6")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShlbvJW0ffI/AAAAAAAABqc/kv21ucxKpug/s800/DSC_0343.JPG", fileChksum = "2741073cabe851ab5f85e641b3127b12", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 632, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShlbvJW0ffI/AAAAAAAABqc/kv21ucxKpug/s800/DSC_0343.JPG", fileChksum = "2741073cabe851ab5f85e641b3127b12", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 632, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "571790e957786b7b125ee1197cb262be", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 178, imageFileHeight = 141, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "39b9eb87169cf6ce97850c8c4b23e3ae", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2278, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/ShlbvJW0ffI/AAAAAAAABqc/kv21ucxKpug/s800/DSC_0343.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Male Figure, lithograph, 1902, edition of 50 offered for $950 at Davidsons.\nDeux Nus recently sold at Davidson's for $950\nDeux Nus offered for sale at Figaro Gallery for $1800.\nDavidson's recently sold a print for $950"),(ItemCashValue,"$1800"),(ItemMarketAnalysis,"A French painter, illustrator and printmaker, member of the group of artists called the Nabis and afterward a leader of the Intimists; he is generally regarded as one of the greatest colourists of modern art. He also produced lithographs and dry-point. In the 1920s he undertook a series of paintings on one of his most famous themes\8212a nude in a bath, the subject of this lithograph. Les Baigneuses  was probably engraved for Ambrose Vollard who was planing an Album de Nus with various artists. Value characteristics for Bonnard prints include quality of impression, state and condition. A search of auction prices indicate that lithographs of nudes by Bonnard have sold for between $1200 and $2880. At the retail level lithographs of nudes, including 2 examples of Les Baigneuses, have sold and are available at the retail level for between $950 and $1800. Les Baigneuses is currently available the The Figaro Gallery and is the basis for estimating the subject property's Replacement Value. \n\n"),(ItemDescription,"Two nude women are situated in a landscape. One lies on the grass while the other stands in profile next to a small tree or bush. Rendered in scratch"),(ItemCost,"$232"),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Galleries, Mercer Island, WA"),(ItemCondition,"Very Good. Nothing to note."),(ItemMediumMaterialsTechniques,"Drypoint and roulette. Handmade wove paper (Japanese paper)"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Edition of 50"),(ItemSignatureInscriptionsMarkings,""),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 26 cm x 31 cm 10\" x 12\"; Impression:  16 cm x 23.2 cm  6 1/2\" x 23.2 cm. 9 1/4\" \t\t\t\n"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Deux Nus (Les Baigneuses)"),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1867-1947"),(ItemArtistOrMaker,"Pierre Bonnard"),(ItemItemDate,""),(ItemDateOrPeriod,"c. 1927-29"),(ItemTypeOfObject,"Print (Davidson)"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"7, Hanson #18")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/ShlbvtyXvFI/AAAAAAAABqk/cUWgS7DuEH8/s800/DSC_0345.JPG", fileChksum = "86cc8540086f6ff5101b3f370297ae92", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 624, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/ShlbvtyXvFI/AAAAAAAABqk/cUWgS7DuEH8/s800/DSC_0345.JPG", fileChksum = "86cc8540086f6ff5101b3f370297ae92", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 624, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "33c6d8738ebaf0cf5ede0f8b1903896d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 179, imageFileHeight = 140, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "66cd9fdb2fc7a56c64c2fc4f760b85e3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2308, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh3.ggpht.com/_s-XE562xJtM/ShlbvtyXvFI/AAAAAAAABqk/cUWgS7DuEH8/s800/DSC_0345.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,""),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Good: Ares of loss upper right, upper left quadrant and center"),(ItemMediumMaterialsTechniques,"Oil on canvas"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,""),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"9 1/2\" x 13 1/12\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"15 1/8\" x 19 1/4\""),(ItemSupportBaseFrame,"Glazed gilt wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,""),(ItemArtistDate,""),(ItemArtistOrMaker,"Hudson River School"),(ItemItemDate,""),(ItemDateOrPeriod,""),(ItemTypeOfObject,"Painting"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"14, Hanson #43")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SiBTr4P6xmI/AAAAAAAAB4Q/hY6AhRf1thQ/s720/DSC_0374.JPG", fileChksum = "6bf4434d44c3c6ee243eea34bfdf73e3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 479, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SiBTr4P6xmI/AAAAAAAAB4Q/hY6AhRf1thQ/s720/DSC_0374.JPG", fileChksum = "6bf4434d44c3c6ee243eea34bfdf73e3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 479, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "4547a89fcc964f7278edf8f86427dcd3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 194, imageFileHeight = 129, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "1b0078d0ead16f4279a50a75ce3f7f41", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2706, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/SiBTr4P6xmI/AAAAAAAAB4Q/hY6AhRf1thQ/s720/DSC_0374.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Oil on panel works available at auction in the Netherlands and Germany as well as Sotheby's and Christie's, Amsterdam, Bonham's and Butterfields, Leslie Hindman in the USfor between $250 and $6000 (Bonham's, 5/16/07 oil on board 15 3/4\" x 12\" A Fisherman on a dock) with the majority selling in the $1000-$2500 range. \n\nlots 606, 607 oil on panel, portraits'65 Leslie Hindman:  1/18/09, $1464\nGlerum, Amsterdam, boy with goat, 1970 $930\nGlerum, Amsterdam, 6/2/08 $1085\n\n\n\nA painting by the artist is available at Kamp Galleries in Winnetka, IL. \nDear Lydia,\nThanks for your inquiry about our Van d. Berg Still Life.  This is the only painting we have by the artist\nand we believe it to be an early example, between the wars, ca 1925.  The dealer price is $2,900.00\n(Retail $5,900.00).  It's a beautiful example of his work, and a good representation of art from that era.\nYou should buy it.\nRegards,\nNick Vahlkamp, mgr.\n- Show quoted text -\n-- \nKamp Gallery\nDrake Hotel Arcade\n140 East Walton Street\nChicago, IL 60611\n312-664-0090\nfax 312-664-0014\n\n\nHis name is also listed on Covington Fine Arts Gallery in Tuscon, AZ.\n6530 E Tanque Verde Rd Suite 140\nTucson Arizona, 85715 USA\nTel: \t(520) 298-7878\nAlso see M. Klein Auctions in Chicago. \n400 W. Hubbard \nChicago Illinois, 60622 USA\nTel: \t312-948-0080\nFax: \t312-948-0090\nListed on Artnet\n"),(ItemCashValue,"$2,900"),(ItemMarketAnalysis,"Willem van den Berg studied art at the Academie voor Beeldende Kunst in The Hague and traveled extensively to Belgium, France, Italy, Austria and England. \nKnown as  a draughtsman, painter and graphic artist, Van Den Berg's style is characterized by a preference for strong closed contours which reflects his interest in Japanese wood block prints. He was director of the Rijksacademie in Amsterdam from 1938 to 1953. He died in Amsterdam in 1970. His paintings are in the collections of museums in Holland, Europe and the United States. The majority of Van Den Berg's works are available through auctions in Amsterdam and Germany. They have sold for between $732 (7.5\" x 5.25\"), Portrait on panel and $6000. Van Den Berg is particularly known for his oil on panel portraits of fishermen and farmers, and these images have brought the highest prices with majority n the $1000-$2500 range. Paintings by the artist comparable to the subject property's pastoral subject that have come up at auction in recent years have all failed to sell indicating a weak market for this subject. A search for Van Den Berg's works for sale in galleries the United States and Europe turned up a single example - an oil on canvas painting still life of garden flowers which is available for $2900 and is the basis for the value conclusion for the subject property."),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very good: surface crackelure"),(ItemMediumMaterialsTechniques,"oil on panel"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"W. V. D. Berg"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"11 1/4\" x 15 1/4\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"16 3/4\" x 20 3/4\""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,"Holland/Netherlands"),(ItemArtistDate,"1886-1970"),(ItemArtistOrMaker,"Willem van den Berg"),(ItemItemDate,""),(ItemDateOrPeriod,"early 20th century"),(ItemTypeOfObject,"Painting"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"15")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwmvncL_sI/AAAAAAAABww/33e5fwELMLY/s720/DSC03318.JPG", fileChksum = "49dd80467b4515462303f794b29ffe58", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 504, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwmvncL_sI/AAAAAAAABww/33e5fwELMLY/s720/DSC03318.JPG", fileChksum = "49dd80467b4515462303f794b29ffe58", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 504, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "90b383820e81928940f738d65cf7a70f", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 189, imageFileHeight = 132, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "91cfb0af5d0e0f17e6f33b88eb4842d6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2571, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/ShwmvncL_sI/AAAAAAAABww/33e5fwELMLY/s720/DSC03318.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Label on back has biography of artist\nSothebys 2006 1860 \"Garden views with green houses, figures, and gardeners\", 5.3 in x 8.3 in, watercolor over black chalk, $1,566\nthere are additional drawing comps on Artprice, also for Jean-Victor Bertin"),(ItemCashValue,"$1,566?"),(ItemMarketAnalysis,"Charles Nicolas Ransonnette was a painter, engraver and draughtsman active in the mid-19th century. He served as official draughtsman and engraver to the Duchess of Berry, and exhibited regularly at the French Salon from 1822 to 1857. He was a pupil of neo-classical painter Jean-Victor Bertin, and his father, Pierre Nicolas Ransonette, also a painter, engraver, and draughtsman, served the French royal family. \n\nRansonnette worked in a finely detailed and charming style, with many works in small scale. Works by Ransonnette rarely appear in the American market. No current retail offerings could be located and only a few auction sales. The most recent, is for a large scale watercolor over chalk on paper landscape, 5.3 in. x 8.3 in., which sold for $1,566. The subject property is comprised of two ink on paper works framed together. Each work is approximately one half of the comparable. "),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very good"),(ItemMediumMaterialsTechniques,"pen and ink and wash on laid paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"said to be signed on verso under dust jacket"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Mat opening: Upper and Lower: 4\" x 5"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"20 1/2\" x 15 1/2\" "),(ItemSupportBaseFrame,"Gilt and black paint glazed wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1793-1877"),(ItemArtistOrMaker,"Charles-Nicolas Ransonnette"),(ItemItemDate,""),(ItemDateOrPeriod,"1856"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"16, Hanson #38")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5A3amjI/AAAAAAAABx0/IKqO3Zh6LHU/s512/DSC03332.JPG", fileChksum = "e1aca3d54e41127c581be92762f4b8bd", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5A3amjI/AAAAAAAABx0/IKqO3Zh6LHU/s512/DSC03332.JPG", fileChksum = "e1aca3d54e41127c581be92762f4b8bd", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "9e7179bac603aebdd740d2f5d8bd9074", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 129, imageFileHeight = 194, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "cd0cd9d06928104877468460ea075fc6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1199, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh3.ggpht.com/_s-XE562xJtM/Shwm5A3amjI/AAAAAAAABx0/IKqO3Zh6LHU/s512/DSC03332.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Jean-Luc Baroni, Ltd.of London has a drawing for sale. As yet I have not located galleries that sell his prints..\n\nMost recent sale of print, Elefanten, ein Lowe unde ein Eber, 33.2 x 47.7 cm. Bassenge, 04/06/09 $1050\n\nEtchings by are available at Spaightwood Galleries for between $1750- $4500. Most are much smaller than the subject property. "),(ItemCashValue,"$4,500"),(ItemMarketAnalysis,"Known as a Mannerist painter, draughtsman and printmaker, Battista Franco spent the first years of his independent career in Rome, but by 1536 had settled in Florence. Franco is better known, and has usually been more highly regarded, as a draughtsman rather than as a painter. He is also well known for his etching having produced over 100 during his life-time. An assiduous copyist, among his drawings and etchings are numerous studies of antique subjects, usually taken from reliefs on Roman sarcophagi. According to the Illustrated Bartsch, The Deluge was likely after a work designed by Polidoro da Caravaggio.\n\nValue characteristics of prints include state, quality of impression, edition size, subject matter and condition. The subject property is in good condition and is a fine impression. A search for etchings by this artist indicate that the primary  market for this artist is in Europe with most sales taking place at Galerie Bassenge Kunst Und Buchauktionene and Reiss & Sohn in Berlin.  Prints of similar biblical subjects, size, quality of impression, condition and subject matter range from $1401-$2424. The subject property is most comparable in quality of impression to prints at the high end of this range. A search of etchings by the artist in galleries in the United States only turned up a single venue, Spaightwood Galleries in Upton, Massachusetts where a number etchings are available for between $1750-$4500. Most of the etchings available for sale are considerably smaller than the subject property. The etching most closely approximating the subject property in size and quality of impression is available for $4,500 and is the basis for estimating the subject property's replacement value\n"),(ItemDescription,"A biblical scene of the flood."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Good: Some stains and creases. Good impression."),(ItemMediumMaterialsTechniques,"Etching"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signed in the plate lower left, Battista Franco Inu"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 10 1/2\" x 17\" (27 cm x 43 cm)"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"25 1/4\" x 30 1/4\""),(ItemSupportBaseFrame,"Glazed silver gilt wood frame"),(ItemSubjectOrTitle,"The Deluge, Le deluge universel, 3, Bartsch"),(ItemCountryOrNationality,"Italy"),(ItemArtistDate,"1498-1561"),(ItemArtistOrMaker,"Giovanni Battista Franco Inu"),(ItemItemDate,""),(ItemDateOrPeriod,"16th century"),(ItemTypeOfObject,"Print"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"17, Hanson #66")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG", fileChksum = "dac291fe9972381edd30621a4739a12c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG", fileChksum = "dac291fe9972381edd30621a4739a12c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "a699a3975a3ee1fb393e47e1ab5b3ca1", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 194, imageFileHeight = 129, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "592c3b006543a18d7373ab57901c5832", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2707, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"pastel \" Standing Female Nude \" is available, framed and in excellent conditions and as described on our site. The asking price is $ 9.500, but because we are having a special summer promotion, I can give you a discount of 25%, for a net of $ 7.125, including packing, shipping and insurance Pasquale Ianetti Gallery, San Francisco, CA\n\npastel Nude, was offered for $14,000 at Jack Rutberg Gallery, Los Angeles, CA\n"),(ItemCashValue,"$10,000?"),(ItemMarketAnalysis,"Hans Burhardt worked in a variety of media during his long career, including painting, drawing and print making. Known primarily as an abstract impressionist, Burkhardt also worked in a figurative style. Born in Switzerland, Burkhardt came to the United States in ??. He moved to Los Angeles in 1937 and was a well known and well regarded figure in the West Coast art scene. Burkhardt\8217s work has received recent attention due to two major projects, a retrospective of his work mounted at the University Gallery at California State University Northridge, and a comprehensive exhibition of his 1960s paintings presented at the Jack Rutberg Gallery in Los Angeles in 2008-2009. These projects have brought renewed public attention to the artists\8217s work reflected in a greater wolume of sales at auction.\n\nValue characteristics for Burhardt prints include date, size, condition and quality. In the case of his works on paper, a review of auction results in the past five years indicates that earlier works from the 1930s and 1940s appear to be more desirable, reaching higher prices than later drawings from the 1970s and 1980s. The subject property is an earlier work in the abstract style in very good condition. A survey of retail offerings located two pastel works on paper, one offered for $7,125 and the second for $14,000\n"),(ItemDescription,"Abstract composition of three figures rendered in colors on ivory paper"),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good"),(ItemMediumMaterialsTechniques,"graphite and pastel on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signed, dated lower center quadrant"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"26\" x 19 1/2\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"27\" x 20 3/4\""),(ItemSupportBaseFrame,"Red painted and gilt wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,""),(ItemArtistDate,""),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1969"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"18 Hanson #64")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5yzTHVI/AAAAAAAABx4/WErzxTrjaa8/s512/DSC03333.JPG", fileChksum = "a65da9e6609f6590ef0449079fbef5f1", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 400, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5yzTHVI/AAAAAAAABx4/WErzxTrjaa8/s512/DSC03333.JPG", fileChksum = "a65da9e6609f6590ef0449079fbef5f1", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 400, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "f3a8178401afff79c8e85a4dd92155de", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 140, imageFileHeight = 179, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "43029cacdeccdc39f5725592e3fca02f", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1406, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh3.ggpht.com/_s-XE562xJtM/Shwm5yzTHVI/AAAAAAAABx4/WErzxTrjaa8/s512/DSC03333.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes," \8220Flowers and Tears\8221 excellent condition offered for $8,500, The Caldwell Gallery\n\8220Pink Sage\8221 1960, 18 x 25, excellent condition, offered for $34,000, Jack Rutberg Gallery, Los Angeles\nLandscape, 1968, 20 x 24, excellent condition, offered for $?, Jack Rutberg Gallery, Los Angeles\n"),(ItemCashValue,""),(ItemMarketAnalysis,"Hans Burhardt worked in a variety of media during his long career, including painting, drawing and print making. Known primarily as an abstract impressionist, Burkhardt also worked in a figurative style. Born in Switzerland, Burkhardt came to the United States in ??. He moved to Los Angeles in 1937 and was a well known and well regarded figure in the West Coast art scene. Burkhardt\8217s work has received recent attention due to two major projects, a retrospective of his work mounted at the University Gallery at California State University Northridge, and a comprehensive exhibition of his 1960s paintings presented at the Jack Rutberg Gallery in Los Angeles in 2008-2009. These projects have brought renewed public attention to the artists\8217s work reflected in a greater wolume of sales at auction.\n\nValue characteristics for Burhardt prints include date, size, condition and quality. In the case of his works on paper, a review of auction results in the past five years indicates that earlier works from the 1930s and 1940s appear to be more desirable, reaching higher prices than later drawings from the 1970s and 1980s. The subject property is an earlier work in the abstract style in very good condition. A survey of retail offerings located \n"),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good. No evidence of damage"),(ItemMediumMaterialsTechniques,"oil on canvas"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signature, date, lower left quadrant"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"15 1/2\" x 19 1/4\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"23 1/2\" x 27 1/2\""),(ItemSupportBaseFrame,"Gilt wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,""),(ItemArtistDate,""),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1975"),(ItemTypeOfObject,"Painting"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"19 Hanson #63")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SnoTIWHxquI/AAAAAAAABzk/D_Os7LpS43U/IMG00021_2.jpg", fileChksum = "61f9a7508ceddcbb780fe757b3cc48bf", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 622, imageFileHeight = 511, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SnoTIWHxquI/AAAAAAAABzk/D_Os7LpS43U/IMG00021_2.jpg", fileChksum = "61f9a7508ceddcbb780fe757b3cc48bf", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 622, imageFileHeight = 511, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "c13a71aec66c814dd585fc6e417fad20", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 174, imageFileHeight = 143, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "ed1f081afe85b605eace3b171c14252c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2191, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_EV171eW_Cmg/SnoTIWHxquI/AAAAAAAABzk/D_Os7LpS43U/IMG00021_2.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"pastel \" Standing Female Nude \" is available, framed and in excellent conditions and as described on our site. The asking price is $ 9.500, but because we are having a special summer promotion, I can give you a discount of 25%, for a net of $ 7.125, including packing, shipping and insurance Pasquale Ianetti Gallery, San Francisco\n\npastel Nude, was offered for $14,000\n"),(ItemCashValue,""),(ItemMarketAnalysis,"Hans Burhardt worked in a variety of media during his long career, including painting, drawing and print making. Known primarily as an abstract impressionist, Burkhardt also worked in a figurative style. Born in Switzerland, Burkhardt came to the United States in ??. He moved to Los Angeles in 1937 and was a well known and well regarded figure in the West Coast art scene. Burkhardt\8217s work has received recent attention due to two major projects, a retrospective of his work mounted at the University Gallery at California State University Northridge, and a comprehensive exhibition of his 1960s paintings presented at the Jack Rutberg Gallery in Los Angeles in 2008-2009. These projects have brought renewed public attention to the artists\8217s work reflected in a greater wolume of sales at auction.\n\nValue characteristics for Burhardt prints include date, size, condition and quality. In the case of his works on paper, a review of auction results in the past five years indicates that earlier works from the 1930s and 1940s appear to be more desirable, reaching higher prices than later drawings from the 1970s and 1980s. The subject property is an earlier work in the abstract style in very good condition. A survey of retail offerings located .. \n"),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good"),(ItemMediumMaterialsTechniques,"Pastel and graphite on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signature, date, lower right quadrant"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"23 1/2\" x 17 1/2\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"25 1/2\" x 19 1/2\""),(ItemSupportBaseFrame,"Glazed silver gilt frame"),(ItemSubjectOrTitle,"Untitled"),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1976"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"20, Hanson #62")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/ShwnA8lJxmI/AAAAAAAAByg/3D8KSo25_5Q/s512/DSC03343.JPG", fileChksum = "eb55c9e8f2b87448ad3fe1b7a5aaf54e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/ShwnA8lJxmI/AAAAAAAAByg/3D8KSo25_5Q/s512/DSC03343.JPG", fileChksum = "eb55c9e8f2b87448ad3fe1b7a5aaf54e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "fdd47b358ebab5ec1b0b0a51cbe41fbc", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 129, imageFileHeight = 194, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "940e99372ee848707da068ff9a2e03a9", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1199, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_s-XE562xJtM/ShwnA8lJxmI/AAAAAAAAByg/3D8KSo25_5Q/s512/DSC03343.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"pastel \" Standing Female Nude \" is available, framed and in excellent conditions and as described on our site. The asking price is $ 9.500, but because we are having a special summer promotion, I can give you a discount of 25%, for a net of $ 7.125, including packing, shipping and insurance Pasquale Ianetti Gallery, San Francisco\n\npastel Nude, was offered for $14,000\n"),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good: Some creasing along the top edge and bottom left corner"),(ItemMediumMaterialsTechniques,"Charcoal on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signature, date L/R"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 17 1/4\" x 23\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"27 1/4\" x 32 3/4\""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Untitled"),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1941"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"21 Hanson #61")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwnC1KiGAI/AAAAAAAABys/EtMJnBzM5AU/s576/DSC03346.JPG", fileChksum = "10d359b37f2a0b3c109762e7f7cd7de6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 576, imageFileHeight = 469, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwnC1KiGAI/AAAAAAAABys/EtMJnBzM5AU/s576/DSC03346.JPG", fileChksum = "10d359b37f2a0b3c109762e7f7cd7de6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 576, imageFileHeight = 469, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "6365c7b50dbfa8a39588c266a3123f1d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 175, imageFileHeight = 143, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "3d50f7e328332edf554de26815f2b282", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2211, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/ShwnC1KiGAI/AAAAAAAABys/EtMJnBzM5AU/s576/DSC03346.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Paid $150 to Cal State Northridge\nOne was offered by Delorme-Collin du Bocage in Paris on 1 July 2009. It was estimated at 2,500-3,000 euros but failed to sell. The second example was offered by Christie's New York on 24 November 2008. It was estimated at $4,000-$5,000, and sold at $3,000.\n\nOne was offered by Delorme-Collin du Bocage in Paris on 1 July 2009. It was estimated at 2,500-3,000 euros but failed to sell. The second example was offered by Christie's New York on 24 November 2008. It was estimated at $4,000-$5,000, and sold at $3,000.\n\n"),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,"Portrait of Native American activist and celebrity Russell Means depicted in a bust view, wearing braids and a traditional bone choker. The background features shades of blue that contrast with the brown skinned figure and the peach colored shirt he is wearing."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,""),(ItemMediumMaterialsTechniques,""),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signed vertically, Andy Warhol"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"49\" x 27\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"61\" x 38 1/2\""),(ItemSupportBaseFrame,"Stained wood frame, glazed"),(ItemSubjectOrTitle,"\"Russell Means\" from the \"American Indian\" series"),(ItemCountryOrNationality,"American"),(ItemArtistDate,""),(ItemArtistOrMaker,"Andy Warhol"),(ItemItemDate,""),(ItemDateOrPeriod,"1972"),(ItemTypeOfObject,"Poster"),(ItemLocation,"Stairwell"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"22 Hanson #65")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/ShwnGFc7K9I/AAAAAAAABzA/0hx7n8lKdoI/s800/DSC03351.JPG", fileChksum = "a5c3128491db2ae76606a856354db49d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 534, imageFileHeight = 800, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/ShwnGFc7K9I/AAAAAAAABzA/0hx7n8lKdoI/s800/DSC03351.JPG", fileChksum = "a5c3128491db2ae76606a856354db49d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 534, imageFileHeight = 800, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "437579f98bb162682cd04bb705fac5be", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 129, imageFileHeight = 194, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "df596a1cec9291a0afd8a0e3cfed0fcb", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1202, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh4.ggpht.com/_s-XE562xJtM/ShwnGFc7K9I/AAAAAAAABzA/0hx7n8lKdoI/s800/DSC03351.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Painted screens available at Ark Antiques are and Edo period, 6 panel screen, 4'6\" high in very good condition and very good quality offered for $2600.\nA small screen, approx. 3'5\" high, 6 panels,  Japanese, 19th century of birds, duck, chickens and foliage available for $1800.\n\n8 panel painted Chinese screen sold at Millea Brothers for $2000 + BP 10/29/06\n\n"),(ItemCashValue,"$2000"),(ItemMarketAnalysis,"This screen was produced for export to the Western markets.  Vintage screens from China are available from importers and at estate sales and regional auction houses. Value characteristics include age, quality of painting, size and condition. The subject property was purchased at an art sale at California State Northridge, therefore the appropriate market for this piece is high-end consignment stores and regional auction houses. Vintage screens comparable to the subject property in size, age and quality were available at local consignment stores and at regional auction houses throughout the country for between $575 and $3,500. A pair of four panel painted Chinese screens depicting boaters on river amid mountains sold at an auction house in New Jersey, Millea Brothers, for $2000 is the basis for the subject property's replacement value.\n "),(ItemDescription,"A six-panel painted screen of ink and colors on paper with a lacquer frame depicting a landscape of misty mountains, tiny villages and boaters along a river."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"California State University at Northridge"),(ItemCondition,"Good: First panel: buckling, a hole in upper quarter; second panel, some slight buckling; third panel: one small buckle in center; fourth panel: horizontal buckling with creases and some loss of pigment; fifth panel: some buckling in center and lower half; sixth panel: slight abrasions towards left center border; gilt paint border: abrasions along areas of borders. "),(ItemMediumMaterialsTechniques,"ink and colors on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"L/L of leftmost panel, signature and title\nSame 2 seals on each panel, alternates lower right and lower left."),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"53 1/2\" x 20\" each of 6 panels 120\" wide overall"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"68\" x 24 1/2\" (each panel)"),(ItemSupportBaseFrame,"Lacquer frame and silk backing"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,"China"),(ItemArtistDate,""),(ItemArtistOrMaker,""),(ItemItemDate,""),(ItemDateOrPeriod,"20th century"),(ItemTypeOfObject,"Chinese screen"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"24")], images = []}},ReportParagraph {elemText = ""},ReportItem {elemItem = Item {itemName = "", fields = [], images = []}},ReportUndecided], reportGlossary = [("Aquatint","The whole plate is covered with grains of resin called an aquatint ground, allowing acid to bite into the entire area, creating an overall grainy, tonal effect. This technique is often combined with etching."),("Drypoint","As in an engraving, the drypoint needle draws the image directly onto the plate. The residue copper is left on the side of the etched lines, which then collect the ink, creating a furry effect called burr."),("Etching","The plate is covered in an acid-resistant layer of wax called an etching ground. The image is then drawn into this surface with an etching needle. When covered with printing ink the lines hold the ink whilst the rest of the plate repels it."),("Lithograph","From the Greek lithos, stone and graphe, writing. This printing process is unlike both intaglio and relief processes, both of which involve cutting into the plate. Lithography relies on the principle that grease and water will repel each other. The image is drawn in a greasy substance onto a lithographic stone. The stone is then dampened with water and the greasy printing ink adheres only to the drawing."),("Second edition","Often an artist will work on a composition to a certain point, and stop to print an impression of it. This single stage in the evolution of this image is called a state. Each time the composition is changed a new state of the print is created. These changes can range from the addition of a plate signature to drastic alterations in the composition. Today artist\8217s will frequently choose to edition a state before moving on with the composition. "),("State","Often an artist will work on a composition to a certain point, and stop to print an impression of it. This single stage in the evolution of this image is called a state. Each time the composition is changed a new state of the print is created. These changes can range from the addition of a plate signature to drastic alterations in the composition. Today artist\8217s will frequently choose to edition a state before moving on with the composition. ")], reportSources = [("Experts Consulted","Jennifer M. Wood\nConsulting and Appraisal Services\n3122 Arizona Avenue, Suite C\nSanta Monica, CA  90404\nPhone: 310 633 1736\nEmail: jenniferwood@roadrunner.com\n"),("Internet Sources","Artfact.com\nhttp://www.artfact.com\n\nArtnet.com\nhttp://www.artnet.com\n\nInternational Fine Print Dealers Association\nhttp://www.printdealers.com\n\nTrocadero.com\nhttp://www.trocadero.com"),("Print Sources","Bartsch, Adam, Zerner, Henri ed., <em> The Illustrated Bartsch: Italian Artists of the Sixteenth Century School of Fountainbleau<\\em> New York: Abaris Books, 1979.\n\nBouvet, Francis <em>Bonnard: The Complete Graphic Work </em> New York: Rizzoli, 1981. \n\nDelteil, Loys, <em> Le Peintre-Graveur Illustre (XIX et XX Siecles) Tome Cinquieme, Corot <\\em> republication of the 1910 Paris edition, New York: Da Capo Press, 1968.\n\nIves, Colta, Giambruni, Newman Sasha M. <em> Pierre Bonnard: The Graphic Art<\\em> New York: The Metropolitan Museum of Art, 1989\n\nLeymarie, Jean, Melot, Michel, <em> The Graphic Works of the Impressionists: Manet, Pissarro, Renoir, Cezanne, Sisley<\\em> New York: Harry N. Abrams, Inc., 1971.\n\nSoucy, Patricia and Janella N. Smyth, eds. <em>The Appraisal of Personal Property: Principles, Theories and Practice Methods for the Professional Appraiser.</em> Washington, D.C.: American Society of Appraisers, 1994.\n\nStella, Joseph G., <em> The Graphic Works of Renoir </em> London: Lund Humphries, 1975.\n\n<em>Uniform Standards of Professional Appraisal Practice, the 2008-2009 Edition</em>(USPAP)(Washington, D.C.: The Appraisal Foundation, 2008)."),("Galleries and Auction Houses","Allinson Gallery, Inc.\n3 Dunham Pond Road East \nStorrs, CT 06268-2022 \n860-429-2322\nhttp://www.allinsongallery.com\n\nBonham's & Butterfield's\n220 San Bruno Avenue\nSan Francisco, CA 94203\n415-86-7500\nhttp://www.bonhams.com\n\nChristie's\n20 Rockefeller Plaza\nNew York, NY 10020\n212-636-2000\nhttp://www.christies.com\n\nDavidson Galleries\n313 Occidental Ave S.\nSeattle, WA 98104\n206 624.6700\nhttp://www.davidsongalleries.com\n\nFigaro Gallery\n703-491-1000\nwww.justfineart.com\nE-mail: inquire@figarogallery.com\n\nGalerie Bassenge Kunst-und Buchauktionen\nErdener Strabe 5a\n14193 Berlin-Grunewald\nGerman\n+49-(0)30-893 80 29-0\n\nI. M. Chait Gallery\n9330 Civic Center Drive\nBeverly Hills, CA 90210\n(800) 775-5020\nhttp://www.chait.com\n\nKamp Galleries\n563 Lincoln Avenue\nWinnetka, IL 60093\n847-441-7999\nhttp://www.kampgallery.com\n\nLeslie Hindman Auctioneers\n\nMasterworks Fine Art Inc.\n13470 Campus Drive\nOakland, CA 94619\nhttp://www.masterworksfineart.com\n510-777-9970\nSpaightwood Galleries")], reportLetterOfTransmittal = "According to our contract signed by you on [ContractDate], [we] have\nprepared the following [ReportType] report for the property\nrepresented as belonging to you.  The subject properties consists of:\n[BriefDescription].\n\nYou provided [us] with the information that indicates that you own the\nsubject [properties] listed above in their entirety.  [We] examined the\n[properties] in the presence of [Client] at [InspectionAddress]\non [InspectionDate].\n\nThis appraisal report provides the appraisal conclusion of\n[ValueType] for the intended use of [IntendedUse].  \n[ValueTypeDescription]\n\nThis report will function as documented evidence of [theseappraisers]\nhaving witnessed the subject [properties], and identified, described and\nvalued the subject [properties] in the context of estimating [ValueType].\n\n\nAfter thorough consideration of available facts and statistics, on\n[EffectiveDate], the [ValueType] for the subject [properties] is\n<b>[TotalValueAsText] ([TotalValue])</b>.\nThe effective date of value for this appraisal is [EffectiveDate].\nThis figure does not reflect a consideration for related sales\ntaxes and delivery/installation charges, which can vary.  This\nprofessional opinion is based upon [our] knowledge and experience in the\nvaluation of fine arts.  This appraisal is based on [our] research and\nanalysis of the [properties] and knowledge of relevant markets where the\n[properties] [are] traded.\n\nThis is a [ReportType] report that was prepared to conform to\nthe Code of Ethics of the American Society of Appraisers and the\ncurrent version of the [USPAP],\npublished by the Appraisal Foundation.  We have\nno past, present or contemplated future interest in the [properties]\nincluded in this report.  Our fee was not contingent on any value\nfinding.\n\nThis [ReportLength] page report consists of a letter of transmittal, scope of\nwork, definitions, privacy statement, assumptions and limiting\nconditions, certification, description, photographs, narrative,\nglossary, bibliography and credentials.  All these elements must be\npresent for the value conclusion to be valid.\n\nThe information in this report is confidential.  All notes,\nelectronic, work copies and photographs remain the property of the\nappraiser and are retained for five years, and/or two years after the\nconclusion of any litigation.  All documents are shredded or otherwise\ndefaced prior to disposal.\n\nThe submission of this report completes [our] assignment and\nobligations.  Any further consultation, preparation of additional\nreports, testimony, or appearances in court or before any governmental\nforum in reference to the [properties] described in this appraisal will\nrequire additional fees to cover the required services.  The\naforementioned fees will be billed at the rate commensurate with the\ntiming of the appraisal.\n\nThank you for allowing [us] to be of service to you.\n", reportScopeOfWork = "Thompson & Martinez Fine Art Appraisals, Inc. was hired by [Client] to reach an opinion of the [ValueType] of [BriefDescription], for the intended use of [IntendedUse].  Based on [our] discussion with [Client], as well as the intended use of the report, the type and location of the subject [properties] and the effective date of valuation, the [appraisers] [have] selected the type of value, the relevant market, the information researched and the analysis to be applied.  To complete this assignment [we] consulted authoritative publications and on-line sources, reviewed published and on-line market data and interviewed knowledgeable dealers and others regarding the artists, their work, and the relevant markets. Kathi Jablonsky ISA, CAPP assisted in the research of the display cabinet and chairs. Jo Ellen Cole of Cole Appraisal Services consulted on the identification of the jade and bowenite sculptures.\n\nThis report is intended to be an objective, independent, unbiased and credible analysis of known facts.  During the course of our examination, research and analysis, [we] have endeavored to be accurate and to use due diligence.  This opinion of value was developed using appraisal methods that are within the accepted standard of care for the appraisal profession and within the boundaries of typical practice. \n\nThis appraisal document summarizes the majority of information analyzed, the appraisal procedures followed, and the reasoning that supports the analysis, opinions and conclusions under the requirements of the [USPAP].  It is considered a summary report.  The documentation substantiating the value reported is held in the appraiser's file. ", reportCertification = ["To the best of [our] knowledge and belief, the statements of fact continued in this report are true and correct.  The value conclusions are based on [our] background and experience outlined in [our] credentials attached at the end of this report.","The reported analyses, opinions and conclusions are limited only by the reported assumptions and liming conditions.  They are [our] personal, impartial and unbiased professional analyses, opinions and conclusions.","[Our] analyses, opinions and conclusions in this report were developed and prepared in conformity with the [USPAP]. ","Unless otherwise stated, the value given in the Appraisal Report represents the opinion of value as of the date of inspection.","[We] have no bias with respect to the property that is the subject of this report, or to the parties involved with this assignment.","[Our] compensation for completing this assignment is not contingent upon development or reporting of a predetermined value or direction in value that favors the cause of the client, the amount of the value opinion, the attainment of a stipulated result, or the occurrence of a subsequent event directly related to the intended use of this appraisal.","[We] have made a personal inspection of the property that is the subject of this report.","[These] [appraisers] prepared most conclusions and opinions set forth in this Appraisal Report. Kathi Jablonsky, ISA, CAPP assisted in the research of the display cabinet and pair of chairs.","Although all due care has been taken to ensure that any statement as to the authorship, attribution, origin, date, age, provenance, condition, and value of any item of personal property included in this report is accurate, all such statements are to the best of the knowledge of the [appraisers] who completed this report, and are not to be relied upon as representation of fact."], reportLimitingConditions = ["Title to the property and present ownership of the property is assumed to be good and marketable unless otherwise stated.  Further, the appraisal is made with the understanding that the present ownership of the subject property includes all rights that may be lawfully owned and is not subject to any encumbrances or indebtedness.  The Clients stated that they own the property; ownership was not researched or confirmed.","The [appraisers] [have] no present or prospective interest in the property that is the subject of this report, and no specified personal interest with respect to the parties involved.  The reported analyses, opinions and conclusions are limited only by the reported assumptions and limiting conditions, and are personal, impartial and unbiased professional analyses, opinions and conclusions.","The [appraisers'] analyses, opinions, and conclusions [were] developed, and this report has been prepared, in conformity with the Code of Ethics as set forth by the American Society of Appraisers and the [USPAP] of the Appraisal Foundation, Washington, D.C.","The Appraisal is based on visual inspection of the appraised item.  No tests of any kind have been administered unless so stated.  Measurements and photographs are made under ``field conditions,'' and therefore will be as accurate as possible.  Measurements are listed in the order of height, width and depth, unless otherwise indicated.","The opinion of value expressed in this Appraisal Report is the result of and is subject to the data and research described in detail in the Appraisal Report.  In the preparation of this report, the [appraisers] [have] consulted a variety of sources that are documented either in the report or are available in the [appraisers'] files.  These sources are assumed to be reliable and accurate; however, the [appraisers] [take] no responsibility for errors or omissions contained in these sources.","This Appraisal Report, or a copy of this Appraisal Report, may be transmitted to a third party or legal entity only in its entirety and only with the expressed written consent of the appraisal client.","Disclosure of the contents of the Appraisal Report is governed by the Bylaws and Regulations of the Appraisal Foundation [USPAPShort].  Neither all nor any part of the contents of this report (especially any conclusions as to the value or the identity of the Appraisers) shall be disseminated to the public through advertising, public relations, news, sales, or the media without the prior written consent and approval of the appraiser.","Information furnished by the owner regarding property history is believed to be reliable; however, no warranty is given for its accuracy.  ","The [appraisers] assumes no responsibility for unforeseen changes in market conditions, nor is the inability to locate a potential buyer or replacement item, at the appraised value, if such is attempted by the client or any other person. The value conclusion is not an offer to buy or sell the property.","In the event of a dispute involving interpretation or application of this report, the American Society of Appraisers recommends that such a dispute shall be referred to a neutral third-party mediation service.  The cost of such mediation shall be borne equally by both parties.  In the event mediation is not successful, the parties should agree to submit the dispute to binding arbitration under the laws of the State of California.","The [appraisers] [have] personally inspected the subject property.  The condition of the property is described sufficiently for its type, with any visible and/or relevant repairs noted. The condition of the property may change after inspection and the effective date of valuation.","The opinion of value expressed in the Appraisal Report is the result of and is subject to the data and research described in detail in the Appraisal Report and on data retained in the [appraisers'] files.  The [appraisers] [have] not knowingly withheld any significant information.","Appraisers witness, identify and value subject property but do not claim authentication of the subject properties, unless otherwise noted.","This Appraisal report is valid only for the intended use, intended users and for the type of value stated in the appraisal.  The values reported in this appraisal are valid only on the effective date of the appraisal [EffectiveDate].","The submission of this report completes [our] assignment and obligations.  Any further consultation, preparation of additional reports, testimony, or appearances in court or before any governmental forum in reference to the property described in this appraisal will require additional fees to cover the required services."], reportPrivacyPolicy = "Thompson & Martinez Fine Art Appraisals, Inc. makes significant efforts to keep all appraisal-related data confidential in compliance with the current version of the [USPAP] as published by The Appraisal Foundation. We will not disclose confidential appraisal data to any third party without the written permission of the rightful owner of that data. We will not release a copy of an appraisal you have contracted through Thompson & Martinez Fine Art Appraisals, Inc. to any third party without the written permission of the client.\n\nOur policy specifies the protection of client information in the following ways:\n * We limit access to all information only to those who have a business or professional reason to know and only to nonaffiliated parties as required by law.\n * We maintain a secure office and computer environment to ensure that your information is not placed at unreasonable risk.\n * The categories of non-public personal information that we collect from a client depend upon the scope of the client's appraisal engagement.  This includes information about personal property assets, information about tax identification numbers, and other non-public information necessary.  Additionally, it may include information about transactions between clients and third parties and information form third party sources.\n * Unaffiliated third parties that require access to our clients' personal information, such as Federal and State tax regulators, insurance companies, and the like may only review company records as permitted under law.\n * We do not provide client information to mailing list vendors or solicitors for any reason.\n * Personally identifiable information will be maintained during the time a client is a client of the company and for 5 years after the appraisal has been completed and/or 2 years after litigation as well as the ethical considerations as specified by [USPAPShort].  After this required period of record retention, all such information will be destroyed.\n", reportMessages = "", reportPerms = Just (Permissions {owner = Username {unUsername = "lydiat"}, writers = [Username {unUsername = "Natasha"},Username {unUsername = "david"},Username {unUsername = "lydia"}], readers = []}), revisionInfo = defaultValue}
r = Report {reportFolder = "", reportName = "Laurie Hanson", reportDate = "August 28th, 2009", reportContractDate = "May 6th, 2009", reportInspectionDate = "May 28th, 2009", reportEffectiveDate = "May 28th, 2009", reportAuthors = [Author {authorName = "Lydia Thompson, Ph.D.", authorCredentials = "====Education\n * Ph.D., New York University (NYU), Institute of Fine Arts (IFA), New York, NY, 1998.  Art and Archaeology.  Specialized China\n * M.A., NYU, Institute of Fine Arts (IFA), New York, NY, 1989.  Art and Archaeology\n * B.A., Middlebury College, Middlebury, VT, 1983.  East Asian Studies\n====Professional Study\n * University of California, Irvine, Certificate Program in Appraisal Studies, UCI in partnership with American Society of Appraisers (ASA).  Completion of all P.O.V. courses in certificate program\n * Shandong University, Jinan, Shandong, PRC.  Department of Archaeology, 1993-94.\n====Awards\n * Kress Foundation, Dissertation Fellowship, 1995-96.  Center for Advanced Study in the Visual Arts (CASVA), National Gallery of Art, Ittleson Fellow, 1993-95. Samuel F.B. Morse Fellowship, 1991-92\n====Memberships\n * Candidate Member, American Society of Appraisers\n====Selected List of Professional Experience 2000-2008\n * President, Thompson and Martinez Fine Art Appraisals, Inc., La Jolla, CA\n * Curator and Docent Educator for the exhibition: ``ymbols of Power:  Masterpieces from the Nanjing Museum'', the Bowers Museum of Cultural Art, Santa Ana, CA\n * Adjunct Faculty in Art History, UC Los Angeles and UC San Diego, CA.  \n====Selected Publications\n * ``Ritual, Art and Agency:  Consecrating the Burial Ground in the Han Period'' <em>Recarving China's Past:  Art, Archaeology, and Architecture of the Wu Family Shrines</em>.  Proceedings of a symposium held at the Princeton University Art Museum, (Princeton U.P.:  forthcoming)\n * Entries on artists and translation of essay by Art Critic Li Xianting, Betti-Sue Hertz ed. <em>Past in Reverse:  Contemporary Art in East Asia</em> (San Diego:  San Diego Museum of Art, 2004)\n * Co-author with Xu Huping, <em>Symbols of Power:  Masterpieces from the Nanjing Museum</em>, (Santa Ana:  The Bowers Museum of Cultural Art, 2002)\n"},Author {authorName = "Natasha Bonilla Martinez, MA", authorCredentials = "====Education\n * Certificate, University of California, Irvine Extension, Certificate Program in Appraisal Studies, UCI in partnership with American Society of Appraisers (ASA), 2007\n * Certificate, USPAP, University of California, Irvine Extension, Certificate Program in Appraisal Studies, UCI in partnership with American Society of Appraisers (ASA), 2006.\n * M.A. San Diego State University, Latin American Studies, San Diego, CA. 2001 Specialization in Latin American and Latino photography and art.\n * Graduate study. New York University, Department of Anthropology, New York, NY. 1982 to 1986, photographic anthropology and Native American art and culture.\n * B.A.  Vassar College, Poughkeepsie, NY, 1981. Independent Major.\n====Awards\n * Gateways Rockefeller Fellowship, 1997, Guadalupe Cultural Arts Center, San Antonio, TX, for the study of contemporary photography by Mexican and Mexican American women.\n====Selected Professional Experience 2000-2008\n * Vice-President, Thompson and Martinez Fine Art Appraisals, Inc., La Jolla, CA.\n * Director of Education, Director of Visual Arts and Education and Director of Program Development, successively, California Center for the Arts, Escondido, Escondido CA. \n * Co-Curator of Niki de Saint Phalle: California Dreaming, California Center for the Arts, Escondido, Escondido CA.\n * Co-Curator, Spirit Capture: Photography and Native America exhibition, National Museum of the American Indian, Smithsonian Institution, New York, NY.\n * Adjunct faculty, Department of Art, University of San Diego, San Diego, CA. Course on US/Mexico border art.\n====Selected Publications\n * ``An Indian Americas'' and ``Camera Shots'' in <em>Spirit Capture: Photographs from the National Museum of the American Indian</em>. (Washington, D.C.: Smithsonian Press, 1998).\n====Memberships\n * Candidate member, American Society of Appraisers\n * American Association of Museums \n * Association for Latin American Art, College Art Association\n"}], reportPreparer = "Thompson & Martinez Fine Art Appraisals, Inc.", reportPreparerEIN = "(preparer EIN)", reportPreparerAddress = "(preparer address)", reportPreparerEMail = "(preparer Email)", reportPreparerWebsite = "(preparer Website)", reportAbbrevs = [("[EffectiveDate]","May 28th, 2009"),("agree","agree"),("appraisers","Appraisers"),("appraisers'","Appraisers'"),("are","are"),("assume","assume"),("certify","certify"),("Client","Ms. Laurie Hanson"),("EIN","20-8308057"),("have","have"),("InspectionAddress","3618 West Estates Lane, #D\nPalos Verdes Estates, CA 90274"),("intendeduse","Insurance purposes"),("LOTInspectedBy","Lydia Thompson, Ph.D., Natasha Martinez, M.A."),("Our","Our"),("our","our"),("Preparer","Thompson & Martinez Fine Art Appraisals, Inc.\nEIN: 20-8308057\n"),("preparer Email","appraiser@thompsonandmartinez.com"),("preparer Website","http://www.thompsonandmartinez.com"),("properties","properties"),("ReportType","summary"),("ReportUsers","Ms. Laurie Hanson and her insurance company"),("soucysmith","(from Soucy and Smith, eds, <em>The Appraisal of Personal Property - Principles, Theories, and \nPractice Methods for the Professional Appraiser\n</em>, 1994.)"),("take","take"),("These","These"),("these","these"),("theseappraisers","theseappraisers"),("Us","Us"),("us","us"),("USPAP","<em>2008-2009 Uniform Standards of Professional Practice (USPAP)</em>"),("USPAPShort","USPAP 2008-2009"),("ValueTypeLongDescription","Replacement Value is the price required to replace a property with another of similar age, quality, origin, appearance and condition, within a reasonable length of time in an appropriate and relevant market (from Soucy and Smith, eds, <em>The Appraisal of Personal Property - Principles, Theories, and \nPractice Methods for the Professional Appraiser\n</em>, 1994.) \n\nBecause of the time element involved, the appropriate market for replacement is deemed to be that of primary sources, such as art galleries carrying comparable and immediately available personal properties. If the only source for replacement can be demonstrated to be that of auction, that is the source utilized. Replacement Value is used as the basis for insurance purposes and damage/loss reports. "),("We","We"),("we","we"),("were","were")], reportTitle = "", reportIntendedUse = Just Insurance, reportValueType = Just ReplacementValue, reportValueApproach = Just MarketComparisonApproach, reportClient = "3618 West Estates Lane, #D\nPalos Verdes Estates, CA 90274", reportClientGreeting = "Ms. Hanson", reportItemsOwnerFull = "", reportItemsOwner = "Laurie Hanson", reportBriefItems = "prints by Jean Baptiste Camille Corot, Auguste Renoir, Henri Toulouse-Lautrec, Francisco de Goya, Pierre Bonnard, Will Barnet, Giovanni Battista Franco, Charles-Nicolas Ransonnette, 1 painting by Willem van den Berg, 1 painting by a member of the Hudson River School, 2 prints, 3 works on paper and 1 painting by Hans Burkhardt, 1 signed poster by Andy Warhol, 1 painted Chinese screen", reportInspectionLocation = "3618 West Estates Lane, #D\nPalos Verdes Estates, CA 90274", reportBody = [ReportParagraph {elemText = "Since the inception of the financial crisis and economic downturn in the fall of 2008, the art market has dropped precipitously. Well-known auction houses have seen their sales volume drop 70% or more since their highs in the spring of 2008 and, in several categories of art, prices have returned to 2002-2003 levels (5/21/09 Editorial by Swann President Nicholas Lowry). \n\nSome categories of art, however, have fared better in the downturn than others. Contemporary art, which realized spectacular price increases in the past 5 years, has been the hardest hit with average prices of high-end contemporary art dropping two-thirds in the evening sales. On the other hand, average prices of antiquities and Old Masters have softened but are still double what they were in 2003 (Art & Auction, June 2009, databank, p. 204). Sales of the impressionists at spring auctions showed surprising strength generally exceeding expectations (Art & Auction, July/August, 2009) Auction Reviews, p. 72.\n\nMs. Hanson's collection consists of prints by impressionists and old Masters as well as prints and paintings by well-known 20th century artists. Also included is a decorative Chinese painted screen.\n\nThere is a limited market for antique European prints in southern California, with only a few galleries on the West Coast, with one of the largest the Davidson Gallery in Seattle. The majority of galleries carrying this type of material is based in the East Coast of the United States and Europe. Although the customary practice is to seek comparable sales in the market where the insured shops, the unavailability of Old Master prints in the region required the Appraisers to seek comparable sales elsewhere in the national market. Given the quality of the works comparables were sought among dealers of high quality prints and members of the International Fine Art Print Dealer's Association, as well as in the auction market as appropriate.\n\nA significant element of the Hanson collection are several works by Hans Burkhardt (1904-1994), the Swiss American abstract impressionist, including a painting, works on paper and prints. A well known figure in Southern California, Burkhardt is especially known for his large scale anti-war paintings that expresses his response to the Spanish Civil War, World War II, Vietnam and Desert Storm. Recent major installations by the Jack Rutberg Gallery in Los Angeles and the California State University Northridge gallery have brought renewed attention to Burkhardt's work, which appears to be reflected in interest and activity at recent auction offerings at Sotheby's and Christie's.\n\n"},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Venator & Hanstein KG (Butcher Graphik Autographen Auuktion 100) 3/23/07 #548 estimated $1330 (EU 1000) bought in\n\nSame print currently available for $950 at the Davidson Gallery.\n\nAuction records of print published in Gazette de Beux-Arts indicate this print sell for between $431, $447, $632.\n\nDelteil 1/iv\n\nC&J Goodfriend offers Souvenir de Toscane 2nd state print, dated 1845 for $2500\n\nMasterworks Fine Art Inc. has recently sold a 4th state Souvenir de Toscane but don't know price.\nRecently sold (4th state, A. Salmon printer, Gazette des Beaux-Arts at Masterworks Fine Art Inc. \n"),(ItemCashValue,"$950"),(ItemMarketAnalysis,"Jean-Baptiste-Camille Corot was the leading painter of the Barbizon school of France in the mid-nineteenth century. He is a pivotal figure in landscape painting and his vast output simultaneously references the Neo-Classical tradition and anticipates the plein-air innovations of Impressionism. He was also an engraver. ``Souvenir de Toscane'' was originally executed in 1845 and produced in four states.\n\nValue characteristics include: state, edition, subject matter and condition. State is a significant characteristic of value for this print with impressions published before the Gazette des Beaux-Arts edition (state 4) being especially scarce. A second state version of Souvenir de Toscane is available at C&J. Goodfriend Galleries for $2500. The subject property is an example of the 4th state published in Gazette des Beaux-Arts in 1875 whA second state version of Souvenir de Toscane is available at C&J. Goodfriend Galleries for $2500. between $250 and $630.  A 4th state print in similar condition and impression to the subject property is currently offered for sale at the Davidson Galleries for $950 and is the basis for estimating its replacement value."),(ItemDescription,"Landscape of Tuscan countryside with rolling hills and trees in the foreground framing a domed structure in the center middle-ground."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Gallery"),(ItemCondition,"Very Good: dark spot (mold?) lower margin center. Paper appears toned, beige, almost sepia"),(ItemMediumMaterialsTechniques,"Etching on laid paper "),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Edition of 750, 4th state, Delteil 1, iv., Melot 1,v."),(ItemSignatureInscriptionsMarkings,"Bottom left: Corot Del. & Next line Gazette des Beaux-Arts bottom right: Imp. A. Salmon, Paris."),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Plate impression: 5\" x 7\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"11 3/4\" x 14 1/2\""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Souvenir de Toscane"),(ItemCountryOrNationality,"France"),(ItemArtistDate,"1796-1875"),(ItemArtistOrMaker,"Jean Baptiste Camille Corot"),(ItemItemDate,""),(ItemDateOrPeriod,"c. 1843, published April 1st, 1875. Printed by A. Salmon, Paris"),(ItemTypeOfObject,"Print"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"1 Hanson #41")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/Shlbi8G5awI/AAAAAAAABoU/NxQ-H2USxAo/s800/DSC_0281.JPG", fileChksum = "b743dc61d9721ee357c2b0d8555d5dac", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/Shlbi8G5awI/AAAAAAAABoU/NxQ-H2USxAo/s800/DSC_0281.JPG", fileChksum = "b743dc61d9721ee357c2b0d8555d5dac", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "30cbc4b81ec19aa1ba6b1441a264df7b", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 194, imageFileHeight = 129, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "da929ce67ba02b3f079f047d0a712aba", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2707, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/Shlbi8G5awI/AAAAAAAABoU/NxQ-H2USxAo/s800/DSC_0281.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Gallery label on back. Original print 1904. Published in 1919.\n\nAccording to the Catalog Raisonne, the original lithograph is from 1904.\n\nThis print has sold at auction for between $1000 and $6000 mostly from this edition.\n\nA single print from the edition of 950 (``on wove paper'') sold at Christie's New York 5/2/07 for $6000.\n\nAnother print ``Ambroise Vollard\" of the print-maker Vollard from the same series: ``Douze Lithographies Originales\" of 1919 is available from Davidson for $3800.\n\nMasterworks Fine Art: Claude Renoir, La Tete Baissee. Original print on Japon paper, edition of 50. Offered for $8700.\n\nArmstrong Fine Art\n``I currently have an impression of \"Claude Renoir, la Tete Baissee\" in pristine condition for $5,500.  It is a beautiful impression on a full sheet of Arches ``wove paper'' (implies vellum??, see above).  There are no tears, creases, discolorations, etc.  There are only two acid-free cloth hinges on the verso at the top of the sheet, which could easily be removed.  Aside from that, it is perfect.''\n\nIf you have further questions, please do not hesitate to contact me.\n \nSincerely, Bernard\n \nBernard Derroitte\nArmstrong Fine Art\n200 East Walton Place\nChicago, IL  60611\nT.  312.664.9312; F.  312.664.9862\nwww.armstrongfineart.com\n\nLouis Valtat. c. 1904. Lithograph. Delteil, Stella 38. 11 5/8 x 9 1/4 (sheet 13 x 9 3/4). Series: Douze Lithographies Originales de Pierre-Auguste Renoir. A fine impression apart from the edition of 1,000 printed on cream laid paper with deckle edges on three sides. Signed in the stone. $4,500.\n"),(ItemCashValue,"$4500"),(ItemMarketAnalysis,"Renowned French Impressionist painter Pierre-Auguste Renoir took up printmaking later in life after 1890, and is thought to have executed some 59 lithographs and etchings, some of which are regarded as some of the best of the period. Renoir was the first of the Impressionists to enjoy popular success and his prints were produced in recognition of that success. Stella lists 60 prints (25 etchings and 35 lithographs) in his catalogue of Renoir's prints. Michel Melot, The Impressionist Print notes that most of Renoir's prints were produced after he had became famous and represent an attempt to satisfy the demand for his works, almost 3/4 of them having been produced after 1900 for inclusion in books and portfolios produced by Theodore Duret, the chief chronicler of the Impressionists, and Ambroise Vollard, Renoir's dealer and a great lover of the art of printmaking. One of Renoir's subjects of choice was portraits of children and friends.\n\nValue characteristics for Renoir lithographs include quality of impression, state and condition and type of paper. Executed in 1904, this is an impression from 2nd and final state which was published for ``L'Album des Douze Lithographies Originales de Pierre August Renoir\" by A. Vollard (Paris, 1919) in and edition of 1000, with 50 printed on Japon paper and 950 printed on Arches wove paper. Some images from this series were also printed separately on laid paper. At Allinson Gallery, a print from this series on laid paper is available for $4500.\n\n\nExamples of the print are available at retail with the print from the edition of 950 available for  $5500, and the two from the edition of 50 for $6500 and $8700."),(ItemDescription,"Three-quarter view bust portrait of a young boy looking down."),(ItemCost,"$785"),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Galleries"),(ItemCondition,"Very good. Discoloration to upper horizontal and right vertical margins. No tape residue on reverse. UV light damage to the sheet. Deckled edges."),(ItemMediumMaterialsTechniques,"On laid paper. No watermark."),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Second state. Edition: Ambroise Vollard, Paris, 1919. Printer: Auguste Clot. Edition of 1000. One from the series, \"Douze Lithographies Originales\". D. 39"),(ItemSignatureInscriptionsMarkings,"Signature L/R, signed in stone"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Mat opening: 9 3/4'' x 8''\nSheet: 13\" x 10\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"21 5/8'' x 17 5/8''"),(ItemSupportBaseFrame,"Glazed gilt wood frame"),(ItemSubjectOrTitle,"Claude Renoir, La Tete Baissee"),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1841-1919"),(ItemArtistOrMaker,"Auguste Renoir"),(ItemItemDate,""),(ItemDateOrPeriod,"c. 1904"),(ItemTypeOfObject,"Print (Davidson)"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"2 Hanson #37")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SmTiIV_kBoI/AAAAAAAAC-4/GfT1RJYXZXc/s800/Renoir.jpg", fileChksum = "340ae2f791e548f90b9a345b32cf894e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 600, imageFileHeight = 678, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SmTiIV_kBoI/AAAAAAAAC-4/GfT1RJYXZXc/s800/Renoir.jpg", fileChksum = "340ae2f791e548f90b9a345b32cf894e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 600, imageFileHeight = 678, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "d8a06582fa6ab7548c1a1d3f3a1c8a03", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 149, imageFileHeight = 168, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "7b3d6e27ca4467778a146c3df16d84b2", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1593, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/SmTiIV_kBoI/AAAAAAAAC-4/GfT1RJYXZXc/s800/Renoir.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"\"A Basement Shop in Cracow\" $2,000. Lockport Street Gallery\n\nAu Pied du Sinai: Schlome Fuss A La Synague, ?, Lewis and Maese Auctions, Houston, TX, auction est $3,750 - $4,250 lithograph in black-brown on china paper, signed in pencil Lower right. Rare - 25 proofs, deluxe book edition iii.\n"),(ItemCashValue,"$2,000"),(ItemMarketAnalysis,"Henri Toulouse-Lautrec was a French painter, printmaker, and illustrator best known for his depictions of life in the bohemian world of the Parisian theater and its environs at the end of the 19th century. The subject property is from a series of works depicting Jewish life that illustrated a volume of sketches or travel essays on Eastern European Jewish life by Georges Clemenceau, who later served as prime minister of France. Toulouse-Lautrec is said to have observed Jewish life in Paris as the basis for his illustrations.\n\nThere is one state only of this lithograph, one of a set of eight published by Henry Floury. The group of eight lithographs were published in a deluxe book edition with four variations, and in a regular book edition in two variations.\n\nValue characteristics include edition and condition. The subject property is from the regular book edition, var. ii, as indicated by its sheet size, and is a good quality impression in good condition. This print occasionally appears at auction and at retail, as part of the set of eight and also as a single work. A national survey of print dealers located two offerings ranging from $2000 to $4500.  The higher comparable is for a print from the Deluxe book edition, and the lower comparable for a print from the Regular Book edition similar to the subject property. The lower comparable is the basis for the value conclusion of Replacement Value Comparable of $2,000."),(ItemDescription,"The composition features an elderly man with a long beard and wearing a coat and hat, looking out of a basement level window adjacent to a sidewalk. His hands are resting on the window ledge at the same level as the sidewalk. A small black cat sits at the edge of the window near the man's right hand. The window is lined with large blocks, evoking a an older stone building. The composition emphasizes shades of gray, in the figure of the man window and building, so that the black cat serves as a focal point."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Good impression in very good condition. Minor UV light damage. No adhesive or tape on reverse."),(ItemMediumMaterialsTechniques,"Lithograph on wove paper, no watermark"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Series: \"Au Pied du Sinai\", 1 state only"),(ItemSignatureInscriptionsMarkings,"Unsigned"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Image impression: 170 mm x 140 mm; Sheet size: 260 mm x 196mm"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"16\" x 13 5/8\""),(ItemSupportBaseFrame,"Gilt wood frame, glazed"),(ItemSubjectOrTitle,"\"Une Arriere-boutique a Cracovie\" (A Basement Shop) from the series \"Au pied du Sinai\"(From the Foot of Sinai)"),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1864-1901"),(ItemArtistOrMaker,"Henri Toulouse-Lautrec"),(ItemItemDate,""),(ItemDateOrPeriod,"1898"),(ItemTypeOfObject,"Print"),(ItemLocation,"Living room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"3 Hanson #36")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "\"Une Arriere-boutique a Cracovie\"", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SmThoUC_IxI/AAAAAAAAC-o/T4jaLSFLz0c/s400/Toulouse%20Lautrec1.jpg", fileChksum = "ed9232e7fa5b87ba495cacce48c03ee9", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 321, imageFileHeight = 400, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SmThoUC_IxI/AAAAAAAAC-o/T4jaLSFLz0c/s400/Toulouse%20Lautrec1.jpg", fileChksum = "ed9232e7fa5b87ba495cacce48c03ee9", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 321, imageFileHeight = 400, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "9e82c2f080523b6b49be954d3aea9a4e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 142, imageFileHeight = 177, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "20b6fc9e14bd2f4041c1c071c35a6b10", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1445, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_s-XE562xJtM/SmThoUC_IxI/AAAAAAAAC-o/T4jaLSFLz0c/s400/Toulouse%20Lautrec1.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Etching, \8220Tampoco\8221 from \8220Desastres de la Guerra\8221,  is $2,250., David Tunick, Inc.\n\nEtching \8220Bien te esta\8221 from from \8220Desastres de la Guerra\8221, proof impression is $140,000, Harris Schrank\n\n"),(ItemCashValue,""),(ItemMarketAnalysis,"Francisco de Goya, the Spanish master painter, illustrator and engraver of the 19th century is well known for his series of prints describing the human cost of the Napoleonic Wars in Spain called \8220Los desastres de la Guerra,\8221 (The Disasters of War). The series includes some 80 prints of which only some proofs and individual prints were made during the artist\8217s lifetime. The first published edition was issued posthumously in 1864, and ultimately seven editions were printed, in varying degrees of quality.\n\nValue characteristics for Goya prints from this series include edition, condition, quality of impression, and plate or subject. At auction the value range is dramatic, with a rare proof of \"Bien te esta\" reaching more than $140,000 to later edition lesser quality impressions selling for less than $500. The subject property, an impression of plate 45 titled \8220Tambien estos\8221, is from the first edition. It  is a good quality impression in very good condition. A national survey of galeries did not locate any prints of this plate, but did locate several first edition prints of other plates in the series. These range from ? to ?. The subject property, is a very good impression in generally good condition.\n"),(ItemDescription,"The composition depicts a room full of wounded soldiers, some lying on the floor and on beds, others seated and standing. Three figures form a triangle in the composition: the seated figure on a bed in the foreground, the standing figure looking over his shoulder in the center and the figure lying on the right in the middle."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very good impression in good condition. Upper and lower margins have been trimmed. Mat burn but no foxing and UV light exposure. Remnants of tape residue on reverse of the sheet. Inscribed along bottom margin. \"1-13-37\". Inscription on reverse of sheet L/R corner: ``Disasters of war, 1st edition''. "),(ItemMediumMaterialsTechniques,"Etching on wove paper with no watermark."),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Series: Los Desastres de la Guerra, first edition"),(ItemSignatureInscriptionsMarkings,"Unsigned, Composition L/L; no. 25 in U/R margin"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Image size: 4 1/2'' x 7 1/2'' (11 1/2 cm. x 19.2 cm); \nPlate impression: 6 1/8'' x 9 1/2'' (15 1/2 cm x 23 1/2 cm)\nSheet size: ???"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"19\" x 20 3/4''"),(ItemSupportBaseFrame,"Glaze gilt wood frame"),(ItemSubjectOrTitle,"\"Tambien Estos\" (Also these) from the \"Los desastres de la guerra\" (Disasters of War) series"),(ItemCountryOrNationality,"Spanish"),(ItemArtistDate,"1746-1828"),(ItemArtistOrMaker,"Francisco de Goya "),(ItemItemDate,""),(ItemDateOrPeriod,"1898"),(ItemTypeOfObject,"Print"),(ItemLocation,"Living room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"4 Hanson #32")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/SmThoXPWNmI/AAAAAAAAC-I/4U5lZrDuWkA/s640/Goya1.jpg", fileChksum = "10c5286b55fa1707077e866f3c8bbad5", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 640, imageFileHeight = 479, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/SmThoXPWNmI/AAAAAAAAC-I/4U5lZrDuWkA/s640/Goya1.jpg", fileChksum = "10c5286b55fa1707077e866f3c8bbad5", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 640, imageFileHeight = 479, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "9ad9cea61d166acdf7a95b491f2faa36", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 183, imageFileHeight = 137, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "f3b70a8be1092118de3857829f4f6465", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2405, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh4.ggpht.com/_s-XE562xJtM/SmThoXPWNmI/AAAAAAAAC-I/4U5lZrDuWkA/s640/Goya1.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Gallery label on back\n\nJayne Haslem Gallery, The Zither, 9 7/8'' x 10 3/4\" Aquatint and etching, $1,500\n\nThe Annex Gallery, The Zither, 24/60 published 1982. $750.\n\nKiechel Fine Art, August, 1940 etching 9 3/4\" x 8\" $1,200\n\nEtchings and acquatints of the same era have sold at auction for between $418 and $3,200 with the majority in the $1100- $2400 range."),(ItemCashValue,"$1,500"),(ItemMarketAnalysis,"Born in Beverly, Massachusettes in 1911 American contemporary artist, Will Barnet is an acclaimed painter and printmaker, who has also taught art at the leading art schools in the U.S.\n\nTrained at the Boston Museum of Fine Arts School and the Art Students League New York, he later taught art at such leading American schools as Yale University, Cornell University, and the Art Students League.\n\nA prolific graphic artist, Barnet changed his style significantly at different points in his career. His earliest works were influenced by expressionism; they were followed by abstract works in the 1950s and 1960s, and finally evolved into more figurative works of silhouetted forms set against geometrically designed backgrounds.\n\nHis work has been exhibited in prominent museums and galleries in the United States and Canada and is included in many prestigious collections, including the Metropolitan Museum of Art and the Guggenheim Museum in New York and the Boston Museum of Fine Arts. The subject property is an example of Barnet's expressionistic early work. A search of auctions and galleries did not turn up any available examples of ``The Bath''. Other aquatints and etchings of this period have sold at auction for between $418 - $3200. At retail, prints of this era have sold for between $750 and $1500 depending on edition. Because ``The Bath'' is currently scarce, the subject property is valued at the high end of this range."),(ItemDescription,"A view from above of a woman taking a bath with towels draped on a rack in the background."),(ItemCost,"$107"),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Galleries. Mercer Island, WA"),(ItemCondition,"Very good"),(ItemMediumMaterialsTechniques,"Acquatint etching"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Edition of 12"),(ItemSignatureInscriptionsMarkings,"L/R signature, title II ``The Bath''; Edition 12"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 9 3/4\" x   12\"; Impression: 6 1/2\" x 8 3/4\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"The Bath"),(ItemCountryOrNationality,"American"),(ItemArtistDate,"b. 1911"),(ItemArtistOrMaker,"Will Barnet"),(ItemItemDate,""),(ItemDateOrPeriod,"1937, Cole 46"),(ItemTypeOfObject,"Print, Davidson"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"5")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShlbvJW0ffI/AAAAAAAABqc/kv21ucxKpug/s800/DSC_0343.JPG", fileChksum = "2741073cabe851ab5f85e641b3127b12", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 632, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShlbvJW0ffI/AAAAAAAABqc/kv21ucxKpug/s800/DSC_0343.JPG", fileChksum = "2741073cabe851ab5f85e641b3127b12", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 632, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "571790e957786b7b125ee1197cb262be", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 178, imageFileHeight = 141, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "39b9eb87169cf6ce97850c8c4b23e3ae", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2278, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/ShlbvJW0ffI/AAAAAAAABqc/kv21ucxKpug/s800/DSC_0343.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Male Figure, lithograph, 1902, edition of 50 offered for $950 at Davidsons.\nDeux Nus recently sold at Davidson's for $950\nDeux Nus offered for sale at Figaro Gallery for $1800.\nDavidson's recently sold a print for $950"),(ItemCashValue,"$1,800"),(ItemMarketAnalysis,"A French painter, illustrator and printmaker, member of the group of artists called the Nabis and afterward a leader of the Intimists; he is generally regarded as one of the greatest colourists of modern art. He also produced lithographs and dry-point. In the 1920s he undertook a series of paintings on one of his most famous themes\8212a nude in a bath, the subject of this lithograph. Les Baigneuses  was probably engraved for Ambrose Vollard who was planing an Album de Nus with various artists. Value characteristics for Bonnard prints include quality of impression, state and condition. A search of auction prices indicate that lithographs of nudes by Bonnard have sold for between $1200 and $2880. At the retail level lithographs of nudes, including 2 examples of Les Baigneuses, have sold and are available at the retail level for between $950 and $1800. Les Baigneuses is currently available the The Figaro Gallery and is the basis for estimating the subject property's Replacement Value. \n\n"),(ItemDescription,"Impressionistic rendering of two nude women situated in a landscape. One lies on the grass while the other stands in profile next to a small tree or bush. Rendered in a loose, impressionistic line. "),(ItemCost,"$232"),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"Davidson Galleries, Mercer Island, WA"),(ItemCondition,"Very Good. Nothing to note."),(ItemMediumMaterialsTechniques,"Drypoint and roulette. Handmade wove paper (Japanese paper)"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"Edition of 50"),(ItemSignatureInscriptionsMarkings,""),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 26 cm x 31 cm 10\" x 12\"; Impression:  16 cm x 23.2 cm  6 1/2\" x 23.2 cm. 9 1/4\" \t\t\t\n"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Deux Nus (Les Baigneuses)"),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1867-1947"),(ItemArtistOrMaker,"Pierre Bonnard"),(ItemItemDate,""),(ItemDateOrPeriod,"c. 1927-29"),(ItemTypeOfObject,"Print (Davidson)"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"6, Hanson #18")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/ShlbvtyXvFI/AAAAAAAABqk/cUWgS7DuEH8/s800/DSC_0345.JPG", fileChksum = "86cc8540086f6ff5101b3f370297ae92", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 624, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/ShlbvtyXvFI/AAAAAAAABqk/cUWgS7DuEH8/s800/DSC_0345.JPG", fileChksum = "86cc8540086f6ff5101b3f370297ae92", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 624, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "33c6d8738ebaf0cf5ede0f8b1903896d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 179, imageFileHeight = 140, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "66cd9fdb2fc7a56c64c2fc4f760b85e3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2308, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh3.ggpht.com/_s-XE562xJtM/ShlbvtyXvFI/AAAAAAAABqk/cUWgS7DuEH8/s800/DSC_0345.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Jean-Luc Baroni, Ltd.of London has a drawing for sale. As yet I have not located galleries that sell his prints..\n\nMost recent sale of print, Elefanten, ein Lowe unde ein Eber, 33.2 x 47.7 cm. Bassenge, 04/06/09 $1050\n\nEtchings by are available at Spaightwood Galleries for between $1750- $4500. Most are much smaller than the subject property. "),(ItemCashValue,"$4,500"),(ItemMarketAnalysis,"Known as a Mannerist painter, draughtsman and printmaker, Battista Franco spent the first years of his independent career in Rome, but by 1536 had settled in Florence. Franco is better known, and has usually been more highly regarded, as a draughtsman rather than as a painter. He is also well known for his etching having produced over 100 during his life-time. An assiduous copyist, among his drawings and etchings are numerous studies of antique subjects, usually taken from reliefs on Roman sarcophagi. According to the Illustrated Bartsch, The Deluge was likely after a work designed by Polidoro da Caravaggio.\n\nValue characteristics of prints include state, quality of impression, edition size, subject matter and condition. The subject property is in good condition and is a fine impression. A search for etchings by this artist indicate that the primary  market for this artist is in Europe with most sales taking place at Galerie Bassenge Kunst Und Buchauktionene and Reiss & Sohn in Berlin.  Prints of similar biblical subjects, size, quality of impression, condition and subject matter range from $1401-$2424. The subject property is most comparable in quality of impression to prints at the high end of this range. A search of etchings by the artist in galleries in the United States only turned up a single venue, Spaightwood Galleries in Upton, Massachusetts where a number etchings are available for between $1750-$4500. Most of the etchings available for sale are considerably smaller than the subject property. The etching most closely approximating the subject property in size and quality of impression is available for $4,500 and is the basis for estimating the subject property's replacement value\n"),(ItemDescription,"A biblical scene of the flood said to be taken from a work by Polidoro da Caravaggio depicting people fleeing to higher ground to escape the rising waters. Figures are seen pulling themselves from the water and helping women and children to higher ground. The sky is heavy with clouds with heavy vertical lines depicting torrents of rain. The arc is shown in the middle-ground of the composition. "),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Good: Some stains and creases. Good impression."),(ItemMediumMaterialsTechniques,"Etching"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signed in the plate lower left, Battista Franco Inu"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 10 1/2\" x 17\" (27 cm x 43 cm)"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"25 1/4\" x 30 1/4\""),(ItemSupportBaseFrame,"Glazed silver gilt wood frame"),(ItemSubjectOrTitle,"The Deluge, Le deluge universel, 3, Bartsch"),(ItemCountryOrNationality,"Italy"),(ItemArtistDate,"1498-1561"),(ItemArtistOrMaker,"Giovanni Battista Franco Inu"),(ItemItemDate,""),(ItemDateOrPeriod,"16th century"),(ItemTypeOfObject,"Print"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"7, Hanson #66")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG", fileChksum = "dac291fe9972381edd30621a4739a12c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG", fileChksum = "dac291fe9972381edd30621a4739a12c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "a699a3975a3ee1fb393e47e1ab5b3ca1", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 194, imageFileHeight = 129, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "592c3b006543a18d7373ab57901c5832", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2707, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG"},Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG", fileChksum = "dac291fe9972381edd30621a4739a12c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG", fileChksum = "dac291fe9972381edd30621a4739a12c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 800, imageFileHeight = 532, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "a699a3975a3ee1fb393e47e1ab5b3ca1", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 194, imageFileHeight = 129, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "592c3b006543a18d7373ab57901c5832", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2707, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_s-XE562xJtM/SiBTq5JRCeI/AAAAAAAAB4I/sAJBgb3dMcc/s800/DSC_0383.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Label on back has biography of artist\nSothebys 2006 1860 \"Garden views with green houses, figures, and gardeners\", 5.3 in x 8.3 in, watercolor over black chalk, $1,566\nthere are additional drawing comps on Artprice, also for Jean-Victor Bertin"),(ItemCashValue,"$1,566?"),(ItemMarketAnalysis,"Charles Nicolas Ransonnette was a painter, engraver and draughtsman active in the mid-19th century. He served as official draughtsman and engraver to the Duchess of Berry, and exhibited regularly at the French Salon from 1822 to 1857. He was a pupil of neo-classical painter Jean-Victor Bertin, and his father, Pierre Nicolas Ransonette, also a painter, engraver, and draughtsman, served the French royal family. \n\nRansonnette worked in a finely detailed and charming style, with many works in small scale. Works by Ransonnette rarely appear in the American market. No current retail offerings could be located and only a few auction sales. The most recent, is for a large scale watercolor over chalk on paper landscape, 5.3 in. x 8.3 in., which sold for $1,566. The subject property is comprised of two ink on paper works framed together. Each work is approximately one half of the comparable. "),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very good"),(ItemMediumMaterialsTechniques,"pen and ink and wash on laid paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"said to be signed on verso under dust jacket"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Mat opening: Upper and Lower: 4\" x 5"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"20 1/2\" x 15 1/2\" "),(ItemSupportBaseFrame,"Gilt and black paint glazed wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,"French"),(ItemArtistDate,"1793-1877"),(ItemArtistOrMaker,"Charles-Nicolas Ransonnette"),(ItemItemDate,""),(ItemDateOrPeriod,"1856"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"8, Hanson #38")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5A3amjI/AAAAAAAABx0/IKqO3Zh6LHU/s512/DSC03332.JPG", fileChksum = "e1aca3d54e41127c581be92762f4b8bd", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5A3amjI/AAAAAAAABx0/IKqO3Zh6LHU/s512/DSC03332.JPG", fileChksum = "e1aca3d54e41127c581be92762f4b8bd", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "9e7179bac603aebdd740d2f5d8bd9074", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 129, imageFileHeight = 194, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "cd0cd9d06928104877468460ea075fc6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1199, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh3.ggpht.com/_s-XE562xJtM/Shwm5A3amjI/AAAAAAAABx0/IKqO3Zh6LHU/s512/DSC03332.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,""),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Good: Ares of loss upper right, upper left quadrant and center"),(ItemMediumMaterialsTechniques,"Oil on canvas"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,""),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"9 1/2\" x 13 1/12\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"15 1/8\" x 19 1/4\""),(ItemSupportBaseFrame,"Glazed gilt wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,""),(ItemArtistDate,""),(ItemArtistOrMaker,"Hudson River School"),(ItemItemDate,""),(ItemDateOrPeriod,""),(ItemTypeOfObject,"Painting"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"9, Hanson #43")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SiBTr4P6xmI/AAAAAAAAB4Q/hY6AhRf1thQ/s720/DSC_0374.JPG", fileChksum = "6bf4434d44c3c6ee243eea34bfdf73e3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 479, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/SiBTr4P6xmI/AAAAAAAAB4Q/hY6AhRf1thQ/s720/DSC_0374.JPG", fileChksum = "6bf4434d44c3c6ee243eea34bfdf73e3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 479, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "4547a89fcc964f7278edf8f86427dcd3", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 194, imageFileHeight = 129, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "1b0078d0ead16f4279a50a75ce3f7f41", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2706, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/SiBTr4P6xmI/AAAAAAAAB4Q/hY6AhRf1thQ/s720/DSC_0374.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Oil on panel works available at auction in the Netherlands and Germany as well as Sotheby's and Christie's, Amsterdam, Bonham's and Butterfields, Leslie Hindman in the USfor between $250 and $6000 (Bonham's, 5/16/07 oil on board 15 3/4\" x 12\" A Fisherman on a dock) with the majority selling in the $1000-$2500 range. \n\nlots 606, 607 oil on panel, portraits'65 Leslie Hindman:  1/18/09, $1464\nGlerum, Amsterdam, boy with goat, 1970 $930\nGlerum, Amsterdam, 6/2/08 $1085\n\n\n\nA painting by the artist is available at Kamp Galleries in Winnetka, IL. \nDear Lydia,\nThanks for your inquiry about our Van d. Berg Still Life.  This is the only painting we have by the artist\nand we believe it to be an early example, between the wars, ca 1925.  The dealer price is $2,900.00\n(Retail $5,900.00).  It's a beautiful example of his work, and a good representation of art from that era.\nYou should buy it.\nRegards,\nNick Vahlkamp, mgr.\n- Show quoted text -\n-- \nKamp Gallery\nDrake Hotel Arcade\n140 East Walton Street\nChicago, IL 60611\n312-664-0090\nfax 312-664-0014\n\n\nHis name is also listed on Covington Fine Arts Gallery in Tuscon, AZ.\n6530 E Tanque Verde Rd Suite 140\nTucson Arizona, 85715 USA\nTel: \t(520) 298-7878\nAlso see M. Klein Auctions in Chicago. \n400 W. Hubbard \nChicago Illinois, 60622 USA\nTel: \t312-948-0080\nFax: \t312-948-0090\nListed on Artnet\n"),(ItemCashValue,"$2,900"),(ItemMarketAnalysis,"Willem van den Berg studied art at the Academie voor Beeldende Kunst in The Hague and traveled extensively to Belgium, France, Italy, Austria and England. \nKnown as  a draughtsman, painter and graphic artist, Van Den Berg's style is characterized by a preference for strong closed contours which reflects his interest in Japanese wood block prints. He was director of the Rijksacademie in Amsterdam from 1938 to 1953. He died in Amsterdam in 1970. His paintings are in the collections of museums in Holland, Europe and the United States. The majority of Van Den Berg's works are available through auctions in Amsterdam and Germany. They have sold for between $732 (7.5\" x 5.25\"), Portrait on panel and $6000. Van Den Berg is particularly known for his oil on panel portraits of fishermen and farmers, and these images have brought the highest prices with majority n the $1000-$2500 range. Paintings by the artist comparable to the subject property's pastoral subject that have come up at auction in recent years have all failed to sell indicating a weak market for this subject. A search for Van Den Berg's works for sale in galleries the United States and Europe turned up a single example - an oil on canvas painting still life of garden flowers which is available for $2900 and is the basis for the value conclusion for the subject property."),(ItemDescription,"Gentle pastoral landscape of Dutch countryside. A woman drives two cows on a tree-lined road next to a canal set against a cloud-scudded sky."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very good: surface crackelure"),(ItemMediumMaterialsTechniques,"oil on panel"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"W. V. D. Berg"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"11 1/4\" x 15 1/4\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"16 3/4\" x 20 3/4\""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,"Holland/Netherlands"),(ItemArtistDate,"1886-1970"),(ItemArtistOrMaker,"Willem van den Berg"),(ItemItemDate,""),(ItemDateOrPeriod,"early 20th century"),(ItemTypeOfObject,"Painting"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"10")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwmvncL_sI/AAAAAAAABww/33e5fwELMLY/s720/DSC03318.JPG", fileChksum = "49dd80467b4515462303f794b29ffe58", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 504, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwmvncL_sI/AAAAAAAABww/33e5fwELMLY/s720/DSC03318.JPG", fileChksum = "49dd80467b4515462303f794b29ffe58", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 720, imageFileHeight = 504, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "90b383820e81928940f738d65cf7a70f", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 189, imageFileHeight = 132, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "91cfb0af5d0e0f17e6f33b88eb4842d6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2571, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/ShwmvncL_sI/AAAAAAAABww/33e5fwELMLY/s720/DSC03318.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Paid $150 to Cal State Northridge\nOne was offered by Delorme-Collin du Bocage in Paris on 1 July 2009. It was estimated at 2,500-3,000 euros but failed to sell. The second example was offered by Christie's New York on 24 November 2008. It was estimated at $4,000-$5,000, and sold at $3,000.\n\nOne was offered by Delorme-Collin du Bocage in Paris on 1 July 2009. It was estimated at 2,500-3,000 euros but failed to sell. The second example was offered by Christie's New York on 24 November 2008. It was estimated at $4,000-$5,000, and sold at $3,000.\n\n"),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,"Portrait of Native American activist and celebrity Russell Means depicted in a bust view, wearing braids and a traditional bone choker. The background features shades of blue that contrast with the brown skinned figure and the peach colored shirt he is wearing."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,""),(ItemMediumMaterialsTechniques,""),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signed vertically, Andy Warhol"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"49\" x 27\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"61\" x 38 1/2\""),(ItemSupportBaseFrame,"Stained wood frame, glazed"),(ItemSubjectOrTitle,"\"Russell Means\" from the \"American Indian\" series"),(ItemCountryOrNationality,"American"),(ItemArtistDate,""),(ItemArtistOrMaker,"Andy Warhol"),(ItemItemDate,""),(ItemDateOrPeriod,"1972"),(ItemTypeOfObject,"Poster"),(ItemLocation,"Stairwell"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"11 Hanson #65")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/ShwnGFc7K9I/AAAAAAAABzA/0hx7n8lKdoI/s800/DSC03351.JPG", fileChksum = "a5c3128491db2ae76606a856354db49d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 534, imageFileHeight = 800, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/ShwnGFc7K9I/AAAAAAAABzA/0hx7n8lKdoI/s800/DSC03351.JPG", fileChksum = "a5c3128491db2ae76606a856354db49d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 534, imageFileHeight = 800, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "437579f98bb162682cd04bb705fac5be", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 129, imageFileHeight = 194, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "df596a1cec9291a0afd8a0e3cfed0fcb", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1202, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh4.ggpht.com/_s-XE562xJtM/ShwnGFc7K9I/AAAAAAAABzA/0hx7n8lKdoI/s800/DSC03351.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"Painted screens available at Ark Antiques are and Edo period, 6 panel screen, 4'6\" high in very good condition and very good quality offered for $2600.\nA small screen, approx. 3'5\" high, 6 panels,  Japanese, 19th century of birds, duck, chickens and foliage available for $1800.\n\n8 panel painted Chinese screen sold at Millea Brothers for $2000 + BP 10/29/06\n\n"),(ItemCashValue,"$2,000"),(ItemMarketAnalysis,"This screen was produced for export to the Western markets.  Vintage screens from China are available from importers and at estate sales and regional auction houses. Value characteristics include age, quality of painting, size and condition. The subject property was purchased at an art sale at California State University at Northridge. Therefore the appropriate market level for this piece are works found at high-end consignment stores and regional auction houses. Vintage screens similar to the subject property in size, age and quality are available for between $575 and $3,500. The subject property's relatively large-size and good quality make it most comparable to examples in the middle of this value range. A pair of four-panel painted Chinese screens depicting boaters on river amid mountains sold at an Millea Brothers auction house in New Jersey for $2,000 is the basis for the subject property's replacement value.\n "),(ItemDescription,"A six-panel painted screen of ink and colors on paper depicting a landscape of misty mountains, tiny villages and boaters along a river. It is mounted on a lacquer frame. "),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"California State University at Northridge"),(ItemCondition,"Good: First panel: buckling, a hole in upper quarter; second panel, some slight buckling; third panel: one small buckle in center; fourth panel: horizontal buckling with creases and some loss of pigment; fifth panel: some buckling in center and lower half; sixth panel: slight abrasions towards left center border; gilt paint border: abrasions along areas of borders. "),(ItemMediumMaterialsTechniques,"ink and colors on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"L/L of leftmost panel, signature and title\nSame 2 seals on each panel, alternates lower right and lower left."),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"53 1/2\" x 20\" each of 6 panels 120\" wide overall"),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"68\" x 24 1/2\" (each panel)"),(ItemSupportBaseFrame,"Lacquer frame and silk backing"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,"China"),(ItemArtistDate,""),(ItemArtistOrMaker,""),(ItemItemDate,""),(ItemDateOrPeriod,"20th century"),(ItemTypeOfObject,"Chinese screen"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"12")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwnguXPCWI/AAAAAAAAB0s/JQX8q9C2o8A/s800/DSC03377.JPG", fileChksum = "61e3acb3509a7850aafa1a9774ac144a", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 558, imageFileHeight = 800, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwnguXPCWI/AAAAAAAAB0s/JQX8q9C2o8A/s800/DSC03377.JPG", fileChksum = "61e3acb3509a7850aafa1a9774ac144a", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 558, imageFileHeight = 800, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "50299ad1391ec8e70fb1e8321436f211", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 132, imageFileHeight = 189, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "38f5d18e2e63a03ba4fbc50b06517b6e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1256, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/ShwnguXPCWI/AAAAAAAAB0s/JQX8q9C2o8A/s800/DSC03377.JPG"},Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/ShW9p3SjUNI/AAAAAAAABnA/jZoUWkJzQU4/s800/DSC03368.JPG", fileChksum = "e7b6294ca559ae677fb48b6d3f55337a", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 534, imageFileHeight = 800, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh4.ggpht.com/_s-XE562xJtM/ShW9p3SjUNI/AAAAAAAABnA/jZoUWkJzQU4/s800/DSC03368.JPG", fileChksum = "e7b6294ca559ae677fb48b6d3f55337a", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 534, imageFileHeight = 800, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "cfe486ec7a52f24205108d9774a04eb9", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 129, imageFileHeight = 194, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "1a022d14bb0e0f4fb719b44183fa7fee", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1202, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh4.ggpht.com/_s-XE562xJtM/ShW9p3SjUNI/AAAAAAAABnA/jZoUWkJzQU4/s800/DSC03368.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,""),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,"Reverse composition features abstracted seated nude female figure on red field. Emphasis is on breasts and thighs with no detailing of face or lower extremities."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"purchased from the artist ca. 1985"),(ItemCondition,"Excellent"),(ItemMediumMaterialsTechniques,"woodcut on handmade paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"12/16"),(ItemSignatureInscriptionsMarkings,"Numbered lower left margin; signed and dated lower right margin; embossed stamp lower right margin"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Untitled"),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1972"),(ItemTypeOfObject,"Print"),(ItemLocation,"Dining room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"13 Hanson #26")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SoSr8iQyWJI/AAAAAAAAB0E/hscFnMAb58c/s512/IMG00013.jpg", fileChksum = "bec5bacaf92f4d696a29e2e9e7744f01", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 384, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SoSr8iQyWJI/AAAAAAAAB0E/hscFnMAb58c/s512/IMG00013.jpg", fileChksum = "bec5bacaf92f4d696a29e2e9e7744f01", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 384, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "8e585423d482abea53f36ac243f87b67", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 137, imageFileHeight = 183, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "33d652f5435a93658161916f5ef91d11", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1350, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_EV171eW_Cmg/SoSr8iQyWJI/AAAAAAAAB0E/hscFnMAb58c/s512/IMG00013.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"print acquired from the artist in 1985 according to Ms. Hanson"),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,"Three color composition features man wearing hat bending down to pick up a pepper on a sidewalk. The pepper is red and yellow, while the balance of the composition is black and white."),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,"purchased from the artist ca 1985"),(ItemCondition,"Very Good"),(ItemMediumMaterialsTechniques,"woodcut"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,"8/12"),(ItemSignatureInscriptionsMarkings,"signed and numbered, lower right margin; embossed stamp lower right margin"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"\"Mark Tobey\""),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1975"),(ItemTypeOfObject,"Print"),(ItemLocation,"Dining room"),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"14 Hanson #27")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_EV171eW_Cmg/SoSrvxqq2JI/AAAAAAAAB0A/8y9y53D2YKo/s512/IMG00016_2.jpg", fileChksum = "79221c6448716b08b8ab19085b6bc582", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 392, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_EV171eW_Cmg/SoSrvxqq2JI/AAAAAAAAB0A/8y9y53D2YKo/s512/IMG00016_2.jpg", fileChksum = "79221c6448716b08b8ab19085b6bc582", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 392, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "90c3fc977a59abfda729d9fb50b0137f", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 138, imageFileHeight = 181, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "d21aad082a06bb4e5790de5ed9166967", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1378, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_EV171eW_Cmg/SoSrvxqq2JI/AAAAAAAAB0A/8y9y53D2YKo/s512/IMG00016_2.jpg"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"pastel \" Standing Female Nude \" is available, framed and in excellent conditions and as described on our site. The asking price is $ 9.500, but because we are having a special summer promotion, I can give you a discount of 25%, for a net of $ 7.125, including packing, shipping and insurance Pasquale Ianetti Gallery, San Francisco, CA\n\npastel Nude, was offered for $14,000 at Jack Rutberg Gallery, Los Angeles, CA\n"),(ItemCashValue,"$10,000?"),(ItemMarketAnalysis,"Hans Burhardt worked in a variety of media during his long career, including painting, drawing and print making. Known primarily as an abstract impressionist, Burkhardt also worked in a figurative style. Born in Switzerland, Burkhardt came to the United States in ??. He moved to Los Angeles in 1937 and was a well known and well regarded figure in the West Coast art scene. Burkhardt\8217s work has received recent attention due to two major projects, a retrospective of his work mounted at the University Gallery at California State University Northridge, and a comprehensive exhibition of his 1960s paintings presented at the Jack Rutberg Gallery in Los Angeles in 2008-2009. These projects have brought renewed public attention to the artists\8217s work reflected in a greater wolume of sales at auction.\n\nValue characteristics for Burhardt prints include date, size, condition and quality. In the case of his works on paper, a review of auction results in the past five years indicates that earlier works from the 1930s and 1940s appear to be more desirable, reaching higher prices than later drawings from the 1970s and 1980s. The subject property is an earlier work in the abstract style in very good condition. A survey of retail offerings located two pastel works on paper, one offered for $7,125 and the second for $14,000\n"),(ItemDescription,"Abstract composition of three figures rendered in colors on ivory paper"),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good"),(ItemMediumMaterialsTechniques,"graphite and pastel on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signed, dated lower center quadrant"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"26\" x 19 1/2\""),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"27\" x 20 3/4\""),(ItemSupportBaseFrame,"Red painted and gilt wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,""),(ItemArtistDate,""),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1969"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"15 Hanson #64")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5yzTHVI/AAAAAAAABx4/WErzxTrjaa8/s512/DSC03333.JPG", fileChksum = "a65da9e6609f6590ef0449079fbef5f1", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 400, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh3.ggpht.com/_s-XE562xJtM/Shwm5yzTHVI/AAAAAAAABx4/WErzxTrjaa8/s512/DSC03333.JPG", fileChksum = "a65da9e6609f6590ef0449079fbef5f1", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 400, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "f3a8178401afff79c8e85a4dd92155de", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 140, imageFileHeight = 179, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "43029cacdeccdc39f5725592e3fca02f", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1406, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh3.ggpht.com/_s-XE562xJtM/Shwm5yzTHVI/AAAAAAAABx4/WErzxTrjaa8/s512/DSC03333.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"pastel \" Standing Female Nude \" is available, framed and in excellent conditions and as described on our site. The asking price is $ 9.500, but because we are having a special summer promotion, I can give you a discount of 25%, for a net of $ 7.125, including packing, shipping and insurance Pasquale Ianetti Gallery, San Francisco\n\npastel Nude, was offered for $14,000\n"),(ItemCashValue,""),(ItemMarketAnalysis,"Hans Burhardt worked in a variety of media during his long career, including painting, drawing and print making. Known primarily as an abstract impressionist, Burkhardt also worked in a figurative style. Born in Switzerland, Burkhardt came to the United States in ??. He moved to Los Angeles in 1937 and was a well known and well regarded figure in the West Coast art scene. Burkhardt\8217s work has received recent attention due to two major projects, a retrospective of his work mounted at the University Gallery at California State University Northridge, and a comprehensive exhibition of his 1960s paintings presented at the Jack Rutberg Gallery in Los Angeles in 2008-2009. These projects have brought renewed public attention to the artists\8217s work reflected in a greater wolume of sales at auction.\n\nValue characteristics for Burhardt prints include date, size, condition and quality. In the case of his works on paper, a review of auction results in the past five years indicates that earlier works from the 1930s and 1940s appear to be more desirable, reaching higher prices than later drawings from the 1970s and 1980s. The subject property is an earlier work in the abstract style in very good condition. A survey of retail offerings located .. \n"),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good"),(ItemMediumMaterialsTechniques,"Pastel and graphite on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signature, date, lower right quadrant"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"23 1/2\" x 17 1/2\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"25 1/2\" x 19 1/2\""),(ItemSupportBaseFrame,"Glazed silver gilt frame"),(ItemSubjectOrTitle,"Untitled"),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1976"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"16 Hanson #62")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/ShwnA8lJxmI/AAAAAAAAByg/3D8KSo25_5Q/s512/DSC03343.JPG", fileChksum = "eb55c9e8f2b87448ad3fe1b7a5aaf54e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_s-XE562xJtM/ShwnA8lJxmI/AAAAAAAAByg/3D8KSo25_5Q/s512/DSC03343.JPG", fileChksum = "eb55c9e8f2b87448ad3fe1b7a5aaf54e", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 341, imageFileHeight = 512, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "fdd47b358ebab5ec1b0b0a51cbe41fbc", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 129, imageFileHeight = 194, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "940e99372ee848707da068ff9a2e03a9", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 1199, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_s-XE562xJtM/ShwnA8lJxmI/AAAAAAAAByg/3D8KSo25_5Q/s512/DSC03343.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes,"pastel \" Standing Female Nude \" is available, framed and in excellent conditions and as described on our site. The asking price is $ 9.500, but because we are having a special summer promotion, I can give you a discount of 25%, for a net of $ 7.125, including packing, shipping and insurance Pasquale Ianetti Gallery, San Francisco\n\npastel Nude, was offered for $14,000\n"),(ItemCashValue,""),(ItemMarketAnalysis,""),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good: Some creasing along the top edge and bottom left corner"),(ItemMediumMaterialsTechniques,"Charcoal on paper"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signature, date L/R"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"Sheet: 17 1/4\" x 23\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"27 1/4\" x 32 3/4\""),(ItemSupportBaseFrame,""),(ItemSubjectOrTitle,"Untitled"),(ItemCountryOrNationality,"Swiss American"),(ItemArtistDate,"1904-1994"),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1941"),(ItemTypeOfObject,"Work on paper"),(ItemLocation,""),(ItemInspectionDate,"5/28/09"),(ItemProjectNumber,""),(ItemClient,"Hanson"),(ItemInspector,""),(ItemDataSheetNumber,"17 Hanson #61")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwnC1KiGAI/AAAAAAAABys/EtMJnBzM5AU/s576/DSC03346.JPG", fileChksum = "10d359b37f2a0b3c109762e7f7cd7de6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 576, imageFileHeight = 469, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh6.ggpht.com/_s-XE562xJtM/ShwnC1KiGAI/AAAAAAAABys/EtMJnBzM5AU/s576/DSC03346.JPG", fileChksum = "10d359b37f2a0b3c109762e7f7cd7de6", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 576, imageFileHeight = 469, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "6365c7b50dbfa8a39588c266a3123f1d", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 175, imageFileHeight = 143, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "3d50f7e328332edf554de26815f2b282", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2211, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh6.ggpht.com/_s-XE562xJtM/ShwnC1KiGAI/AAAAAAAABys/EtMJnBzM5AU/s576/DSC03346.JPG"}]}},ReportItem {elemItem = Item {itemName = "", fields = [(ItemAdditionalNotes," \8220Flowers and Tears\8221 excellent condition offered for $8,500, The Caldwell Gallery\n\8220Pink Sage\8221 1960, 18 x 25, excellent condition, offered for $34,000, Jack Rutberg Gallery, Los Angeles\nLandscape, 1968, 20 x 24, excellent condition, offered for $?, Jack Rutberg Gallery, Los Angeles\n"),(ItemCashValue,""),(ItemMarketAnalysis,"Hans Burhardt worked in a variety of media during his long career, including painting, drawing and print making. Known primarily as an abstract impressionist, Burkhardt also worked in a figurative style. Born in Switzerland, Burkhardt came to the United States in ??. He moved to Los Angeles in 1937 and was a well known and well regarded figure in the West Coast art scene. Burkhardt\8217s work has received recent attention due to two major projects, a retrospective of his work mounted at the University Gallery at California State University Northridge, and a comprehensive exhibition of his 1960s paintings presented at the Jack Rutberg Gallery in Los Angeles in 2008-2009. These projects have brought renewed public attention to the artists\8217s work reflected in a greater wolume of sales at auction.\n\nValue characteristics for Burhardt prints include date, size, condition and quality. In the case of his works on paper, a review of auction results in the past five years indicates that earlier works from the 1930s and 1940s appear to be more desirable, reaching higher prices than later drawings from the 1970s and 1980s. The subject property is an earlier work in the abstract style in very good condition. A survey of retail offerings located \n"),(ItemDescription,""),(ItemCost,""),(ItemFax,""),(ItemTel,""),(ItemCity,""),(ItemAddress,""),(ItemExhibitionsAndPublications,""),(ItemProvenance,""),(ItemCondition,"Very Good. No evidence of damage"),(ItemMediumMaterialsTechniques,"oil on canvas"),(ItemVerso,""),(ItemRecto,""),(ItemPhotographNumbers,""),(ItemSignatureInscriptionsMarkings,"Signature, date, lower left quadrant"),(ItemItemDepth,""),(ItemItemWidth,""),(ItemItemHeight,""),(ItemItemMeasurement,"15 1/2\" x 19 1/4\" "),(ItemFrameDepth,""),(ItemFrameWidth,""),(ItemFrameHeight,""),(ItemFrameMeasurement,"23 1/2\" x 27 1/2\""),(ItemSupportBaseFrame,"Gilt wood frame"),(ItemSubjectOrTitle,""),(ItemCountryOrNationality,""),(ItemArtistDate,""),(ItemArtistOrMaker,"Hans Burkhardt"),(ItemItemDate,""),(ItemDateOrPeriod,"1975"),(ItemTypeOfObject,"Painting"),(ItemLocation,""),(ItemInspectionDate,""),(ItemProjectNumber,""),(ItemClient,""),(ItemInspector,""),(ItemDataSheetNumber,"18 Hanson #63")], images = [Pic {picSize = ImageSize {dim = TheHeight, size = 3.0, units = Inches}, picCrop = ImageCrop {topCrop = 0, bottomCrop = 0, leftCrop = 0, rightCrop = 0, rotation = 0}, picCaption = "", picImage = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SnoTIWHxquI/AAAAAAAABzk/D_Os7LpS43U/IMG00021_2.jpg", fileChksum = "61f9a7508ceddcbb780fe757b3cc48bf", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 622, imageFileHeight = 511, imageFileMaxVal = 255}), picEdited = Just (ImageFile {imageFile = File {fileSource = TheURI "http://lh5.ggpht.com/_EV171eW_Cmg/SnoTIWHxquI/AAAAAAAABzk/D_Os7LpS43U/IMG00021_2.jpg", fileChksum = "61f9a7508ceddcbb780fe757b3cc48bf", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 622, imageFileHeight = 511, imageFileMaxVal = 255}), picThumb = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "c13a71aec66c814dd585fc6e417fad20", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 174, imageFileHeight = 143, imageFileMaxVal = 255}), picPrinter = Just (ImageFile {imageFile = File {fileSource = TheBytes, fileChksum = "ed1f081afe85b605eace3b171c14252c", fileBytes = Nothing, fileMessages = []}, imageFileType = JPEG, imageFileWidth = 2191, imageFileHeight = 1800, imageFileMaxVal = 255}), picMessages = "<br>\nupdateOriginal http://lh5.ggpht.com/_EV171eW_Cmg/SnoTIWHxquI/AAAAAAAABzk/D_Os7LpS43U/IMG00021_2.jpg"}]}}], reportGlossary = [("Aquatint","The whole plate is covered with grains of resin called an aquatint ground, allowing acid to bite into the entire area, creating an overall grainy, tonal effect. This technique is often combined with etching."),("Drypoint","As in an engraving, the drypoint needle draws the image directly onto the plate. The residue copper is left on the side of the etched lines, which then collect the ink, creating a furry effect called burr."),("Etching","The plate is covered in an acid-resistant layer of wax called an etching ground. The image is then drawn into this surface with an etching needle. When covered with printing ink the lines hold the ink whilst the rest of the plate repels it."),("Lithograph","From the Greek lithos, stone and graphe, writing. This printing process is unlike both intaglio and relief processes, both of which involve cutting into the plate. Lithography relies on the principle that grease and water will repel each other. The image is drawn in a greasy substance onto a lithographic stone. The stone is then dampened with water and the greasy printing ink adheres only to the drawing."),("Second edition","Often an artist will work on a composition to a certain point, and stop to print an impression of it. This single stage in the evolution of this image is called a state. Each time the composition is changed a new state of the print is created. These changes can range from the addition of a plate signature to drastic alterations in the composition. Today artist\8217s will frequently choose to edition a state before moving on with the composition. "),("State","Often an artist will work on a composition to a certain point, and stop to print an impression of it. This single stage in the evolution of this image is called a state. Each time the composition is changed a new state of the print is created. These changes can range from the addition of a plate signature to drastic alterations in the composition. Today artist\8217s will frequently choose to edition a state before moving on with the composition. ")], reportSources = [("Experts Consulted","Jennifer M. Wood\nConsulting and Appraisal Services\n3122 Arizona Avenue, Suite C\nSanta Monica, CA  90404\nPhone: 310 633 1736\nEmail: jenniferwood@roadrunner.com\n"),("Internet Sources","Artfact.com\nhttp://www.artfact.com\n\nArtnet.com\nhttp://www.artnet.com\n\nInternational Fine Print Dealers Association\nhttp://www.printdealers.com\n\nTrocadero.com\nhttp://www.trocadero.com"),("Print Sources","Bartsch, Adam, Zerner, Henri ed., <em> The Illustrated Bartsch: Italian Artists of the Sixteenth Century School of Fountainbleau</em> New York: Abaris Books, 1979.\n\nBouvet, Francis <em>Bonnard: The Complete Graphic Work </em> New York: Rizzoli, 1981. \n\nBrown, Betty Ann, ed. <em> Hans Burkhardt: The California State University, Northridge Collection</em> Northridge: California State University, Northridge, 2008.\n\nDelteil, Loys, <em> Le Peintre-Graveur Illustre (XIX et XX Siecles) Tome Cinquieme, Corot </em> republication of the 1910 Paris edition, New York: Da Capo Press, 1968.\n\nGascoigne, Bamber, <em> How to Identify Prints: A Complete Guide to Manual and mechanical Processes from Woodcut to Inkjet</em>, 2nd edition, London: Thames & Hudson,2004.\n\nIves, Colta, Giambruni, Newman Sasha M. <em> Pierre Bonnard: The Graphic Art</em> New York: The Metropolitan Museum of Art, 1989\n\nLeymarie, Jean, Melot, Michel, <em> The Graphic Works of the Impressionists: Manet, Pissarro, Renoir, Cezanne, Sisley</em> New York: Harry N. Abrams, Inc., 1971.\n\nRutberg, Jack, <em> Hans Burkhardt: Paintings of the 1960s</em> Los Angeles: Jack Rutberg Fine Arts, Inc., 2008.\n\nSoucy, Patricia and Janella N. Smyth, eds. <em>The Appraisal of Personal Property: Principles, Theories and Practice Methods for the Professional Appraiser.</em> Washington, D.C.: American Society of Appraisers, 1994.\n\nStella, Joseph G., <em> The Graphic Works of Renoir </em> London: Lund Humphries, 1975.\n\n<em>Uniform Standards of Professional Appraisal Practice, the 2008-2009 Edition</em> (USPAP) Washington, D.C.: The Appraisal Foundation, 2008."),("Galleries and Auction Houses","Allinson Gallery, Inc.\n3 Dunham Pond Road East \nStorrs, CT 06268-2022 \n860-429-2322\nhttp://www.allinsongallery.com\n\nBonham's & Butterfield's\n220 San Bruno Avenue\nSan Francisco, CA 94203\n415-86-7500\nhttp://www.bonhams.com\n\nChristie's\n20 Rockefeller Plaza\nNew York, NY 10020\n212-636-2000\nhttp://www.christies.com\n\nDavidson Galleries\n313 Occidental Ave S.\nSeattle, WA 98104\n206 624.6700\nhttp://www.davidsongalleries.com\n\nFigaro Gallery\n703-491-1000\nwww.justfineart.com\nE-mail: inquire@figarogallery.com\n\nGalerie Bassenge Kunst-und Buchauktionen\nErdener Strabe 5a\n14193 Berlin-Grunewald\nGerman\n+49-(0)30-893 80 29-0\n\nI. M. Chait Gallery\n9330 Civic Center Drive\nBeverly Hills, CA 90210\n(800) 775-5020\nhttp://www.chait.com\n\nJack Rutberg Fine Arts, Inc.\n357 North La Brea Avenue\nLos Angeles, CA 90036\n323-938-5222\nhttp://www.jackrutbergfinearts.com\n\nKamp Galleries\n563 Lincoln Avenue\nWinnetka, IL 60093\n847-441-7999\nhttp://www.kampgallery.com\n\nLeslie Hindman Auctioneers\n\nMasterworks Fine Art Inc.\n13470 Campus Drive\nOakland, CA 94619\nhttp://www.masterworksfineart.com\n510-777-9970\nSpaightwood Galleries")], reportLetterOfTransmittal = "According to our contract signed by you on [ContractDate], [we] have\nprepared the following [ReportType] report for the property\nrepresented as belonging to you.  The subject properties consists of:\n[BriefDescription].\n\nYou provided [us] with the information that indicates that you own the\nsubject [properties] listed above in their entirety.  [We] examined the\n[properties] in the presence of [Client] at [InspectionAddress]\non [InspectionDate].\n\nThis appraisal report provides the appraisal conclusion of\n[ValueType] for the intended use of [IntendedUse].  \n[ValueTypeDescription]\n\nThis report will function as documented evidence of [theseappraisers]\nhaving witnessed the subject [properties], and identified, described and\nvalued the subject [properties] in the context of estimating [ValueType].\n\n\nAfter thorough consideration of available facts and statistics, on\n[EffectiveDate], the [ValueType] for the subject [properties] is\n<b>[TotalValueAsText] ([TotalValue])</b>.\nThe effective date of value for this appraisal is [EffectiveDate].\nThis figure does not reflect a consideration for related sales\ntaxes and delivery/installation charges, which can vary.  This\nprofessional opinion is based upon [our] knowledge and experience in the\nvaluation of fine arts.  This appraisal is based on [our] research and\nanalysis of the [properties] and knowledge of relevant markets where the\n[properties] [are] traded.\n\nThis is a [ReportType] report that was prepared to conform to\nthe Code of Ethics of the American Society of Appraisers and the\ncurrent version of the [USPAP],\npublished by the Appraisal Foundation.  We have\nno past, present or contemplated future interest in the [properties]\nincluded in this report.  Our fee was not contingent on any value\nfinding.\n\nThis [ReportLength] page report consists of a letter of transmittal, scope of\nwork, definitions, privacy statement, assumptions and limiting\nconditions, certification, description, photographs, narrative,\nglossary, bibliography and credentials.  All these elements must be\npresent for the value conclusion to be valid.\n\nThe information in this report is confidential.  All notes,\nelectronic, work copies and photographs remain the property of the\nappraiser and are retained for five years, and/or two years after the\nconclusion of any litigation.  All documents are shredded or otherwise\ndefaced prior to disposal.\n\nThe submission of this report completes [our] assignment and\nobligations.  Any further consultation, preparation of additional\nreports, testimony, or appearances in court or before any governmental\nforum in reference to the [properties] described in this appraisal will\nrequire additional fees to cover the required services.  The\naforementioned fees will be billed at the rate commensurate with the\ntiming of the appraisal.\n\nThank you for allowing [us] to be of service to you.\n", reportScopeOfWork = "Thompson & Martinez Fine Art Appraisals, Inc. was hired by [Client] to reach an opinion of the [ValueType] of [BriefDescription], for the intended use of [IntendedUse].  Based on [our] discussion with [Client], as well as the intended use of the report, the type and location of the subject [properties] and the effective date of valuation, the [appraisers] [have] selected the type of value, the relevant market, the information researched and the analysis to be applied.  To complete this assignment [we] consulted authoritative publications and on-line sources, reviewed published and on-line market data and interviewed knowledgeable dealers and others regarding the artists, their work, and the relevant markets. Jennifer Wood, ASA assisted in the inspection of the prints by Francisco de Goya, Henri Toulouse-Lautrec and Auguste Renoir.\n\nThis report is intended to be an objective, independent, unbiased and credible analysis of known facts.  During the course of our examination, research and analysis, [we] have endeavored to be accurate and to use due diligence.  This opinion of value was developed using appraisal methods that are within the accepted standard of care for the appraisal profession and within the boundaries of typical practice. \n\nThis appraisal document summarizes the majority of information analyzed, the appraisal procedures followed, and the reasoning that supports the analysis, opinions and conclusions under the requirements of the [USPAP].  It is considered a summary report.  The documentation substantiating the value reported is held in the appraiser's file. ", reportCertification = ["To the best of [our] knowledge and belief, the statements of fact continued in this report are true and correct.  The value conclusions are based on [our] background and experience outlined in [our] credentials attached at the end of this report.","The reported analyses, opinions and conclusions are limited only by the reported assumptions and liming conditions.  They are [our] personal, impartial and unbiased professional analyses, opinions and conclusions.","[Our] analyses, opinions and conclusions in this report were developed and prepared in conformity with the [USPAP]. ","Unless otherwise stated, the value given in the Appraisal Report represents the opinion of value as of the date of inspection.","[We] have no bias with respect to the property that is the subject of this report, or to the parties involved with this assignment.","[Our] compensation for completing this assignment is not contingent upon development or reporting of a predetermined value or direction in value that favors the cause of the client, the amount of the value opinion, the attainment of a stipulated result, or the occurrence of a subsequent event directly related to the intended use of this appraisal.","[We] have made a personal inspection of the property that is the subject of this report.","[These] [appraisers] prepared most conclusions and opinions set forth in this Appraisal Report. Jennifer M. Wood, ASA, assisted in the inspection of prints by Auguste Renoir, Henri Toulouse-Lautrec and Francisco de Goya.","Although all due care has been taken to ensure that any statement as to the authorship, attribution, origin, date, age, provenance, condition, and value of any item of personal property included in this report is accurate, all such statements are to the best of the knowledge of the [appraisers] who completed this report, and are not to be relied upon as representation of fact."], reportLimitingConditions = ["Title to the property and present ownership of the property is assumed to be good and marketable unless otherwise stated.  Further, the appraisal is made with the understanding that the present ownership of the subject property includes all rights that may be lawfully owned and is not subject to any encumbrances or indebtedness.  The Clients stated that they own the property; ownership was not researched or confirmed.","The [appraisers] [have] no present or prospective interest in the property that is the subject of this report, and no specified personal interest with respect to the parties involved.  The reported analyses, opinions and conclusions are limited only by the reported assumptions and limiting conditions, and are personal, impartial and unbiased professional analyses, opinions and conclusions.","The [appraisers'] analyses, opinions, and conclusions [were] developed, and this report has been prepared, in conformity with the Code of Ethics as set forth by the American Society of Appraisers and the [USPAP] of the Appraisal Foundation, Washington, D.C.","The Appraisal is based on visual inspection of the appraised item.  No tests of any kind have been administered unless so stated.  Measurements and photographs are made under ``field conditions,'' and therefore will be as accurate as possible.  Measurements are listed in the order of height, width and depth, unless otherwise indicated.","The opinion of value expressed in this Appraisal Report is the result of and is subject to the data and research described in detail in the Appraisal Report.  In the preparation of this report, the [appraisers] [have] consulted a variety of sources that are documented either in the report or are available in the [appraisers'] files.  These sources are assumed to be reliable and accurate; however, the [appraisers] [take] no responsibility for errors or omissions contained in these sources.","This Appraisal Report, or a copy of this Appraisal Report, may be transmitted to a third party or legal entity only in its entirety and only with the expressed written consent of the appraisal client.","Disclosure of the contents of the Appraisal Report is governed by the Bylaws and Regulations of the Appraisal Foundation [USPAPShort].  Neither all nor any part of the contents of this report (especially any conclusions as to the value or the identity of the Appraisers) shall be disseminated to the public through advertising, public relations, news, sales, or the media without the prior written consent and approval of the appraiser.","Information furnished by the owner regarding property history is believed to be reliable; however, no warranty is given for its accuracy.  ","The [appraisers] assumes no responsibility for unforeseen changes in market conditions, nor is the inability to locate a potential buyer or replacement item, at the appraised value, if such is attempted by the client or any other person. The value conclusion is not an offer to buy or sell the property.","In the event of a dispute involving interpretation or application of this report, the American Society of Appraisers recommends that such a dispute shall be referred to a neutral third-party mediation service.  The cost of such mediation shall be borne equally by both parties.  In the event mediation is not successful, the parties should agree to submit the dispute to binding arbitration under the laws of the State of California.","The [appraisers] [have] personally inspected the subject property.  The condition of the property is described sufficiently for its type, with any visible and/or relevant repairs noted. The condition of the property may change after inspection and the effective date of valuation.","The opinion of value expressed in the Appraisal Report is the result of and is subject to the data and research described in detail in the Appraisal Report and on data retained in the [appraisers'] files.  The [appraisers] [have] not knowingly withheld any significant information.","Appraisers witness, identify and value subject property but do not claim authentication of the subject properties, unless otherwise noted.","This Appraisal report is valid only for the intended use, intended users and for the type of value stated in the appraisal.  The values reported in this appraisal are valid only on the effective date of the appraisal [EffectiveDate].","The submission of this report completes [our] assignment and obligations.  Any further consultation, preparation of additional reports, testimony, or appearances in court or before any governmental forum in reference to the property described in this appraisal will require additional fees to cover the required services."], reportPrivacyPolicy = "Thompson & Martinez Fine Art Appraisals, Inc. makes significant efforts to keep all appraisal-related data confidential in compliance with the current version of the [USPAP] as published by The Appraisal Foundation. We will not disclose confidential appraisal data to any third party without the written permission of the rightful owner of that data. We will not release a copy of an appraisal you have contracted through Thompson & Martinez Fine Art Appraisals, Inc. to any third party without the written permission of the client.\n\nOur policy specifies the protection of client information in the following ways:\n * We limit access to all information only to those who have a business or professional reason to know and only to nonaffiliated parties as required by law.\n * We maintain a secure office and computer environment to ensure that your information is not placed at unreasonable risk.\n * The categories of non-public personal information that we collect from a client depend upon the scope of the client's appraisal engagement.  This includes information about personal property assets, information about tax identification numbers, and other non-public information necessary.  Additionally, it may include information about transactions between clients and third parties and information form third party sources.\n * Unaffiliated third parties that require access to our clients' personal information, such as Federal and State tax regulators, insurance companies, and the like may only review company records as permitted under law.\n * We do not provide client information to mailing list vendors or solicitors for any reason.\n * Personally identifiable information will be maintained during the time a client is a client of the company and for 5 years after the appraisal has been completed and/or 2 years after litigation as well as the ethical considerations as specified by [USPAPShort].  After this required period of record retention, all such information will be destroyed.\n", reportMessages = "runLatex -> success", reportPerms = Just (Permissions {owner = Username {unUsername = "lydiat"}, writers = [Username {unUsername = "Natasha"},Username {unUsername = "david"},Username {unUsername = "lydia"}], readers = []}), revisionInfo = defaultValue}

test8 = TestCase $ assertEqual "merging reports" Nothing $
        trace "\n\nStarting test8" (return ()) >> 
        threeWayMerge' o l r
